#!/usr/bin/env python3

# Copyright (C) 2012-2015  ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

import unittest
from time import sleep
import datetime

import logging
logger = logging.getLogger(__name__)

from lofar.sas.tmss.test.test_environment import TMSSTestEnvironment

from lofar.sas.tmss.services.feedback_handling import HybridFeedbackMessageHandler, create_service
from lofar.common.test_utils import integration_test, exit_with_skipped_code_if_skip_integration_tests

exit_with_skipped_code_if_skip_integration_tests()

@integration_test
class TestFeedbackHandlingService(unittest.TestCase):
    '''
    Tests for the FeedbackHandlingService
    '''

    # a chunk of feedback as it comes from the correlator for each dataproduct
    # string can be filled in with .format(subband=...)
    feedback_chunk = """Observation.DataProducts.Output_Correlated_[{subband}].SAP=0
Observation.DataProducts.Output_Correlated_[{subband}].centralFrequency=102734375.000000
Observation.DataProducts.Output_Correlated_[{subband}].channelWidth=3051.757812
Observation.DataProducts.Output_Correlated_[{subband}].channelsPerSubband=64
Observation.DataProducts.Output_Correlated_[{subband}].duration=0
Observation.DataProducts.Output_Correlated_[{subband}].fileFormat=AIPS++/CASA
Observation.DataProducts.Output_Correlated_[{subband}].filename=L2000000_SAP000_SB{subband:03d}_uv.MS
Observation.DataProducts.Output_Correlated_[{subband}].integrationInterval=1.006633
Observation.DataProducts.Output_Correlated_[{subband}].location=CEP4:/data/test-projects/high/L2000000/uv
Observation.DataProducts.Output_Correlated_[{subband}].percentageWritten=0
Observation.DataProducts.Output_Correlated_[{subband}].size=0
Observation.DataProducts.Output_Correlated_[{subband}].startTime=2021-01-29 12:39:00
Observation.DataProducts.Output_Correlated_[{subband}].stationSubband={subband}
Observation.DataProducts.Output_Correlated_[{subband}].storageWriter=LOFAR
Observation.DataProducts.Output_Correlated_[{subband}].storageWriterVersion=3
Observation.DataProducts.Output_Correlated_[{subband}].subband={subband}"""

    @classmethod
    def setUpClass(cls) -> None:
        cls.tmss_test_env = TMSSTestEnvironment(populate_schemas=True)
        cls.tmss_test_env.start()
        cls.test_data_creator = cls.tmss_test_env.create_test_data_creator()


    @classmethod
    def tearDownClass(cls) -> None:
        cls.tmss_test_env.stop()

    def test_01_hybrid_feedback_handler_business_logic_without_messagebuses(self):
        handler = HybridFeedbackMessageHandler(rest_client_creds_id=self.tmss_test_env.client_credentials.dbcreds_id)

        # do not start_handling, cause it would connect to the old and obsolete qpid bus, for which we don't have a test broker.
        # So, we skip testing the qpid messagebus part, and only test the handling of raw feedback in this service and tmss.
        with handler._tmss_client:
            with self.tmss_test_env.create_tmss_client() as tmss_client:
                # create a subtask with some output dataproducts with initial emtpy feedback
                dataproduct_feedback_templates = tmss_client.get_path_as_json_object('dataproduct_feedback_template')
                empty_dataproduct_feedback_template = next(x for x in dataproduct_feedback_templates if x['name']=='empty')

                subtask_templates = tmss_client.get_path_as_json_object('subtask_template')
                obs_subtask_template = next(x for x in subtask_templates if x['name']=='observation control')

                subtask = self.test_data_creator.post_data_and_get_response_as_json_object(self.test_data_creator.Subtask(specifications_template_url=obs_subtask_template['url']), '/subtask/')
                subtask_id = subtask['id']
                subtask_output = self.test_data_creator.post_data_and_get_response_as_json_object(self.test_data_creator.SubtaskOutput(subtask_url=subtask['url']), '/subtask_output/')
                NUM_DATAPRODUCTS = 4
                for i in range(NUM_DATAPRODUCTS):
                    self.test_data_creator.post_data_and_get_response_as_json_object(self.test_data_creator.Dataproduct(subtask_output_url=subtask_output['url'],
                                                                                                                        filename="L%d_SAP000_SB%03d_uv.MS" % (subtask_id, i),
                                                                                                                        dataproduct_feedback_template_url=empty_dataproduct_feedback_template['url']),
                                                                                     '/dataproduct/')

                # check the intial dataproducts have empty feedback
                dataproducts = tmss_client.get_subtask_output_dataproducts(subtask_id=subtask_id)
                self.assertEqual(NUM_DATAPRODUCTS, len(dataproducts))
                for dataproduct in dataproducts:
                    self.assertEqual(empty_dataproduct_feedback_template['url'], dataproduct['feedback_template'])

                # TMSS only accepts feedback in finishing state
                from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions, Subtask
                set_subtask_state_following_allowed_transitions(Subtask.objects.get(id=subtask_id), 'finishing')

                # test handler busines logic without messagebuses
                # assume the old qpid messagebus just works, and delivers proper feedback chuncks in the payload.

                # create the chunks of actual feedback and upload/process them
                for i in range(NUM_DATAPRODUCTS):
                    feedback_dp = self.feedback_chunk.format(subband=i)
                    handler.process_feedback_and_set_to_finished_if_complete(subtask_id, feedback_dp)
                    if i < NUM_DATAPRODUCTS-1:
                        self.assertEqual('finishing', tmss_client.get_subtask(subtask_id=subtask_id)['state_value'])

                # check the updated dataproducts have filled in feedback
                feedback_dataproduct_feedback_template = next(x for x in dataproduct_feedback_templates if x['name']=='feedback')
                dataproducts = tmss_client.get_subtask_output_dataproducts(subtask_id=subtask_id)
                self.assertEqual(NUM_DATAPRODUCTS, len(dataproducts))
                for dataproduct in dataproducts:
                    self.assertEqual(feedback_dataproduct_feedback_template['url'], dataproduct['feedback_template'])
                    # check a few keys, specific for this test.
                    # see t_adapter for full conversion unit test
                    self.assertEqual('lofarstman', dataproduct['feedback_doc']['samples']['writer'])
                    subband_from_filename = int(dataproduct['filename'].split('_')[2][2:])
                    self.assertEqual([subband_from_filename], dataproduct['feedback_doc']['frequency']['subbands'])



logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

if __name__ == '__main__':
    #run the unit tests
    unittest.main()
