#!/usr/bin/env python3

# Copyright (C) 2012-2015  ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

import unittest
import time
import datetime
import logging
logger = logging.getLogger('lofar.' + __name__)
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)


from lofar.sas.tmss.test.test_environment import TMSSTestEnvironment

from lofar.sas.tmss.services.precalculations_service import create_service_job_for_sunrise_and_sunset_calculations
from lofar.common.test_utils import integration_test


@integration_test
class TestPreCalculationService(unittest.TestCase):
    """
    Tests for the TMSSPreCalculationsServiceJob
    It will check the number of items created of the StationTimeline model based on the input of the service to start
    It will not check the content of the sunrise/sunset data of the  StationTimeline model itself
    Note that 1 day calculation will take about 6 seconds (my local developer environment)
    So the assumption was that the calculation takes about 6 sec BUT the build environment tooks ever longer,  11
    a 14 seconds!!!! Some 'timing' parameters where adjusted in this testcase but maybe not robust enough
    On the other hand if the build system is getting even more slower than this  there should be really be a doubt
    about the build system.
    """

    @classmethod
    def setUpClass(cls) -> None:
        """
        Populate schema to be able to retrieve all stations
        """
        cls.tmss_test_env = TMSSTestEnvironment(populate_schemas=True)
        cls.tmss_test_env.start()
        cls.test_data_creator = cls.tmss_test_env.create_test_data_creator()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.tmss_test_env.stop()

    def setUp(self) -> None:
        """
        Start every testcase with 'clean' StationTimeline model
        """
        from lofar.sas.tmss.tmss.tmssapp.models.calculations import StationTimeline
        StationTimeline.objects.all().delete()

    def test_all_stations_calculated_for_one_day(self):
        """
        Test if creating and starting, followed by stopping the (pre)calculation service results in 'one day'
        of StationTimeline data for all stations
        Note that 1 day calculation will take about 6 seconds
        """
        # Import here otherwise you get
        # "django.core.exceptions.ImproperlyConfigured: Requested setting INSTALLED_APPS, but settings are not configured. You must either define the environment variable DJANGO_SETTINGS_MODULE or call settings.configure() before accessing settings."
        from lofar.sas.tmss.tmss.tmssapp.conversions import get_all_stations
        from lofar.sas.tmss.tmss.tmssapp.models.calculations import StationTimeline

        nbr_stations = len(get_all_stations())
        # Initially there should be no data
        self.assertEqual(len(StationTimeline.objects.all()), 0)
        # Now we are going to create and start the calculation service with a wait time of 60 sec,
        # nbr days to calculate ahead is 1 and nbr days before today 1 ->  so only 'yesterday' should be created
        job = create_service_job_for_sunrise_and_sunset_calculations(60, 1, 1)
        job.start()
        job.stop()
        # Check what have been created
        st_objects = StationTimeline.objects.all()
        self.assertEqual(len(st_objects), nbr_stations)
        # lets check with the timestamp of today, that should be zero
        st_objects = StationTimeline.objects.filter(timestamp=datetime.date.today())
        self.assertEqual(len(st_objects), 0)
        # lets check with the timestamp in future, that should be zero
        st_objects = StationTimeline.objects.filter(timestamp__gt=datetime.date.today())
        self.assertEqual(len(st_objects), 0)
        # lets check with the timestamp yesterday, that should be equal to the number of all stations
        st_objects = StationTimeline.objects.filter(timestamp=datetime.date.today()-datetime.timedelta(days=1))
        self.assertEqual(len(st_objects), nbr_stations)

    def test_all_stations_calculated_for_multiple_days_with_one_trigger(self):
        """
        Test if creating and starting, followed by stopping the (pre)calculation service results in 'multiple day'
        of StationTimeline data for all stations
        Note that 4 days calculation will take about 30 seconds
        """
        # Import here otherwise you get
        # "django.core.exceptions.ImproperlyConfigured: Requested setting INSTALLED_APPS, but settings are not configured. You must either define the environment variable DJANGO_SETTINGS_MODULE or call settings.configure() before accessing settings."
        from lofar.sas.tmss.tmss.tmssapp.conversions import get_all_stations
        from lofar.sas.tmss.tmss.tmssapp.models.calculations import StationTimeline

        nbr_stations = len(get_all_stations())
        # Initially there should be no data
        self.assertEqual(len(StationTimeline.objects.all()), 0)
        # Now we are going to create and start the calculation service with a interval of 120 sec,
        # nbr days to calculate ahead is 4 and nbr days before today 2 ->  so 'day before yesterday, 'yesterday',
        # 'today' and 'tomorrow' should be created
        job = create_service_job_for_sunrise_and_sunset_calculations(120, 4, 2)
        job.start()
        job.stop()
        # Check what have been created
        st_objects = StationTimeline.objects.all()
        self.assertEqual(len(st_objects), 4*nbr_stations)
        # lets check with the timestamp of today, that should be equal to the number of all stations
        st_objects = StationTimeline.objects.filter(timestamp=datetime.date.today())
        self.assertEqual(len(st_objects), nbr_stations)
        # lets check with the timestamp in future, that should be equal to the number of all stations
        st_objects = StationTimeline.objects.filter(timestamp__gt=datetime.date.today())
        self.assertEqual(len(st_objects), nbr_stations)
        # lets check with the timestamp in the past, that should be equal to the 2 times number of all stations
        st_objects = StationTimeline.objects.filter(timestamp__lt=datetime.date.today())
        self.assertEqual(len(st_objects), 2*nbr_stations)

    def test_all_stations_calculated_after_interval(self):
        """
        Test if creating and starting, waiting for period (25 seconds), followed by stopping the (pre)calculation service results
        in 'multiple day' of StationTimeline data for all stations.
        It will test the scheduler with interval of 20 seconds, so three days should be calculated
        """
        # Import here otherwise you get
        # "django.core.exceptions.ImproperlyConfigured: Requested setting INSTALLED_APPS, but settings are not configured. You must either define the environment variable DJANGO_SETTINGS_MODULE or call settings.configure() before accessing settings."
        from lofar.sas.tmss.tmss.tmssapp.conversions import get_all_stations
        from lofar.sas.tmss.tmss.tmssapp.models.calculations import StationTimeline

        nbr_stations = len(get_all_stations())
        # Initially there should be no data
        self.assertEqual(len(StationTimeline.objects.all()), 0)
        # Now we are going to create and start the calculation service with a interval of 20 sec (smaller will not make sense),
        # nbr days to calculate ahead is 1 and nbr days before today 0 ->  so it start with 'today' and after 20 seconds
        # 'tomorrow' etc..,
        job = create_service_job_for_sunrise_and_sunset_calculations(20, 1, 0)
        job.start()
        time.sleep(25)
        job.stop()
        # Check what have been created with interval of 20 seconds we should have two days
        st_objects = StationTimeline.objects.all()
        self.assertEqual(len(st_objects), 2*nbr_stations)
        # lets check with the timestamp of today, that should be equal to the number of all stations
        st_objects = StationTimeline.objects.filter(timestamp=datetime.date.today())
        self.assertEqual(len(st_objects), nbr_stations)
        # lets check with the timestamp in future, that should be equal to the number of all stations
        st_objects = StationTimeline.objects.filter(timestamp__gt=datetime.date.today())
        self.assertEqual(len(st_objects), nbr_stations)
        # lets check with the timestamp in the past, that should be equal to zero
        st_objects = StationTimeline.objects.filter(timestamp__lt=datetime.date.today())
        self.assertEqual(len(st_objects), 0)

    def test_all_stations_calculated_for_when_interval_time_is_too_small(self):
        """
        Check that if the interval time < calculation time it does not lead to exception
        Test if creating and starting, waiting for period (20 seconds), followed by stopping the (pre)calculation service results
        in 'multiple day' of StationTimeline data for all stations.
        It will test the scheduler with interval of 2 seconds, which smaller than ~6 seconds
        Stopping after 20 seconds should make 2 days calculated
        """
        # Import here otherwise you get
        # "django.core.exceptions.ImproperlyConfigured: Requested setting INSTALLED_APPS, but settings are not configured. You must either define the environment variable DJANGO_SETTINGS_MODULE or call settings.configure() before accessing settings."
        from lofar.sas.tmss.tmss.tmssapp.conversions import get_all_stations
        from lofar.sas.tmss.tmss.tmssapp.models.calculations import StationTimeline

        nbr_stations = len(get_all_stations())
        # Initially there should be no data
        self.assertEqual(len(StationTimeline.objects.all()), 0)
        # Now we are going to create and start the calculation service with an interval of 2 sec
        # nbr days to calculate ahead is 1 and nbr days before today 0 ->  so it start with 'today' and after ~6 seconds
        # 'tomorrow' etc..
        job = create_service_job_for_sunrise_and_sunset_calculations(2, 1, 0)
        job.start()
        time.sleep(20)
        job.stop()
        # Check what have been created with interval of 2 seconds we should have two days
        st_objects = StationTimeline.objects.all()
        self.assertGreaterEqual(len(st_objects), 2 * nbr_stations)
        # lets check with the timestamp of today, that should be equal to the number of all stations
        st_objects = StationTimeline.objects.filter(timestamp=datetime.date.today())
        self.assertEqual(len(st_objects), nbr_stations)

    @unittest.skip("TODO: fix blinking test due to incorrect synchronization issues.")
    def test_all_stations_calculated_with_two_jobs_started(self):
        """
        Test if starting two jobs of (pre)calculation service results in no Exception, there are no
        duplicate data stored (covered by the Constraints in the model)
        It will test the scheduler with interval of 20 seconds, to make sure one interval after the start has been passed
        """
        # Import here otherwise you get
        # "django.core.exceptions.ImproperlyConfigured: Requested setting INSTALLED_APPS, but settings are not configured. You must either define the environment variable DJANGO_SETTINGS_MODULE or call settings.configure() before accessing settings."
        from lofar.sas.tmss.tmss.tmssapp.conversions import get_all_stations
        from lofar.sas.tmss.tmss.tmssapp.models.calculations import StationTimeline

        nbr_stations = len(get_all_stations())
        # Initially there should be no data
        self.assertEqual(len(StationTimeline.objects.all()), 0)
        # Now we are going to create and start the calculation service with an interval of 20 sec
        # nbr days to calculate ahead is 1 and nbr days before today 0 ->  so it start with 'today' and after ~20 seconds
        # 'tomorrow' etc..
        job = create_service_job_for_sunrise_and_sunset_calculations(20, 1, 0)
        job2 = create_service_job_for_sunrise_and_sunset_calculations(20, 1, 0)

        job.start()
        job2.start()
        time.sleep(22)
        job.stop()
        job2.stop()
        # Check what have been created should only be today and tomorrow
        st_objects = StationTimeline.objects.all()
        self.assertGreaterEqual(len(st_objects), 2 * nbr_stations)


if __name__ == '__main__':
    #run the unit tests
    unittest.main()