#!/usr/bin/env python3

# Copyright (C) 2021
# ASTRON (Netherlands Institute for Radio Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.


import logging
import os
from optparse import OptionParser, OptionGroup
from requests import session

logger = logging.getLogger(__name__)

from lofar.common.dbcredentials import DBCredentials
from lofar.sas.tmss.client.tmssbuslistener import *
from lofar.sas.tmss.client.tmss_http_rest_client import TMSSsession

class TMSSEventMessageHandlerForSlackWebhooks(TMSSEventMessageHandler):
    '''
    '''
    def __init__(self, slack_url: str, rest_client_creds_id: str="TMSSClient"):
        super().__init__(log_event_messages=False)
        self.slack_url = slack_url
        self.slack_session = session()
        self.tmss_client = TMSSsession.create_from_dbcreds_for_ldap(rest_client_creds_id)

    def start_handling(self):
        self.tmss_client.open()
        super().start_handling()

    def stop_handling(self):
        super().stop_handling()
        self.tmss_client.close()
        self.slack_session.close()

    def post_to_slack_webhook(self, message: str):
        logger.info("post_to_slack_webhook: %s", message)
        # post to slack, see https://api.slack.com/messaging/webhooks
        self.slack_session.post(url=self.slack_url, json={"text": message})

    def onTaskBlueprintStatusChanged(self, id: int, status: str):
        task = self.tmss_client.get_path_as_json_object('task_blueprint/%s' % (id,))
        task_ui_url = task['url'].replace('/api/task_blueprint/', '/task/view/blueprint/')
        task_url = "<%s|\'%s\' id=%s>" % (task_ui_url, task['name'], task['id'])
        self.post_to_slack_webhook("%s - Task %s status changed to %s" % (self._get_formatted_project_scheduling_unit_string(task['scheduling_unit_blueprint_id']),
                                                                                         task_url, status))

    def onSchedulingUnitBlueprintCreated(self, id: int):
        scheduling_unit = self.tmss_client.get_path_as_json_object('scheduling_unit_blueprint/%s' % (id,))
        self.post_to_slack_webhook("%s was created\ndescription: %s" % (self._get_formatted_project_scheduling_unit_string(id),
                                                                        scheduling_unit['description'] or "<no description>"))

    def onSchedulingUnitBlueprintStatusChanged(self, id: int, status:str):
        self.post_to_slack_webhook("%s status changed to %s" % (self._get_formatted_project_scheduling_unit_string(id), status))

    def _get_formatted_project_scheduling_unit_string(self, scheduling_unit_blueprint_id: int) -> str:
        scheduling_unit = self.tmss_client.get_path_as_json_object('scheduling_unit_blueprint/%s' % (scheduling_unit_blueprint_id,))
        scheduling_unit_draft = self.tmss_client.get_url_as_json_object(scheduling_unit['draft'])
        scheduling_set = self.tmss_client.get_url_as_json_object(scheduling_unit_draft['scheduling_set'])
        project = self.tmss_client.get_url_as_json_object(scheduling_set['project'])

        su_ui_url = scheduling_unit['url'].replace('/api/scheduling_unit_blueprint/', '/schedulingunit/view/blueprint/')
        project_ui_url = project['url'].replace('/api/project/', '/project/view/')
        result = "Project <%s|\'%s\'> - SchedulingUnit <%s|\'%s\' id=%s>" % (project_ui_url, project['name'],
                                                                             su_ui_url, scheduling_unit['name'], scheduling_unit['id'])
        return result


def create_service(slack_url: str, rest_client_creds_id:str="TMSSClient", exchange: str=DEFAULT_BUSNAME, broker: str=DEFAULT_BROKER):
    return TMSSBusListener(handler_type=TMSSEventMessageHandlerForSlackWebhooks,
                           handler_kwargs={'slack_url': slack_url, 'rest_client_creds_id': rest_client_creds_id},
                           exchange=exchange, broker=broker)


def main():
    # make sure we run in UTC timezone
    os.environ['TZ'] = 'UTC'

    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

    # Check the invocation arguments
    parser = OptionParser('%prog [options]',
                          description='run the tmss_slack_webhook_service which listens for TMSS event messages on the messagebus, and posts the updates on the slack webhook api.')

    group = OptionGroup(parser, 'Slack options')
    parser.add_option_group(group)
    group.add_option('-s', '--slack_credentials', dest='slack_credentials', type='string', default='TMSSSlack', help='credentials name (for the lofar credentials files) containing the TMSS Slack Webhook URL, default: %default')

    group = OptionGroup(parser, 'Django options')
    parser.add_option_group(group)
    group.add_option('-R', '--rest_credentials', dest='rest_credentials', type='string', default='TMSSClient', help='django REST API credentials name, default: %default')

    group = OptionGroup(parser, 'Messaging options')
    group.add_option('-b', '--broker', dest='broker', type='string', default=DEFAULT_BROKER,
                     help='Address of the message broker, default: %default')
    group.add_option('-e', "--exchange", dest="exchange", type="string", default=DEFAULT_BUSNAME,
                     help="exchange where the TMSS event messages are published. [default: %default]")
    parser.add_option_group(group)

    (options, args) = parser.parse_args()

    TMSSsession.check_connection_and_exit_on_error(options.rest_credentials)

    # The TMSS slack app maintenance page (requires astron user creds): https://radio-observatory.slack.com/apps/A01SKUJHNKF-tmss

    # read the secrect slack webhook url from a lofar dbcredentials file.
    slack_url = DBCredentials().get(options.slack_credentials).host

    with create_service(slack_url=slack_url, rest_client_creds_id=options.rest_credentials, exchange=options.exchange, broker=options.broker):
        waitForInterrupt()

if __name__ == '__main__':
    main()
