#!/usr/bin/env python3

# Copyright (C) 2012-2015  ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

import logging
logger = logging.getLogger(__name__)

from optparse import OptionParser, OptionGroup
import json

from lofar.common.postgres import PostgresListener, makePostgresNotificationQueries
from lofar.messaging.messagebus import ToBus
from lofar.sas.tmss.client.tmssbuslistener import *
from lofar.common import dbcredentials
from lofar.common.util import single_line_with_single_spaces
from distutils.util import strtobool
from datetime import datetime, timedelta


class TMSSPGListener(PostgresListener):
    '''This class subscribes to the Subtask, TaskDraft/Blueprint & SchedulingUnitDraft/Blueprint tables in the TMSS database
    and send EventMessages upon each table row action, *Created, *Updated, *Deleted, and for each status update.
    See lofar.sas.tmss.client.tmssbuslistener.TMSSBusListener for the receiving BusListener'''

    def __init__(self,
                 dbcreds,
                 exchange=DEFAULT_BUSNAME,
                 broker=DEFAULT_BROKER):
        super().__init__(dbcreds=dbcreds)
        self.event_bus = ToBus(exchange=exchange, broker=broker)

        # two cache to keep track of the latest task/scheduling_unit (aggregated) statuses,
        # so we can lookup if the (aggregated) status of the task/scheduling_unit actually changes when a subtask's status changes.
        # This saves many (aggregated) status updates, where the (aggregated) status isn't changed.
        # contents of dict is a mapping of the task/su ID to a status,timestamp tuple
        self._task_status_cache = {}
        self._scheduling_unit_status_cache = {}


    def start(self):
        logger.info("Starting to listen for TMSS database changes and publishing EventMessages on %s  db: %s", self.event_bus.exchange, self._dbcreds.stringWithHiddenPassword())
        self.event_bus.open()

        # SubTask
        self.executeQuery(makePostgresNotificationQueries('', 'tmssapp_subtask', 'insert'))
        self.subscribe('tmssapp_subtask_insert', self.onSubTaskInserted)

        self.executeQuery(makePostgresNotificationQueries('', 'tmssapp_subtask', 'update'))
        self.subscribe('tmssapp_subtask_update', self.onSubTaskUpdated)

        self.executeQuery(makePostgresNotificationQueries('', 'tmssapp_subtask', 'delete'))
        self.subscribe('tmssapp_subtask_delete', self.onSubTaskDeleted)

        self.executeQuery(makePostgresNotificationQueries('', 'tmssapp_subtask', 'update', column_name='state_id', quote_column_value=True))
        self.subscribe('tmssapp_subtask_update_column_state_id', self.onSubTaskStateUpdated)


        # TaskBlueprint
        # please note that the status property in the TaskBlueprint model is derived from the subtasks
        # hence we cannot create a postgres notification for this "column".
        # But, see how onSubTaskStateUpdated does send a TaskBlueprint status changed event as well for its parent.
        self.executeQuery(makePostgresNotificationQueries('', 'tmssapp_taskblueprint', 'insert'))
        self.subscribe('tmssapp_taskblueprint_insert', self.onTaskBlueprintInserted)

        self.executeQuery(makePostgresNotificationQueries('', 'tmssapp_taskblueprint', 'update'))
        self.subscribe('tmssapp_taskblueprint_update', self.onTaskBlueprintUpdated)

        self.executeQuery(makePostgresNotificationQueries('', 'tmssapp_taskblueprint', 'delete'))
        self.subscribe('tmssapp_taskblueprint_delete', self.onTaskBlueprintDeleted)

        self.executeQuery(makePostgresNotificationQueries('', 'tmssapp_taskblueprint', 'update', column_name='output_pinned', quote_column_value=False))
        self.subscribe('tmssapp_taskblueprint_update_column_output_pinned', self.onTaskBlueprintOutputPinningUpdated)


        # TaskDraft
        self.executeQuery(makePostgresNotificationQueries('', 'tmssapp_taskdraft', 'insert'))
        self.subscribe('tmssapp_taskdraft_insert', self.onTaskDraftInserted)

        self.executeQuery(makePostgresNotificationQueries('', 'tmssapp_taskdraft', 'update'))
        self.subscribe('tmssapp_taskdraft_update', self.onTaskDraftUpdated)

        self.executeQuery(makePostgresNotificationQueries('', 'tmssapp_taskdraft', 'delete'))
        self.subscribe('tmssapp_taskdraft_delete', self.onTaskDraftDeleted)


        # SchedulingUnitBlueprint
        # please note that the status property in the SchedulingUnitBlueprint model is derived from the tasks, and these are derived from the subtasks
        # hence we cannot create a postgres notification for this "column".
        # But, see how onSubTaskStateUpdated does send a SchedulingUnitBlueprint status changed event as well for its parent.
        self.executeQuery(makePostgresNotificationQueries('', 'tmssapp_schedulingunitblueprint', 'insert'))
        self.subscribe('tmssapp_schedulingunitblueprint_insert', self.onSchedulingUnitBlueprintInserted)

        self.executeQuery(makePostgresNotificationQueries('', 'tmssapp_schedulingunitblueprint', 'update'))
        self.subscribe('tmssapp_schedulingunitblueprint_update', self.onSchedulingUnitBlueprintUpdated)

        self.executeQuery(makePostgresNotificationQueries('', 'tmssapp_schedulingunitblueprint', 'update', column_name='ingest_permission_granted_since', quote_column_value=True))
        self.subscribe('tmssapp_schedulingunitblueprint_update_column_ingest_permission_granted_since', self.onSchedulingUnitBlueprintIngestPermissionGranted)

        self.executeQuery(makePostgresNotificationQueries('', 'tmssapp_schedulingunitblueprint', 'delete'))
        self.subscribe('tmssapp_schedulingunitblueprint_delete', self.onSchedulingUnitBlueprintDeleted)


        # SchedulingUnitDraft
        self.executeQuery(makePostgresNotificationQueries('', 'tmssapp_schedulingunitdraft', 'insert'))
        self.subscribe('tmssapp_schedulingunitdraft_insert', self.onSchedulingUnitDraftInserted)

        self.executeQuery(makePostgresNotificationQueries('', 'tmssapp_schedulingunitdraft', 'update'))
        self.subscribe('tmssapp_schedulingunitdraft_update', self.onSchedulingUnitDraftUpdated)

        self.executeQuery(makePostgresNotificationQueries('', 'tmssapp_schedulingunitdraft', 'delete'))
        self.subscribe('tmssapp_schedulingunitdraft_delete', self.onSchedulingUnitDraftDeleted)

        self.executeQuery(makePostgresNotificationQueries('', 'tmssapp_schedulingunitdraft', 'update', column_name='scheduling_constraints_doc', quote_column_value=False))
        self.subscribe('tmssapp_schedulingunitdraft_update_column_scheduling_constraints_doc', self.onSchedulingUnitDraftConstraintsUpdated)

        # Settings
        self.executeQuery(makePostgresNotificationQueries('', 'tmssapp_setting', 'update', id_column_name='name_id', quote_id_value=True, column_name='value', quote_column_value=True))
        self.subscribe('tmssapp_setting_update_column_value', self.onSettingUpdated)

        # Project
        self.executeQuery(makePostgresNotificationQueries('', 'tmssapp_project', 'insert', id_column_name="name", quote_id_value=True))
        self.subscribe('tmssapp_project_insert', self.onProjectInserted)

        self.executeQuery(makePostgresNotificationQueries('', 'tmssapp_project', 'update', id_column_name="name", quote_id_value=True))
        self.subscribe('tmssapp_project_update', self.onProjectUpdated)

        self.executeQuery(makePostgresNotificationQueries('', 'tmssapp_project', 'delete', id_column_name="name", quote_id_value=True))
        self.subscribe('tmssapp_project_delete', self.onProjectDeleted)

        # ProjectQuotaArchiveLocation
        self.executeQuery(makePostgresNotificationQueries('', 'tmssapp_projectquotaarchivelocation', 'insert'))
        self.subscribe('tmssapp_projectquotaarchivelocation_insert', self.onProjectQuotaArchiveLocationInserted)

        self.executeQuery(makePostgresNotificationQueries('', 'tmssapp_projectquotaarchivelocation', 'update'))
        self.subscribe('tmssapp_projectquotaarchivelocation_update', self.onProjectQuotaArchiveLocationUpdated)

        self.executeQuery(makePostgresNotificationQueries('', 'tmssapp_projectquotaarchivelocation', 'delete'))
        self.subscribe('tmssapp_projectquotaarchivelocation_delete', self.onProjectQuotaArchiveLocationDeleted)


        return super().start()

    def stop(self):
        self.event_bus.close()
        logger.info("Stopped listening for TMSS database changes and publishing EventMessages on %s broker=%s db: %s",
                    self.event_bus.exchange, self.event_bus.broker, self._dbcreds.stringWithHiddenPassword())
        super().stop()

    def _sendNotification(self, subject, contentDict):
        try:
            if isinstance(contentDict, str):
                contentDict = json.loads(contentDict)

            msg = EventMessage(subject=subject, content=contentDict)
            logger.info('Sending %s to %s: %s',
                        subject, self.event_bus.exchange, single_line_with_single_spaces(contentDict))
            self.event_bus.send(msg)
        except Exception as e:
            logger.error(str(e))

    def onSubTaskInserted(self, payload = None):
        self._sendNotification(TMSS_SUBTASK_OBJECT_EVENT_PREFIX+'.Created', payload)

    def onSubTaskUpdated(self, payload = None):
        self._sendNotification(TMSS_SUBTASK_OBJECT_EVENT_PREFIX+'.Updated', payload)

    def onSubTaskDeleted(self, payload = None):
        self._sendNotification(TMSS_SUBTASK_OBJECT_EVENT_PREFIX+'.Deleted', payload)

    def onSubTaskStateUpdated(self, payload = None):
        payload_dict = json.loads(payload)
        # send notification for this subtask...
        from lofar.sas.tmss.tmss.tmssapp.models import Subtask
        subtask = Subtask.objects.get(id=payload_dict['id'])
        self._sendNotification(TMSS_SUBTASK_STATUS_EVENT_PREFIX+'.'+subtask.state.value.capitalize(),
                               {'id': subtask.id, 'status': subtask.state.value})

        # ... and also send status change and object update events for the parent task, and schedulingunit,
        # because their status is implicitly derived from their subtask(s)
        # send both object.updated and status change events

        # check if task status is new or changed... If so, send event.
        for task_blueprint in subtask.task_blueprints.all():
            task_id = task_blueprint.id
            task_status = task_blueprint.status
            if task_id not in self._task_status_cache or self._task_status_cache[task_id][1] != task_status:
                # update cache for this task
                self._task_status_cache[task_id] = (datetime.utcnow(), task_status)
    
                # send event(s)
                self.onTaskBlueprintUpdated( {'id': task_id})
                self._sendNotification(TMSS_TASKBLUEPRINT_STATUS_EVENT_PREFIX+'.'+task_status.capitalize(),
                                       {'id': task_id, 'status': task_status})
    
            # check if scheduling_unit status is new or changed... If so, send event.
            scheduling_unit_id = task_blueprint.scheduling_unit_blueprint.id
            scheduling_unit_status = task_blueprint.scheduling_unit_blueprint.status
            if scheduling_unit_id not in self._scheduling_unit_status_cache or self._scheduling_unit_status_cache[scheduling_unit_id][1] != scheduling_unit_status:
                # update cache for this task
                self._scheduling_unit_status_cache[scheduling_unit_id] = (datetime.utcnow(), scheduling_unit_status)
    
                # send event(s)
                self.onSchedulingUnitBlueprintUpdated( {'id': scheduling_unit_id})
                self._sendNotification(TMSS_SCHEDULINGUNITBLUEPRINT_STATUS_EVENT_PREFIX+'.'+scheduling_unit_status.capitalize(),
                                       {'id': scheduling_unit_id, 'status': scheduling_unit_status})

        try:
            # wipe old entries from cache.
            # This may result in some odd cases that an event is sent twice, even if the status did not change. That's a bit superfluous, but ok.
            for cache in [self._task_status_cache, self._scheduling_unit_status_cache]:
                for id in list(cache.keys()):
                    if datetime.utcnow() - cache[id][0] > timedelta(days=1):
                        del cache[id]
        except Exception as e:
            logger.warning(str(e))


    def onTaskBlueprintInserted(self, payload = None):
        self._sendNotification(TMSS_TASKBLUEPRINT_OBJECT_EVENT_PREFIX+'.Created', payload)

    def onTaskBlueprintUpdated(self, payload = None):
        self._sendNotification (TMSS_TASKBLUEPRINT_OBJECT_EVENT_PREFIX+'.Updated', payload)

    def onTaskBlueprintDeleted(self, payload = None):
        self._sendNotification(TMSS_TASKBLUEPRINT_OBJECT_EVENT_PREFIX+'.Deleted', payload)

    def onTaskBlueprintOutputPinningUpdated(self, payload = None):
        self._sendNotification(TMSS_TASKBLUEPRINT_OBJECT_EVENT_PREFIX+'.OutputPinningUpdated', payload)

        if isinstance(payload, str):
            payload = json.loads(payload)

        from lofar.sas.tmss.tmss.tmssapp.models import TaskBlueprint
        task_blueprint = TaskBlueprint.objects.get(id=payload['id'])
        self._sendNotification(TMSS_SCHEDULINGUNITBLUEPRINT_OBJECT_EVENT_PREFIX+'.Updated', {'id': task_blueprint.scheduling_unit_blueprint.id})

    def onTaskDraftInserted(self, payload = None):
        self._sendNotification(TMSS_TASKDRAFT_OBJECT_EVENT_PREFIX+'.Created', payload)

    def onTaskDraftUpdated(self, payload = None):
        self._sendNotification(TMSS_TASKDRAFT_OBJECT_EVENT_PREFIX+'.Updated', payload)

    def onTaskDraftDeleted(self, payload = None):
        self._sendNotification(TMSS_TASKDRAFT_OBJECT_EVENT_PREFIX+'.Deleted', payload)

    def onSchedulingUnitBlueprintInserted(self, payload = None):
        self._sendNotification(TMSS_SCHEDULINGUNITBLUEPRINT_OBJECT_EVENT_PREFIX+'.Created', payload)

    def onSchedulingUnitBlueprintUpdated(self, payload = None):
        self._sendNotification(TMSS_SCHEDULINGUNITBLUEPRINT_OBJECT_EVENT_PREFIX+'.Updated', payload)

        if isinstance(payload, str):
            payload = json.loads(payload)

        from lofar.sas.tmss.tmss.tmssapp.models import SchedulingUnitBlueprint
        scheduling_unit_blueprint = SchedulingUnitBlueprint.objects.get(id=payload['id'])

        if not scheduling_unit_blueprint.can_proceed:
            self.onSchedulingUnitBlueprintCannotProceed( {'id': scheduling_unit_blueprint.id})

    def onSchedulingUnitBlueprintIngestPermissionGranted(self, payload=None):
        self._sendNotification(TMSS_SCHEDULINGUNITBLUEPRINT_OBJECT_EVENT_PREFIX + '.IngestPermissionGranted', payload)

    def onSchedulingUnitBlueprintDeleted(self, payload = None):
        self._sendNotification(TMSS_SCHEDULINGUNITBLUEPRINT_OBJECT_EVENT_PREFIX+'.Deleted', payload)

    def onSchedulingUnitDraftInserted(self, payload = None):
        self._sendNotification(TMSS_SCHEDULINGUNITDRAFT_OBJECT_EVENT_PREFIX+'.Created', payload)

    def onSchedulingUnitDraftUpdated(self, payload = None):
        self._sendNotification(TMSS_SCHEDULINGUNITDRAFT_OBJECT_EVENT_PREFIX+'.Updated', payload)

    def onSchedulingUnitDraftDeleted(self, payload = None):
        self._sendNotification(TMSS_SCHEDULINGUNITDRAFT_OBJECT_EVENT_PREFIX+'.Deleted', payload)

    def onSchedulingUnitDraftConstraintsUpdated(self, payload = None):
        # convert payload string to nested json doc
        self._sendNotification(TMSS_SCHEDULINGUNITDRAFT_OBJECT_EVENT_PREFIX+'.Constraints.Updated', payload)

    def onProjectInserted(self, payload = None):
        self._sendNotification(TMSS_PROJECT_OBJECT_EVENT_PREFIX+'.Created', payload)

    def onProjectUpdated(self, payload = None):
        self._sendNotification(TMSS_PROJECT_OBJECT_EVENT_PREFIX+'.Updated', payload)

    def onProjectDeleted(self, payload = None):
        self._sendNotification(TMSS_PROJECT_OBJECT_EVENT_PREFIX+'.Deleted', payload)

    def onProjectQuotaArchiveLocationInserted(self, payload = None):
        self._sendNotification(TMSS_PROJECTQUOTAARCHIVELOCATION_OBJECT_EVENT_PREFIX+'.Created', payload)

    def onProjectQuotaArchiveLocationUpdated(self, payload = None):
        self._sendNotification(TMSS_PROJECTQUOTAARCHIVELOCATION_OBJECT_EVENT_PREFIX+'.Updated', payload)

    def onProjectQuotaArchiveLocationDeleted(self, payload = None):
        self._sendNotification(TMSS_PROJECTQUOTAARCHIVELOCATION_OBJECT_EVENT_PREFIX+'.Deleted', payload)

    def onSettingUpdated(self, payload = None):
        payload = json.loads(payload)
        payload['name'] = payload['name_id']
        del payload['name_id']
        payload['value'] = strtobool(payload['value'])
        self._sendNotification(TMSS_SETTING_OBJECT_EVENT_PREFIX+'.Updated', payload)

    def onSchedulingUnitBlueprintCannotProceed(self, payload = None):
        self._sendNotification(TMSS_SCHEDULINGUNITBLUEPRINT_OBJECT_EVENT_PREFIX+'.CannotProceed', payload)


def create_service(dbcreds, exchange=DEFAULT_BUSNAME, broker=DEFAULT_BROKER):
    '''create a TMSSPGListener instance'''
    return TMSSPGListener(dbcreds=dbcreds, exchange=exchange, broker=broker)


def main():
    # make sure we run in UTC timezone
    import os
    os.environ['TZ'] = 'UTC'

    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

    # Check the invocation arguments
    parser = OptionParser("%prog [options]",
                          description='runs the radb postgres listener which listens to changes on some tables in the radb and publishes the changes as notifications on the bus.')

    group = OptionGroup(parser, 'Messaging options')
    group.add_option('-b', '--broker', dest='broker', type='string', default=DEFAULT_BROKER,
                     help='Address of the message broker, default: %default')
    group.add_option('-e', "--exchange", dest="exchange", type="string", default=DEFAULT_BUSNAME,
                     help="Bus or queue where the TMSS messages are published. [default: %default]")
    parser.add_option_group(group)

    group = OptionGroup(parser, 'Django options')
    parser.add_option_group(group)
    group.add_option('-C', '--credentials', dest='dbcredentials', type='string', default=os.environ.get('TMSS_DBCREDENTIALS', 'TMSS'), help='django dbcredentials name, default: %default')

    (options, args) = parser.parse_args()

    from lofar.sas.tmss.tmss import setup_and_check_tmss_django_database_connection_and_exit_on_error
    setup_and_check_tmss_django_database_connection_and_exit_on_error(options.dbcredentials)

    dbcreds = dbcredentials.DBCredentials().get(options.dbcredentials)

    with create_service(dbcreds=dbcreds,
                        exchange=options.exchange,
                        broker=options.broker) as listener:
        listener.waitWhileListening()

if __name__ == '__main__':
    main()
