#!/usr/bin/env python3

# Copyright (C) 2012-2015  ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

import unittest
import uuid

import logging
logger = logging.getLogger(__name__)
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

from lofar.sas.tmss.test.test_environment import TMSSTestEnvironment

from lofar.messaging.messagebus import TemporaryExchange
from lofar.common.test_utils import integration_test
from lofar.common.util import single_line_with_single_spaces
from threading import Lock
import requests
import json
from collections import deque
from datetime import datetime, timedelta

@integration_test
class TestSubtaskSchedulingService(unittest.TestCase):
    '''
    Tests for the SubtaskSchedulingService
    '''
    @classmethod
    def setUpClass(cls) -> None:
        cls.TEST_UUID = uuid.uuid1()

        cls.tmp_exchange = TemporaryExchange("%s_%s" % (cls.__name__, cls.TEST_UUID))
        cls.tmp_exchange.open()

        cls.tmss_test_env = TMSSTestEnvironment(exchange=cls.tmp_exchange.address, start_postgres_listener=False, populate_schemas=False, populate_test_data=False)
        cls.tmss_test_env.start()

        from lofar.sas.tmss.test.tmss_test_data_rest import TMSSRESTTestDataCreator
        cls.test_data_creator = TMSSRESTTestDataCreator(cls.tmss_test_env.django_server.url,
                                                        (cls.tmss_test_env.ldap_server.dbcreds.user,
                                                         cls.tmss_test_env.ldap_server.dbcreds.password))

    @classmethod
    def tearDownClass(cls) -> None:
        cls.tmss_test_env.stop()
        cls.tmp_exchange.close()

    def test_01_for_expected_behaviour(self):
        '''
        This test starts a TMSSPGListener service and TMSS, creates/updates/deletes subtasks/tasks/schedulingunits, and checks if the correct events are sent.
        '''
        logger.info(' -- test_01_for_expected_behaviour -- ')
        from lofar.sas.tmss.services.tmss_postgres_listener import TMSSPGListener, TMSS_SUBTASK_OBJECT_EVENT_PREFIX, \
                TMSS_SUBTASK_STATUS_EVENT_PREFIX, TMSS_TASKBLUEPRINT_OBJECT_EVENT_PREFIX,  TMSS_TASKBLUEPRINT_STATUS_EVENT_PREFIX, \
                TMSS_TASKDRAFT_OBJECT_EVENT_PREFIX, TMSS_SCHEDULINGUNITBLUEPRINT_OBJECT_EVENT_PREFIX, \
                TMSS_SCHEDULINGUNITBLUEPRINT_STATUS_EVENT_PREFIX, TMSS_SCHEDULINGUNITDRAFT_OBJECT_EVENT_PREFIX, TMSS_PROJECT_OBJECT_EVENT_PREFIX

        class TestTMSSPGListener(TMSSPGListener):
            '''Helper TMSSPGListener for this test, storing intermediate results, and providing synchronization threading.Events'''
            def __init__(self, dbcreds, exchange=self.tmp_exchange.address):
                super().__init__(dbcreds, exchange)
                self.subjects = deque()
                self.contentDicts = deque()
                self.lock = Lock()

            def _sendNotification(self, subject, contentDict):
                # instead of sending a notification to the messagebus, record the subject and content in queues
                # so we can check in the test if the correct subjects are recorded
                with self.lock:
                    logger.info("detected db change: %s %s", subject, single_line_with_single_spaces(contentDict))
                    self.subjects.append(subject)
                    self.contentDicts.append(json.loads(contentDict) if isinstance(contentDict, str) else contentDict)

        # create and start the service (the object under test)
        with TestTMSSPGListener(exchange=self.tmp_exchange.address, dbcreds=self.tmss_test_env.database.dbcreds) as service:
            # create a Project
            project = self.test_data_creator.post_data_and_get_response_as_json_object(self.test_data_creator.Project(), '/project/')

            # sync and check
            with service.lock:
                self.assertEqual(TMSS_PROJECT_OBJECT_EVENT_PREFIX +'.Created', service.subjects.popleft())
                self.assertEqual({"name": project['name']}, service.contentDicts.popleft())

            # create a SchedulingUnitDraft
            scheduling_set_url = self.test_data_creator.post_data_and_get_url(self.test_data_creator.SchedulingSet(project_url=project['url']), '/scheduling_set/')
            su_draft = self.test_data_creator.post_data_and_get_response_as_json_object(self.test_data_creator.SchedulingUnitDraft(scheduling_set_url=scheduling_set_url), '/scheduling_unit_draft/')

            # sync and check
            with service.lock:
                self.assertEqual(TMSS_SCHEDULINGUNITDRAFT_OBJECT_EVENT_PREFIX+'.Created', service.subjects.popleft())
                self.assertEqual({"id": su_draft['id']}, service.contentDicts.popleft())


            # create a TaskDraft
            task_draft = self.test_data_creator.post_data_and_get_response_as_json_object(self.test_data_creator.TaskDraft(scheduling_unit_draft_url=su_draft['url']), '/task_draft/')

            # sync and check
            with service.lock:
                self.assertEqual(TMSS_TASKDRAFT_OBJECT_EVENT_PREFIX+'.Created', service.subjects.popleft())
                self.assertEqual({"id": task_draft['id']}, service.contentDicts.popleft())


            # create a SchedulingUnitBlueprint
            su_blueprint = self.test_data_creator.post_data_and_get_response_as_json_object(self.test_data_creator.SchedulingUnitBlueprint(scheduling_unit_draft_url=su_draft['url']), '/scheduling_unit_blueprint/')

            # sync and check
            with service.lock:
                self.assertEqual(TMSS_SCHEDULINGUNITBLUEPRINT_OBJECT_EVENT_PREFIX+'.Created', service.subjects.popleft())
                self.assertEqual({"id": su_blueprint['id']}, service.contentDicts.popleft())


            # create a TaskBlueprint
            task_blueprint = self.test_data_creator.post_data_and_get_response_as_json_object(self.test_data_creator.TaskBlueprint(scheduling_unit_blueprint_url=su_blueprint['url'],
                                                                                                                                   draft_url=task_draft['url']), '/task_blueprint/')

            # sync and check
            with service.lock:
                self.assertEqual(TMSS_TASKBLUEPRINT_OBJECT_EVENT_PREFIX+'.Created', service.subjects.popleft())
                self.assertEqual({"id": task_blueprint['id']}, service.contentDicts.popleft())


            # create a SubTask
            subtask = self.test_data_creator.post_data_and_get_response_as_json_object(self.test_data_creator.Subtask(task_blueprint_urls=[task_blueprint['url']]), '/subtask/')

            # sync and check
            with service.lock:
                self.assertEqual(TMSS_SUBTASK_OBJECT_EVENT_PREFIX+'.Created', service.subjects.popleft())
                self.assertEqual({"id": subtask['id']}, service.contentDicts.popleft())

            # update subtask status, use a nice tmss_client and the rest api.
            with self.tmss_test_env.create_tmss_client() as client:
                client.set_subtask_status(subtask['id'], 'defined')

                # ugly, but functional. Wait for all status updates: 1 object, 1 status. both per each object (3 types) => total 6 events.
                start_wait = datetime.utcnow()
                while True:
                    with service.lock:
                        if len(service.subjects) == 6:
                            break
                    if  datetime.utcnow() - start_wait > timedelta(seconds=5):
                        raise TimeoutError("timeout while waiting for status/object updates")

            # sync and check
            with service.lock:
                self.assertEqual(TMSS_SUBTASK_OBJECT_EVENT_PREFIX + '.Updated', service.subjects.popleft())
                self.assertEqual({'id': subtask['id']}, service.contentDicts.popleft())

                self.assertEqual(TMSS_SUBTASK_STATUS_EVENT_PREFIX+'.Defined', service.subjects.popleft())
                self.assertEqual({'id': subtask['id'], 'status': 'defined'}, service.contentDicts.popleft())

                self.assertEqual(TMSS_TASKBLUEPRINT_OBJECT_EVENT_PREFIX+'.Updated', service.subjects.popleft())
                self.assertEqual({'id': task_blueprint['id']}, service.contentDicts.popleft())

                self.assertEqual(TMSS_TASKBLUEPRINT_STATUS_EVENT_PREFIX+'.Schedulable', service.subjects.popleft())
                self.assertEqual({'id': task_blueprint['id'], 'status': 'schedulable'}, service.contentDicts.popleft())

                self.assertEqual(TMSS_SCHEDULINGUNITBLUEPRINT_OBJECT_EVENT_PREFIX+'.Updated', service.subjects.popleft())
                self.assertEqual({'id': su_blueprint['id']}, service.contentDicts.popleft())

                self.assertEqual(TMSS_SCHEDULINGUNITBLUEPRINT_STATUS_EVENT_PREFIX+'.Schedulable', service.subjects.popleft())
                self.assertEqual({'id': su_blueprint['id'], 'status': 'schedulable'}, service.contentDicts.popleft())

            # delete subtask, use direct http delete request on rest api
            requests.delete(subtask['url'], auth=self.test_data_creator.auth)

            # sync and check subtask deleted
            with service.lock:
                self.assertEqual(TMSS_SUBTASK_OBJECT_EVENT_PREFIX+'.Deleted', service.subjects.popleft())
                self.assertEqual({'id': subtask['id']}, service.contentDicts.popleft())


if __name__ == '__main__':
    #run the unit tests
    unittest.main()
