"""
This file contains the database models
"""

import logging
logger = logging.getLogger(__name__)

from django.db.models import Model, CharField, DateTimeField, BooleanField, ForeignKey, CASCADE, IntegerField, FloatField, SET_NULL, PROTECT, ManyToManyField, UniqueConstraint, QuerySet, OneToOneField
from django.contrib.postgres.fields import JSONField
from enum import Enum
from django.db.models.expressions import RawSQL
from django.db.models.deletion import ProtectedError
from .common import AbstractChoice, BasicCommon, Template, NamedCommon, TemplateSchemaMixin, NamedCommonPK, RefreshFromDbInvalidatesCachedPropertiesMixin, ProjectPropertyMixin
from lofar.common.json_utils import validate_json_against_schema, validate_json_against_its_schema, add_defaults_to_json_object_for_schema
from lofar.sas.tmss.tmss.exceptions import *
from django.core.exceptions import ValidationError
import datetime
from collections import Counter
from django.utils.functional import cached_property
from lofar.sas.tmss.tmss.exceptions import BlueprintCreationException, TMSSException

#
# I/O
#

# choices

class Role(AbstractChoice):
    """Defines the model and predefined list of possible Role's for TaskConnectorType.
    The items in the Choises class below are automagically populated into the database via a data migration.
    When changing/adding/removing items to the Choices, please update the common json schema for tasks as well."""
    class Choices(Enum):
        CORRELATOR = "correlator"
        BEAMFORMER = "beamformer"
        INSPECTION_PLOTS = "inspection plots"
        CALIBRATOR = "calibrator"
        TARGET = "target"
        ANY = "any"


class IOType(AbstractChoice):
    """Defines the model and predefined list of possible IOType's for TaskConnectorType.
    The items in the Choises class below are automagically populated into the database via a data migration."""
    class Choices(Enum):
        INPUT = "input"
        OUTPUT = "output"
        # maybe we can add an IN_PLACE="in_place" option in the future, but for now it's not needed.


class Datatype(AbstractChoice):
    """Defines the model and predefined list of possible Datatype's for TaskConnectorType.
    The items in the Choises class below are automagically populated into the database via a data migration.
    When changing/adding/removing items to the Choices, please update the common json schema for tasks as well."""
    class Choices(Enum):
        VISIBILITIES = "visibilities"
        TIME_SERIES = "time series"
        INSTRUMENT_MODEL = "instrument model"
        IMAGE = "image"
        QUALITY = "quality"
        PULSAR_PROFILE = "pulsar profile"


class Dataformat(AbstractChoice):
    """Defines the model and predefined list of possible Dataformat's for TaskConnectorType.
    The items in the Choises class below are automagically populated into the database via a data migration.
    When changing/adding/removing items to the Choices, please update the common json schema for tasks as well."""
    class Choices(Enum):
        MEASUREMENTSET = "MeasurementSet"
        BEAMFORMED = "Beamformed"
        QA_HDF5 = "QA_HDF5"
        QA_PLOTS = "QA_Plots"
        PULP_SUMMARY = "pulp summary"
        PULP_ANALYSIS = "pulp analysis"


class CopyReason(AbstractChoice):
    """Defines the model and predefined list of possible CopyReason's for SchedulingUnitDraft and TaskDraft.
    The items in the Choises class below are automagically populated into the database via a data migration."""
    class Choices(Enum):
        TEMPLATE = "template"
        REPEATED = "repeated"


class SchedulingRelationPlacement(AbstractChoice):
    """Defines the model and predefined list of possible Placements for Task Scheduling Relation.
        The items in the Choices class below are automagically populated into the database via a data migration."""
    class Choices(Enum):
        AFTER = "after"
        BEFORE = "before"
        PARALLEL = "parallel"

class SystemSettingFlag(AbstractChoice):
    """Defines the model and predefined list of possible Flags to be used in Setting.
    The items in the Choises class below are automagically populated into the database via a data migration."""
    class Choices(Enum):
        DYNAMIC_SCHEDULING_ENABLED = "dynamic_scheduling_enabled"



class Quantity(AbstractChoice):
    """Defines the model and predefined list of possible period categories to be used in a.o. ProjectQuota and CycleQuota.
       Please note that, by agreement, all values of a certain quantity are stored in SI units.
       For example: a duration has the quantity "time" which is stored in "seconds".
       We only store the quatity-type in the database, because we known the correct SI units for each quantity, so there is no need to store the unit as well.
        The items in the Choices class below are automagically populated into the database via a data migration."""

    class Choices(Enum):
        # these are the basic quantities that we currently use. More can be added if needed.
        TIME = "time"
        BYTES = "bytes"
        NUMBER = "number"


class PeriodCategory(AbstractChoice):
    """Defines the model and predefined list of possible period categories to be used in Project as a policy for managing the project's lifetime.
        The items in the Choices class below are automagically populated into the database via a data migration."""

    class Choices(Enum):
        SINGLE_CYCLE = "single_cycle"
        LONG_TERM = "long_term"
        UNBOUNDED = "unbounded"


class ProjectCategory(AbstractChoice):
    """Defines the model and predefined list of possible project categories to be used in Project.
        The items in the Choices class below are automagically populated into the database via a data migration."""

    class Choices(Enum):
        REGULAR = "regular"
        USER_SHARED_SUPPORT = "user_shared_support"
        COMMISSIONING = "commissioning"
        DDT = "ddt"
        TEST = "test"


class TaskType(AbstractChoice):
    """Defines the model and predefined list of possible TaskType's for Task.
    The items in the Choices class below are automagically populated into the database via a data migration."""
    class Choices(Enum):
        OBSERVATION = "observation"
        PIPELINE = "pipeline"
        INGEST = "ingest"
        CLEANUP = 'cleanup'
        MAINTENANCE = "maintenance"
        OTHER = 'other'


class PriorityQueueType(AbstractChoice):
    """Defines the possible priority queues for SchedulingUnits.
    The items in the Choices class below are automagically populated into the database via a data migration."""
    class Choices(Enum):
        A = "A"
        B = "B"

# concrete models

class Setting(BasicCommon):
    name = OneToOneField('SystemSettingFlag', null=False, on_delete=PROTECT, primary_key=True)
    value = BooleanField(null=False)


class TaskConnectorType(BasicCommon):
    ''' Describes the data type & format combinations a Task can accept or produce. The "role" is used to distinguish
        inputs (or outputs) that have the same data type & format, but are used in different ways by the task. For
        example, a calibration pipeline accepts measurement sets only, but distinghuishes between CALIBRATOR and
        TARGET roles.'''
    role = ForeignKey('Role', null=False, on_delete=PROTECT)
    datatype = ForeignKey('Datatype', null=False, on_delete=PROTECT)
    dataformat = ForeignKey('Dataformat', null=False, on_delete=PROTECT)
    task_template = ForeignKey("TaskTemplate", related_name='output_connector_types', null=False, on_delete=CASCADE)
    iotype = ForeignKey('IOType', null=False, on_delete=PROTECT, help_text="Is this connector an input or output")


#
# Templates
#

class CommonSchemaTemplate(Template):
    '''A Template model for common (reusable) schema's'''
    pass


class GeneratorTemplate(Template):
    create_function = CharField(max_length=128, help_text='Python function to call to execute the generator.')


class DefaultGeneratorTemplate(BasicCommon):
    name = CharField(max_length=128, unique=True)
    template = ForeignKey("GeneratorTemplate", on_delete=PROTECT)


class SchedulingUnitObservingStrategyTemplate(NamedCommon):
    '''
    A SchedulingUnitObservingStrategyTemplate is a template in the sense that it serves as a template to fill in json data objects conform its referred scheduling_unit_template.
    It is however not derived from the (abstract) Template super-class, because the Template super class is for JSON schemas, not JSON data objects.
    '''
    version = CharField(max_length=128, help_text='Version of this template (with respect to other templates of the same name).')
    template = JSONField(null=False, help_text='JSON-data compliant with the JSON-schema in the scheduling_unit_template. '
                                               'This observation strategy template like a predefined recipe with all the correct settings, and defines which parameters the user can alter.')
    scheduling_unit_template = ForeignKey("SchedulingUnitTemplate", on_delete=PROTECT, null=False, help_text="")

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        if self.template and self.scheduling_unit_template_id and self.scheduling_unit_template.schema:
            validate_json_against_schema(self.template, self.scheduling_unit_template.schema)

        super().save(force_insert, force_update, using, update_fields)


class SchedulingUnitTemplate(Template):
    pass


class DefaultSchedulingUnitTemplate(BasicCommon):
    name = CharField(max_length=128, unique=True)
    template = ForeignKey("SchedulingUnitTemplate", on_delete=PROTECT)


class SchedulingConstraintsTemplate(Template):
    pass


class DefaultSchedulingConstraintsTemplate(BasicCommon):
    name = CharField(max_length=128, unique=True)
    template = ForeignKey("SchedulingConstraintsTemplate", on_delete=PROTECT)


class TaskTemplate(Template):
    validation_code_js = CharField(max_length=128, blank=True, default="", help_text='JavaScript code for additional (complex) validation.')
    type = ForeignKey('TaskType', null=False, on_delete=PROTECT)


class DefaultTaskTemplate(BasicCommon):
    name = CharField(max_length=128, unique=True)
    template = ForeignKey("TaskTemplate", on_delete=PROTECT)


class TaskRelationSelectionTemplate(Template):
    pass


class DefaultTaskRelationSelectionTemplate(BasicCommon):
    name = CharField(max_length=128, unique=True)
    template = ForeignKey("TaskRelationSelectionTemplate", on_delete=PROTECT)


class ReservationStrategyTemplate(NamedCommon):
    '''
    A ReservationStrategyTemplate is a template in the sense that it serves as a template to fill in json data objects
    conform its referred reservation_template.
    It is however not derived from the (abstract) Template super-class, because the Template super class is for
    JSON schemas, not JSON data objects.
    '''
    version = CharField(max_length=128, help_text='Version of this template (with respect to other templates of the same name).')
    template = JSONField(null=False, help_text='JSON-data compliant with the JSON-schema in the reservation_template. '
                                               'This reservation strategy template like a predefined recipe with all '
                                               'the correct settings, and defines which parameters the user can alter.')
    reservation_template = ForeignKey("ReservationTemplate", on_delete=PROTECT, null=False, help_text="")

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        if self.template and self.reservation_template_id and self.reservation_template.schema:
            validate_json_against_schema(self.template, self.reservation_template.schema)

        super().save(force_insert, force_update, using, update_fields)


class ReservationTemplate(Template):
    pass


class DefaultReservationTemplate(BasicCommon):
    name = CharField(max_length=128, unique=True)
    template = ForeignKey("ReservationTemplate", on_delete=PROTECT)


#
# Instance Objects
#

class Cycle(RefreshFromDbInvalidatesCachedPropertiesMixin, NamedCommonPK):
    start = DateTimeField(help_text='Moment at which the cycle starts, that is, when its projects can run.')
    stop = DateTimeField(help_text='Moment at which the cycle officially ends.')

    @cached_property
    def duration(self) -> datetime.timedelta:
        '''the duration of the cycle (stop-start date)'''
        return self.stop - self.start

    # Project to Cycle references are ManyToMany now, so cannot be protected via on_delete on db level.
    # Instead, explicitly check whether there are projects linked to the Cycle and prevent delete
    # todo: verify correct behavior
    #  - Implemented: Cycles cannot be deleted if they have projects assigned to them
    #  - Also possible: Projects need at least one cycle
    def delete(self, *args, **kwargs):
        if len(self.projects.all()) > 0:
            raise ProtectedError(protected_objects=self.projects.all(), msg='This Cycle is referenced by %s project(s) and cannot be deleted.' % len(self.projects.all()))
        else:
            super().delete(*args, **kwargs)


class CycleQuota(Model):
    cycle = ForeignKey('Cycle', related_name="quota", on_delete=PROTECT, help_text='Cycle to which these quota apply.')
    value = FloatField(help_text='Resource Quota value')
    resource_type = ForeignKey('ResourceType', on_delete=PROTECT, help_text='Resource type.')


class Project(RefreshFromDbInvalidatesCachedPropertiesMixin, NamedCommonPK):
    # todo: cycles should be protected since we have to manually decide to clean up projects with a cycle or keep them without cycle, however, ManyToManyField does not allow for that
    cycles = ManyToManyField('Cycle', related_name='projects', blank=True, help_text='Cycles to which this project belongs (NULLable).')
    priority_rank = FloatField(null=False, help_text='Priority of this project w.r.t. other projects. Projects can interrupt observations of lower-priority projects.') # todo: add if needed: validators=[MinValueValidator(0.0), MaxValueValidator(1.0)]
    trigger_priority = IntegerField(default=1000, help_text='Priority of this project w.r.t. triggers.') # todo: verify meaning and add to help_text: "Triggers with higher priority than this threshold can interrupt observations of projects."
    auto_ingest = BooleanField(default=False, help_text='True if The data is ingested when the other scheduling unit tasks are finished. False if The data is ingested after approval in the QA validation workflow. At the end of this a flag is set that the data can be ingested.')
    can_trigger = BooleanField(default=False, help_text='True if this project is allowed to supply observation requests on the fly, possibly interrupting currently running observations (responsive telescope).')
    private_data = BooleanField(default=True, help_text='True if data of this project is sensitive. Sensitive data is not made public.')
    expert = BooleanField(default=False, help_text='Expert projects put more responsibility on the PI.')
    filler = BooleanField(default=False, help_text='Use this project to fill up idle telescope time.')
    project_category = ForeignKey('ProjectCategory', help_text='Category this project falls under.', null=True, on_delete=PROTECT)
    period_category = ForeignKey('PeriodCategory', help_text='Policy for managing the lifetime of this project.', null=True, on_delete=PROTECT)
    auto_pin = BooleanField(default=False, help_text='True if the output_pinned flag of tasks in this project should be set True on creation.')
    piggyback_allowed_tbb = BooleanField(default=True, help_text='Piggyback key for TBB.')
    piggyback_allowed_aartfaac = BooleanField(default=True, help_text='Piggyback key for AARTFAAC.')
    path_to_project = "project"

    @cached_property
    def duration(self) -> datetime.timedelta:
        '''return the overall duration of all tasks of this scheduling unit
        '''
        return self.relative_stop_time - self.relative_start_time

    @cached_property
    def project(self):
        return self


class ProjectQuota(Model):
    project = ForeignKey('Project', related_name="quota", on_delete=PROTECT, help_text='Project to wich this quota belongs.')  # protected to avoid accidents
    value = FloatField(help_text='Resource Quota value')
    resource_type = ForeignKey('ResourceType', on_delete=PROTECT, help_text='Resource type.')  # protected to avoid accidents


class ProjectQuotaArchiveLocation(RefreshFromDbInvalidatesCachedPropertiesMixin, Model):
    project_quota = ForeignKey('ProjectQuota', null=False, related_name="project_quota_archive_location", on_delete=PROTECT, help_text='The ProjectQuota for this archive location')
    archive_location = ForeignKey('Filesystem', null=False, on_delete=PROTECT, help_text='Location of an archive LTA cluster.')

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        if self.project_quota.resource_type.quantity.value != Quantity.Choices.BYTES.value:
            raise ValueError("A ProjectQuotaArchiveLocation should have its project_quota defined in a 'bytes' quantity")
        if not self.archive_location.cluster.archive_site:
            raise ValueError("The archive_location of a ProjectQuotaArchiveLocation should be an archive site")
        super().save(force_insert, force_update, using, update_fields)

    @cached_property
    def archive_subdirectory(self) -> str:
        '''return name of the subdirectory where this project is stored. By default and convention this the the lower_case project name'''
        return self.project_quota.project.name.lower()

    @cached_property
    def full_archive_uri(self) -> str:
        '''return full uri where this project is stored. By default and convention this the the lower_case project name'''
        return "%s/%s/" % (self.archive_location.directory.rstrip('/'), self.archive_subdirectory)


class ResourceType(NamedCommonPK):
    quantity = ForeignKey('Quantity', null=False, on_delete=PROTECT, help_text='The quantity of this resource type.')


class SchedulingSet(NamedCommon, TemplateSchemaMixin):
    generator_doc = JSONField(null=True, help_text='Parameters for the generator (NULLable).')
    generator_template = ForeignKey('GeneratorTemplate', on_delete=SET_NULL, null=True, help_text='Generator for the scheduling units in this set (NULLable).')
    generator_source = ForeignKey('SchedulingUnitDraft', on_delete=SET_NULL, null=True, help_text='Reference for the generator to an existing collection of specifications (NULLable).')
    project = ForeignKey('Project', related_name="scheduling_sets", on_delete=PROTECT, help_text='Project to which this scheduling set belongs.')  # protected to avoid accidents

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        self.annotate_validate_add_defaults_to_doc_using_template('generator_doc', 'generator_template')
        super().save(force_insert, force_update, using, update_fields)


class SchedulingUnitDraft(NamedCommon, TemplateSchemaMixin):
    requirements_doc = JSONField(help_text='Scheduling and/or quality requirements for this run.')
    copies = ForeignKey('SchedulingUnitDraft', related_name="copied_from", on_delete=SET_NULL, null=True, help_text='Source reference, if we are a copy (NULLable).')
    copy_reason = ForeignKey('CopyReason', null=True, on_delete=PROTECT, help_text='Reason why source was copied (NULLable).')
    generator_instance_doc = JSONField(null=True, help_text='Parameter value that generated this run draft (NULLable).')
    scheduling_set = ForeignKey('SchedulingSet', related_name='scheduling_unit_drafts', on_delete=CASCADE, help_text='Set to which this scheduling unit draft belongs.')
    requirements_template = ForeignKey('SchedulingUnitTemplate', on_delete=CASCADE, help_text='Schema used for requirements_doc.')
    observation_strategy_template = ForeignKey('SchedulingUnitObservingStrategyTemplate', on_delete=PROTECT, null=True, help_text='Observation Strategy Template used to create the requirements_doc.')
    scheduling_constraints_doc = JSONField(help_text='Scheduling Constraints for this run.', null=True)
    scheduling_constraints_template = ForeignKey('SchedulingConstraintsTemplate', on_delete=CASCADE, null=True, help_text='Schema used for scheduling_constraints_doc.')
    ingest_permission_required = BooleanField(default=False, help_text='Explicit permission is needed before the task.')
    piggyback_allowed_tbb = BooleanField(help_text='Piggyback key for TBB.', null=True)
    piggyback_allowed_aartfaac = BooleanField(help_text='Piggyback key for AARTFAAC.', null=True)
    priority_rank = FloatField(null=False, default=0.0, help_text='Priority of this scheduling unit w.r.t. other scheduling units within the same queue and project.')
    priority_queue = ForeignKey('PriorityQueueType', null=False, on_delete=PROTECT, default="A", help_text='Priority queue of this scheduling unit. Queues provide a strict ordering between scheduling units.')
    is_triggered = BooleanField(default=False, null=False, help_text='boolean (default FALSE), which indicates whether this observation was triggered (responsive telescope)')

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        if self.requirements_doc is not None and self.requirements_template_id and self.requirements_template.schema is not None:
            # If this scheduling unit was created from an observation_strategy_template,
            # then make sure that the observation_strategy_template validates against this unit's requirements_template.schema
            if self.observation_strategy_template_id and self.observation_strategy_template.template:
                validate_json_against_schema(self.observation_strategy_template.template, self.requirements_template.schema)

        if self.scheduling_constraints_doc is not None and self.scheduling_constraints_template_id and self.scheduling_constraints_template.schema is not None:
            validate_json_against_schema(self.scheduling_constraints_doc, self.scheduling_constraints_template.schema)

        # This code only happens if the objects is not in the database yet. self._state.adding is True creating
        if self._state.adding and hasattr(self, 'scheduling_set') and self.scheduling_set.project.auto_ingest is False:
            # When project.auto_ingest=False, the scheduling units will be created with ingest_permission_required = True
            self.ingest_permission_required=True

        # Propagate project piggyback values as default for scheduling_unit_draft
        if self._state.adding and self.piggyback_allowed_tbb is None and hasattr(self, 'scheduling_set'):
            self.piggyback_allowed_tbb = self.scheduling_set.project.piggyback_allowed_tbb
        if self._state.adding and self.piggyback_allowed_aartfaac is None and hasattr(self, 'scheduling_set'):
            self.piggyback_allowed_aartfaac = self.scheduling_set.project.piggyback_allowed_aartfaac
    
        self.annotate_validate_add_defaults_to_doc_using_template('requirements_doc', 'requirements_template')
        self.annotate_validate_add_defaults_to_doc_using_template('scheduling_constraints_doc', 'scheduling_constraints_template')
        super().save(force_insert, force_update, using, update_fields)

    @cached_property
    def duration(self) -> datetime.timedelta:
        '''return the overall duration of all tasks of this scheduling unit
        '''
        return self.relative_stop_time - self.relative_start_time

    @cached_property
    def relative_start_time(self) -> datetime.timedelta:
        '''return the earliest relative start time of all tasks of this scheduling unit
        '''
        task_drafts = list(self.task_drafts.all())
        if task_drafts:
            return min(task_drafts, key=lambda x: x.relative_start_time).relative_start_time
        else:
            return datetime.timedelta(seconds=0)

    @cached_property
    def relative_stop_time(self) -> datetime.timedelta:
        '''return the latest relative stop time of all tasks of this scheduling unit
        '''
        task_drafts = list(self.task_drafts.all())
        if task_drafts:
            return max(task_drafts, key=lambda x: x.relative_stop_time).relative_stop_time
        else:
            return datetime.timedelta(seconds=0)

    @cached_property
    def project(self) -> Project:
        '''return the related project of this task
        '''
        return self.scheduling_set.project


class SchedulingUnitBlueprint(RefreshFromDbInvalidatesCachedPropertiesMixin, TemplateSchemaMixin, NamedCommon):
    class Status(Enum):
        DEFINED = "defined"
        FINISHED = "finished"
        CANCELLED = "cancelled"
        ERROR = "error"
        OBSERVING = "observing"
        OBSERVED = "observed"
        PROCESSING = "processing"
        PROCESSED = "processed"
        INGESTING = "ingesting"
        SCHEDULED = "scheduled"
        SCHEDULABLE = "schedulable"

    # todo: are many of these fields supposed to be immutable in the database?
    #  Or are we fine to just not allow most users to change them?
    requirements_doc = JSONField(help_text='Scheduling and/or quality requirements for this scheduling unit (IMMUTABLE).')
    do_cancel = BooleanField()
    ingest_permission_required = BooleanField(default=False, help_text='Explicit permission is needed before the task.')
    ingest_permission_granted_since = DateTimeField(auto_now_add=False, null=True, help_text='The moment when ingest permission was granted.')
    requirements_template = ForeignKey('SchedulingUnitTemplate', on_delete=CASCADE, help_text='Schema used for requirements_doc (IMMUTABLE).')
    draft = ForeignKey('SchedulingUnitDraft', related_name='scheduling_unit_blueprints', on_delete=PROTECT, help_text='Scheduling Unit Draft which this run instantiates.')
    output_pinned = BooleanField(default=False, help_text='boolean (default FALSE), which blocks deleting unpinned dataproducts. When toggled ON, backend must pick SUB up for deletion. It also must when dataproducts are unpinned.')
    results_accepted = BooleanField(default=False, help_text='boolean (default NULL), which records whether the results were accepted, allowing the higher-level accounting to be adjusted.')
    piggyback_allowed_tbb = BooleanField(help_text='Piggyback key for TBB.', null=True)
    piggyback_allowed_aartfaac = BooleanField(help_text='Piggyback key for AARTFAAC.', null=True)
    priority_rank = FloatField(null=False, default=0.0, help_text='Priority of this scheduling unit w.r.t. other scheduling units within the same queue and project.')
    priority_queue = ForeignKey('PriorityQueueType', null=False, on_delete=PROTECT, default="A", help_text='Priority queue of this scheduling unit. Queues provide a strict ordering between scheduling units.')
    scheduling_constraints_doc = JSONField(help_text='Scheduling Constraints for this run.', null=True)
    scheduling_constraints_template = ForeignKey('SchedulingConstraintsTemplate', on_delete=CASCADE, null=True, help_text='Schema used for scheduling_constraints_doc.')
    is_triggered = BooleanField(default=False, null=False, help_text='boolean (default FALSE), which indicates whether this observation was triggered (responsive telescope)')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # keep original scheduling constraints to detect changes on save
        # Note: we cannot use self.scheduling_constraints_doc here since that causes an infinite loop of update_from_db
        if 'scheduling_constraints_doc' in kwargs.keys():
            self.__original_scheduling_constraints_doc = kwargs['scheduling_constraints_doc']
        else:
            self.__original_scheduling_constraints_doc = None
        self.__original_scheduling_constraints_template_id = self.scheduling_constraints_template_id

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        self.annotate_validate_add_defaults_to_doc_using_template('requirements_doc', 'requirements_template')

        if self._state.adding:
            # On creation, propagate the following scheduling_unit_draft attributes as default for the new scheduling_unit_blueprint
            for copy_field in ['ingest_permission_required', 'piggyback_allowed_tbb', 'piggyback_allowed_aartfaac',
                               'scheduling_constraints_doc', 'scheduling_constraints_template', 'is_triggered']:
                if hasattr(self, 'draft'):
                    setattr(self, copy_field, getattr(self.draft, copy_field))
        else:
            # On updates, prevent changing the scheduling constraints doc or template if we are past schedulable state
            # todo: This causes a ton of tests to fail, e.g. t_workflow_qaworkflow returns errors 422
            if self.status not in [SchedulingUnitBlueprint.Status.DEFINED.value, SchedulingUnitBlueprint.Status.SCHEDULABLE.value] and \
                    ((self.__original_scheduling_constraints_doc is not None and self.scheduling_constraints_doc != self.__original_scheduling_constraints_doc) or
                     self.scheduling_constraints_template_id != self.__original_scheduling_constraints_template_id):
                raise TMSSException('The scheduling constraints of SchedulingUnitBlueprint pk=%s status=%s cannot be updated since it is not in defined or schedulable state.' % (self.pk, self.status))

        # update the original constraints value for comparison on next save
        self.__original_scheduling_constraints_doc = self.scheduling_constraints_doc
        self.__original_scheduling_constraints_template_id = self.scheduling_constraints_template_id

        if self._state.adding and self.is_triggered:
            if self.project.can_trigger:
                from lofar.sas.tmss.services.scheduling.constraints import can_run_after
                start_time = datetime.datetime.utcnow() + datetime.timedelta(minutes=3)
                if self.scheduling_constraints_template is None or can_run_after(self, start_time):
                    logger.info('Triggered obs name=%s can run after start_time=%s. The scheduler will pick this up and cancel ongoing observations if necessary.' % (self.name, start_time))
                else:
                    logger.info('Triggered obs name=%s cannot run after start_time=%s. Adding it for book-keeping, but it will be unschedulable.' % (self.name, start_time))
                    # todo: set to unschedulable? This is a derived state and we do not have subtasks at this point.
            else:
                msg = 'Triggered obs name=%s is rejected because its project name=%s does not allow triggering.' % (self.name, self.project.name)
                logger.warning(msg)
                raise BlueprintCreationException(msg)

        super().save(force_insert, force_update, using, update_fields)

    @cached_property
    def duration(self) -> datetime.timedelta:
        '''return the overall duration of all tasks of this scheduling unit
        '''
        if self.start_time is None or self.stop_time is None:
            return self.relative_stop_time - self.relative_start_time
        else:
            return self.stop_time - self.start_time   # <- todo: do we ever want this?

    @cached_property
    def relative_start_time(self) -> datetime.timedelta:
        '''return the earliest relative start time of all tasks of this scheduling unit
        '''
        task_blueprints = list(self.task_blueprints.all())
        if task_blueprints:
            return min(task_blueprints, key=lambda x: x.relative_start_time).relative_start_time
        else:
            return datetime.timedelta(seconds=0)

    @cached_property
    def relative_stop_time(self) -> datetime.timedelta:
        '''return the latest relative stop time of all tasks of this scheduling unit
        '''
        task_blueprints = list(self.task_blueprints.all())
        if task_blueprints:
            return max(task_blueprints, key=lambda x: x.relative_stop_time).relative_stop_time
        else:
            return datetime.timedelta(seconds=0)

    @cached_property
    def start_time(self) -> datetime or None:
        '''return the earliest start time of all tasks of this scheduling unit
        '''
        tasks_with_start_time = list(filter(lambda x: x.start_time is not None, self.task_blueprints.all()))
        if tasks_with_start_time:
            return min(tasks_with_start_time, key=lambda x: x.start_time).start_time
        else:
            return None

    @cached_property
    def stop_time(self) -> datetime or None:
        '''return the latest stop time of all tasks of this scheduling unit
        '''
        tasks_with_stop_time = list(filter(lambda x: x.stop_time is not None, self.task_blueprints.all()))
        if tasks_with_stop_time:
            return max(tasks_with_stop_time, key=lambda x: x.stop_time).stop_time
        else:
            return None

    @property
    def observed_end_time(self) -> datetime or None:
        """
        return the latest stop time of all (observation) tasks of this scheduling unit with the status observed/finished
        """
        observed_tasks = []
        for task in self.task_blueprints.all():
            if task.specifications_template.type.value == TaskType.Choices.OBSERVATION.value and\
                    (task.status == "observed" or task.status == "finished") and task.stop_time is not None:
                observed_tasks.append(task)
        if observed_tasks:
            return max(observed_tasks, key=lambda x: x.stop_time).stop_time
        else:
            return None

    @property
    def status(self):
        """
        Return the schedulingunit blueprint status which is derived from the taskblueprint status (which is derived
        from the subtasks states)
        See https://support.astron.nl/confluence/display/TMSS/Specification+Flow#SpecificationFlow-TaskBlueprints
        The status is displayed as extra field in rest api of the schedulingunit
        """

        # Determine status per task_type (unfortunately did not manage with updatable view)
        status_overview_counter = Counter()
        status_overview_counter_per_type = {type.value: Counter() for type in TaskType.Choices}
        for tb in self.task_blueprints.select_related('specifications_template').all():
            status_overview_counter[tb.status] += 1
            status_overview_counter_per_type[tb.specifications_template.type.value][tb.status] += 1

        # The actual determination of the SchedulingunitBlueprint status
        if not self._task_graph_instantiated():
            return SchedulingUnitBlueprint.Status.DEFINED.value
        elif self._all_task_finished(status_overview_counter):
            return SchedulingUnitBlueprint.Status.FINISHED.value
        elif self._any_task_cancelled(status_overview_counter):
            return SchedulingUnitBlueprint.Status.CANCELLED.value
        elif self._any_task_error(status_overview_counter):
            return SchedulingUnitBlueprint.Status.ERROR.value
        elif self._any_task_started_observed_finished(status_overview_counter):
            if not self._all_observation_task_observed_finished(status_overview_counter_per_type):
                return SchedulingUnitBlueprint.Status.OBSERVING.value
            elif not self._any_processing_task_started_or_finished(status_overview_counter_per_type):
                return SchedulingUnitBlueprint.Status.OBSERVED.value
            elif not self._all_processing_tasks_and_observation_tasks_finished(status_overview_counter_per_type):
                return SchedulingUnitBlueprint.Status.PROCESSING.value
            elif not self._any_ingest_task_started(status_overview_counter_per_type):
                return SchedulingUnitBlueprint.Status.PROCESSED.value
            else:
                return SchedulingUnitBlueprint.Status.INGESTING.value
        elif self._any_task_scheduled(status_overview_counter):
            return SchedulingUnitBlueprint.Status.SCHEDULED.value
        return SchedulingUnitBlueprint.Status.SCHEDULABLE.value

    @property
    def can_proceed(self) -> bool:
        '''Can this scheduling unit proceed with running its tasks?'''
        return self.status not in [SchedulingUnitBlueprint.Status.ERROR.value, SchedulingUnitBlueprint.Status.FINISHED.value, SchedulingUnitBlueprint.Status.CANCELLED.value]

    def _task_graph_instantiated(self):
        return self._get_total_nbr_tasks() > 0

    def _all_task_finished(self, status_overview_counter):
        return status_overview_counter["finished"] == self._get_total_nbr_tasks()

    @staticmethod
    def _any_task_cancelled(status_overview_counter):
        return status_overview_counter["cancelled"] > 0

    @staticmethod
    def _any_task_error(status_overview_counter):
        return status_overview_counter["error"] > 0

    @staticmethod
    def _any_task_started_observed_finished(status_overview_counter):
        return (status_overview_counter["started"] + status_overview_counter["observed"] + status_overview_counter["finished"]) > 0

    @staticmethod
    def _any_task_scheduled(status_overview_counter):
        return status_overview_counter["scheduled"] > 0

    def _all_observation_task_observed_finished(self, status_overview_counter_per_type):
        total_nbr_observation_tasks = self._get_total_nbr_observation_tasks(status_overview_counter_per_type)
        return (status_overview_counter_per_type["observation"]["observed"] +
                status_overview_counter_per_type["observation"]["finished"]) == total_nbr_observation_tasks

    @staticmethod
    def _any_processing_task_started_or_finished(status_overview_counter_per_type):
        return status_overview_counter_per_type["pipeline"]["started"] + status_overview_counter_per_type["pipeline"]["finished"] > 0

    def _all_processing_tasks_and_observation_tasks_finished(self, status_overview_counter_per_type):
        total_nbr_observation_tasks = self._get_total_nbr_observation_tasks(status_overview_counter_per_type)
        total_nbr_processing_tasks = self._get_total_nbr_processing_tasks(status_overview_counter_per_type)
        return (status_overview_counter_per_type["pipeline"]["finished"] == total_nbr_processing_tasks and
                status_overview_counter_per_type["observation"]["finished"] == total_nbr_observation_tasks)

    @staticmethod
    def _any_ingest_task_started(status_overview_counter_per_type):
        return status_overview_counter_per_type["ingest"]["started"] > 0

    def _get_total_nbr_tasks(self):
        return self.task_blueprints.count()

    @staticmethod
    def _get_total_nbr_observation_tasks(status_overview_counter_per_type):
        return len(tuple(status_overview_counter_per_type[TaskType.Choices.OBSERVATION.value].elements()))

    @staticmethod
    def _get_total_nbr_processing_tasks(status_overview_counter_per_type):
        return len(tuple(status_overview_counter_per_type[TaskType.Choices.PIPELINE.value].elements()))

    @staticmethod
    def _get_total_nbr_ingest_tasks(status_overview_counter_per_type):
        return len(tuple(status_overview_counter_per_type[TaskType.Choices.INGEST.value].elements()))

    @cached_property
    def project(self) -> Project:
        '''return the related project of this task
        '''
        return self.draft.scheduling_set.project

    @property
    def flat_station_list(self):
        """
        Get a flat list of stations of the scheduling unit sorted by name
        """
        lst_stations = []
        for sublist in self._get_recursively(self.requirements_doc, "stations"):
            for item in sublist:
                lst_stations.append(item)
        return list(set(lst_stations))

    @property
    def station_groups(self):
        """
        Get the station groups of the scheduling unit
        """
        lst_station_groups = []
        for sublist in self._get_recursively(self.requirements_doc, "station_groups"):
            for item in sublist:
                lst_station_groups.append(item)
        return lst_station_groups

    def _get_recursively(self, search_dict, field):
        """
        Takes a dict with nested lists and dicts, and searches all dicts for a key of the field provided.
        """
        fields_found = []

        for key, value in search_dict.items():

            if key == field:
                fields_found.append(value)

            elif isinstance(value, dict):
                results = self._get_recursively(value, field)
                for result in results:
                    fields_found.append(result)

            elif isinstance(value, list):
                for item in value:
                    if isinstance(item, dict):
                        more_results = self._get_recursively(item, field)
                        for another_result in more_results:
                            fields_found.append(another_result)

        return fields_found


class TaskDraft(NamedCommon, ProjectPropertyMixin, TemplateSchemaMixin):
    specifications_doc = JSONField(help_text='Specifications for this task.')
    copies = ForeignKey('TaskDraft', related_name="copied_from", on_delete=SET_NULL, null=True, help_text='Source reference, if we are a copy (NULLable).')
    copy_reason = ForeignKey('CopyReason', on_delete=PROTECT, null=True, help_text='Reason why source was copied (NULLable).')
    scheduling_unit_draft = ForeignKey('SchedulingUnitDraft', related_name='task_drafts', on_delete=CASCADE, help_text='Scheduling Unit draft to which this task draft belongs.')
    specifications_template = ForeignKey('TaskTemplate', on_delete=CASCADE, help_text='Schema used for requirements_doc.') # todo: 'schema'?
    output_pinned = BooleanField(default=False, help_text='True if the output of this task is pinned to disk, that is, forbidden to be removed.')
    path_to_project = 'scheduling_unit_draft__scheduling_set__project'

    class Meta:
        # ensure there are no tasks with duplicate names within one scheduling_unit
        constraints = [UniqueConstraint(fields=['name', 'scheduling_unit_draft'], name='TaskDraft_unique_name_in_scheduling_unit')]

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        self.annotate_validate_add_defaults_to_doc_using_template('specifications_doc', 'specifications_template')
        if self._state.adding:  # True on create, False on update, needs to be checked before super().save()
            super().save(force_insert, force_update, using, update_fields)
            if self.scheduling_unit_draft.scheduling_set.project.auto_pin:
                self.output_pinned = True
                self.save()
        else:
            super().save(force_insert, force_update, using, update_fields)

    @cached_property
    def successors(self) -> QuerySet:
        '''return the connect successor taskdraft(s) as queryset (over which you can perform extended queries, or return via the serializers/viewsets)
           If you want the result, add .all() like so: my_taskdraft.successors.all()
        '''
        # see comment on Subtask for why this needs to be raw sql
        return TaskDraft.objects.filter(id__in=RawSQL("SELECT successor_task.id FROM tmssapp_taskdraft as successor_task\n"
                                                      "INNER JOIN tmssapp_taskrelationdraft as task_rel on task_rel.consumer_id = successor_task.id\n"
                                                      "WHERE task_rel.producer_id = %s", params=[self.id]))

    @cached_property
    def predecessors(self) -> QuerySet:
        '''return the connect predecessor taskdraft(s) as queryset (over which you can perform extended queries, or return via the serializers/viewsets)
        If you want the result, add .all() like so: my_taskdraft.predecessors.all()
        '''
        # see comment on Subtask for why this needs to be raw sql
        return TaskDraft.objects.filter(id__in=RawSQL("SELECT successor_task.id FROM tmssapp_taskdraft as successor_task\n"
                                                      "INNER JOIN tmssapp_taskrelationdraft as task_rel on task_rel.producer_id = successor_task.id\n"
                                                      "WHERE task_rel.consumer_id = %s", params=[self.id]))

    @cached_property
    def duration(self) -> datetime.timedelta:
         '''returns the overall duration of this task
         '''
         return self.relative_stop_time - self.relative_start_time

    @cached_property
    def relative_start_time(self) -> datetime.timedelta:
        '''return the earliest relative start time of all subtasks of this task
        '''
        scheduling_relations = list(self.first_scheduling_relation.all()) + list(self.second_scheduling_relation.all())
        for scheduling_relation in scheduling_relations:
            # sometimes self._id does not exist so use self.id instead to avoid Exception
            if hasattr(self, '_id'):
                id = self._id
            else:
                id = self.id
            if scheduling_relation.first.id == id and scheduling_relation.placement_id == "after":
                previous_related_task_draft = TaskDraft.objects.get(id=scheduling_relation.second.id)
                time_offset = scheduling_relation.time_offset
                # todo: max of several relations
                if previous_related_task_draft.relative_stop_time:
                    return previous_related_task_draft.relative_stop_time + datetime.timedelta(seconds=time_offset)

            if scheduling_relation.second.id == id and scheduling_relation.placement_id == "before":
                previous_related_task_draft = TaskDraft.objects.get(id=scheduling_relation.first.id)
                time_offset = scheduling_relation.time_offset
                # todo: max of several relations
                if previous_related_task_draft.relative_stop_time:
                    return previous_related_task_draft.relative_stop_time + datetime.timedelta(seconds=time_offset)
        return datetime.timedelta(seconds=0)

    @cached_property
    def relative_stop_time(self) -> datetime.timedelta:
        '''return the latest relative stop time of all subtasks of this task
        '''
        # todo: when it was added, check if self.specifications_template.type.value == TaskType.Choices.OBSERVATION.value:
        try:
            duration = self.specifications_doc["duration"]
            return self.relative_start_time + datetime.timedelta(seconds=duration)
        except:
            pass
        return self.relative_start_time


    # JK, 28/07/20: After discussion with Sander, we probably only want the
    #  - duration on the scheduling_unit draft (based on relative start/stop times)
    #  - duration plus relative start/stop on the task draft.
    #  This provides an estimate of what is currently planned out in the draft, but does not confuse with timestamps of actual start/stop of the blueprints.
    #  Only on the blueprints, we also aggregate start_stop times as they are in the system
    #  I'll leave these code bits here for now, until we made up our minds about this, but this can probably be removed
    #
    # @cached_property
    # def duration(self) -> datetime.timedelta:
    #     '''returns the overall duration in seconds of all blueprints of this task
    #     # todo: is this the wanted behavior? Do you want to consider all the blueprints created from your draft or do you want to preview a new blueprint?
    #     '''
    #     if self.start_time is None or self.stop_time is None:
    #         # todo: calculate?
    #         return None
    #     else:
    #         return self.stop_time - self.start_time
    #
    # @cached_property
    # def start_time(self) -> datetime or None:
    #     '''return the earliest start time of all blueprints of this task
    #     # todo: is this the wanted behavior? Do you want to consider all the blueprints created from your draft or do you want to preview a new blueprint?
    #     '''
    #     blueprints_with_start_time = list(filter(lambda x: x.start_time is not None, self.task_blueprints.all()))
    #     if blueprints_with_start_time:
    #         return min(blueprints_with_start_time, key=lambda x: x.start_time).start_time
    #     else:
    #         # todo: calculate?
    #         return None
    #
    # @cached_property
    # def stop_time(self) -> datetime or None:
    #     '''return the latest stop time of all blueprints of this task
    #     # todo: is this the wanted behavior? Do you want to consider all the blueprints created from your draft or do you want to preview a new blueprint?
    #     '''
    #     blueprints_with_stop_time = list(filter(lambda x: x.stop_time is not None, self.task_blueprints.all()))
    #     if blueprints_with_stop_time:
    #         return max(blueprints_with_stop_time, key=lambda x: x.stop_time).stop_time
    #     else:
    #         # todo: calculate?
    #         return None


class TaskBlueprint(RefreshFromDbInvalidatesCachedPropertiesMixin, TemplateSchemaMixin, NamedCommon):

    specifications_doc = JSONField(help_text='Schedulings for this task (IMMUTABLE).')
    do_cancel = BooleanField(help_text='Cancel this task.')
    specifications_template = ForeignKey('TaskTemplate', on_delete=CASCADE, help_text='Schema used for specifications_doc (IMMUTABLE).')
    draft = ForeignKey('TaskDraft', related_name='task_blueprints', on_delete=PROTECT, help_text='Task Draft which this task instantiates.')
    scheduling_unit_blueprint = ForeignKey('SchedulingUnitBlueprint', related_name='task_blueprints', on_delete=CASCADE, help_text='Scheduling Unit Blueprint to which this task belongs.')
    output_pinned = BooleanField(default=False, help_text='True if the output of this task is pinned to disk, that is, forbidden to be removed.')

    class Meta:
        # ensure there are no tasks with duplicate names within one scheduling_unit,
        constraints = [UniqueConstraint(fields=['name', 'scheduling_unit_blueprint'], name='TaskBlueprint_unique_name_in_scheduling_unit')]

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        self.annotate_validate_add_defaults_to_doc_using_template('specifications_doc', 'specifications_template')
        super().save(force_insert, force_update, using, update_fields)

    @cached_property
    def successors(self) -> QuerySet:
        '''return the connect successor taskblueprint(s) as queryset (over which you can perform extended queries, or return via the serializers/viewsets)
           If you want the result, add .all() like so: my_taskblueprint.successors.all()
        '''
        # see comment on Subtask for why this needs to be raw sql
        return TaskBlueprint.objects.filter(id__in=RawSQL("SELECT successor_task.id FROM tmssapp_taskblueprint as successor_task\n"
                                                          "INNER JOIN tmssapp_taskrelationblueprint as task_rel on task_rel.consumer_id = successor_task.id\n"
                                                          "WHERE task_rel.producer_id = %s", params=[self.id]))

    @cached_property
    def predecessors(self) -> QuerySet:
        '''return the connect predecessor taskblueprint(s) as queryset (over which you can perform extended queries, or return via the serializers/viewsets)
        If you want the result, add .all() like so: my_taskblueprint.predecessors.all()
        '''
        # see comment on Subtask for why this needs to be raw sql
        return TaskBlueprint.objects.filter(id__in=RawSQL("SELECT predecessor_task.id FROM tmssapp_taskblueprint as predecessor_task\n"
                                                          "INNER JOIN tmssapp_taskrelationblueprint as task_rel on task_rel.producer_id = predecessor_task.id\n"
                                                          "WHERE task_rel.consumer_id = %s", params=[self.id]))

    @cached_property
    def duration(self) -> datetime.timedelta:
        '''return the overall duration of this task
        '''
        if self.start_time is None or self.stop_time is None:
            return self.relative_stop_time - self.relative_start_time
        else:
            return self.stop_time - self.start_time

    @cached_property
    def relative_start_time(self) -> datetime.timedelta:
        '''The relative start time is relative to the start time of the 'start' of the parent scheduling unit.
        It's based on the scheduling_relation's from the scheduling_unit's specification,
        and hence it determines the order in which the tasks within the unit should be executed.
        '''
        scheduling_relations = list(self.first_scheduling_relation.all()) + list(self.second_scheduling_relation.all())
        for scheduling_relation in scheduling_relations:
            # sometimes self._id does not exist so use self.id instead to avoid Exception
            if hasattr(self, '_id'):
                id = self._id
            else:
                id = self.id
            if scheduling_relation.first.id == id and scheduling_relation.placement_id == "after":   # self.id and placement.value will hit the db, this does not
                    previous_related_task_blueprint = TaskBlueprint.objects.get(id=scheduling_relation.second.id)
                    time_offset = scheduling_relation.time_offset
                    # todo: max of several relations
                    if previous_related_task_blueprint.relative_stop_time:
                        return previous_related_task_blueprint.relative_stop_time + datetime.timedelta(seconds=time_offset)

            if scheduling_relation.second.id == id and scheduling_relation.placement_id == "before":   # self.id and placement.value will hit the db, this does not
                    previous_related_task_blueprint = TaskBlueprint.objects.get(id=scheduling_relation.first.id)
                    time_offset = scheduling_relation.time_offset
                    # todo: max of several relations
                    if previous_related_task_blueprint.relative_stop_time:
                        return previous_related_task_blueprint.relative_stop_time + datetime.timedelta(seconds=time_offset)
        return datetime.timedelta(seconds=0)

    @cached_property
    def relative_stop_time(self) -> datetime.timedelta:
        '''The relative_stop_time is the relative_start_time+duration.
        See relative_start_time for an explanation of it's intended usage.
        '''
        # todo: when it was added, check if subtask.specifications_template.type.value == TaskType.Choices.OBSERVATION.value:
        try:
            duration = self.specifications_doc["duration"]
            return self.relative_start_time + datetime.timedelta(seconds=duration)
        except:
            pass
        return self.relative_start_time

    @cached_property
    def start_time(self) -> datetime or None:
        '''return the earliest start time of all subtasks of this task
        '''
        subtasks_with_start_time = list(filter(lambda x: x.start_time is not None, self.subtasks.all()))
        if subtasks_with_start_time:
            return min(subtasks_with_start_time, key=lambda x: x.start_time).start_time
        else:
            return None

    @cached_property
    def stop_time(self) -> datetime or None:
        '''return the latest stop time of all subtasks of this task
        '''
        subtasks_with_stop_time = list(filter(lambda x: x.stop_time is not None, self.subtasks.all()))
        if subtasks_with_stop_time:
            return max(subtasks_with_stop_time, key=lambda x: x.stop_time).stop_time
        else:
            return None

    @property
    def status(self):
        """
        Return the taskblueprint status which is derived from the subtasks status
        See https://support.astron.nl/confluence/display/TMSS/Specification+Flow#SpecificationFlow-TaskBlueprints
        The status is displayed as extra field in rest api of the taskblueprint
        """

        # one call to the db fetching only the needed values
        subtasks = tuple(self.subtasks.values('state', 'specifications_template__type_id').all())

        # simple python filtering counting by state and type. Saves many round-trip-db calls!
        nr_of_subtasks = len(subtasks)
        if nr_of_subtasks == 0:
            return "defined"

        if any(s for s in subtasks if s['state'] == 'defining'):
            return "defined"

        if len([s for s in subtasks if s['state'] == 'finished']) == nr_of_subtasks:
            return "finished"

        if any(s for s in subtasks if s['state'] in ('cancelling', 'cancelled')):
            return "cancelled"

        if any(s for s in subtasks if s['state'] == 'error'):
            return "error"

        observations = [s for s in subtasks if s['specifications_template__type_id'] == 'observation']
        if observations and all(obs and obs['state'] in ('finishing', 'finished') for obs in observations):
            return "observed"

        if any(s for s in subtasks if s['state'] in ('starting','started','queueing','queued','finishing','finished')):
            return "started"

        if any(s for s in subtasks if s['state'] == 'scheduled'):
            return "scheduled"

        return "schedulable"


class TaskRelationDraft(BasicCommon, TemplateSchemaMixin):
    selection_doc = JSONField(help_text='Filter for selecting dataproducts from the output role.')
    selection_template = ForeignKey('TaskRelationSelectionTemplate', on_delete=CASCADE, help_text='Schema used for selection_doc.')

    # caveat: it might look like producer has an incorrect related_name='consumed_by'. But it really is correct, denends on the way you look at it
    producer = ForeignKey('TaskDraft', related_name='consumed_by', on_delete=CASCADE, help_text='Task Draft that has the output connector. NOTE: The producer does typically, but not necessarily, belong to the same Scheduling Unit (or even the same Project) as the consumer.')
    # caveat: it might look like consumer has an incorrect related_name='produced_by'. But it really is correct, denends on the way you look at it
    consumer = ForeignKey('TaskDraft', related_name='produced_by', on_delete=CASCADE, help_text='Task Draft that has the input connector.')

    # this relation descibes a transfer of data from the output_role of the producer to the input_role of the consumer
    input_role = ForeignKey('TaskConnectorType', related_name='taskrelationdraft_input_roles', on_delete=CASCADE, help_text='Input connector type (what kind of data is given to the consumer).')
    output_role = ForeignKey('TaskConnectorType', related_name='taskrelationdraft_output_roles', on_delete=CASCADE, help_text='Output connector type (what kind of data is taken from the producer).')

    class Meta:
        # ensure there are no duplicate relations between tasks with the same in/out roles.
        constraints = [UniqueConstraint(fields=['producer', 'consumer', 'input_role', 'output_role'], name='TaskRelationDraft_unique_relation')]

        # ensure that the roles are compatible, that is, the output we take is suitable for the input we provide to:
        # input_role.dataformat == output_role.dataformat
        # input_role.datatype == outputrole.datatype
        # input_role.output = False
        # output_role.output = True

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        self.annotate_validate_add_defaults_to_doc_using_template('selection_doc', 'selection_template')
        super().save(force_insert, force_update, using, update_fields)


class TaskRelationBlueprint(BasicCommon, TemplateSchemaMixin):
    selection_doc = JSONField(help_text='Filter for selecting dataproducts from the output role.')

    # caveat: it might look like producer has an incorrect related_name='consumed_by'. But it really is correct, denends on the way you look at it
    producer = ForeignKey('TaskBlueprint', related_name='consumed_by', on_delete=CASCADE, help_text='Task Blueprint that has the output connector.')
    # caveat: it might look like consumer has an incorrect related_name='produced_by'. But it really is correct, denends on the way you look at it
    consumer = ForeignKey('TaskBlueprint', related_name='produced_by', on_delete=CASCADE, help_text='Task Blueprint that has the input connector.')

    input_role = ForeignKey('TaskConnectorType', related_name='taskrelationblueprint_input_roles', on_delete=CASCADE, help_text='Input connector type (what kind of data can be taken as input).')
    output_role = ForeignKey('TaskConnectorType', related_name='taskrelationblueprint_output_roles', on_delete=CASCADE, help_text='Output connector type (what kind of data can be created as output).')
    draft = ForeignKey('TaskRelationDraft', on_delete=CASCADE, related_name='related_task_relation_blueprint', help_text='Task Relation Draft which this work request instantiates.')
    selection_template = ForeignKey('TaskRelationSelectionTemplate', on_delete=CASCADE, help_text='Schema used for selection_doc.')  # todo: 'schema'?

    class Meta:
        # ensure there are no duplicate relations between tasks with the same in/out roles.
        constraints = [UniqueConstraint(fields=['producer', 'consumer', 'input_role', 'output_role'], name='TaskRelationBlueprint_unique_relation')]

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        self.annotate_validate_add_defaults_to_doc_using_template('selection_doc', 'selection_template')
        super().save(force_insert, force_update, using, update_fields)


class TaskSchedulingRelationBlueprint(BasicCommon):
    first = ForeignKey('TaskBlueprint', related_name='first_scheduling_relation', on_delete=CASCADE, help_text='First Task Blueprint to connect.')
    second = ForeignKey('TaskBlueprint', related_name='second_scheduling_relation', on_delete=CASCADE, help_text='Second Task Blueprint to connect.')
    placement = ForeignKey('SchedulingRelationPlacement', null=False, default="after", on_delete=PROTECT, help_text='Task scheduling relation placement.')
    time_offset = IntegerField(default=60, help_text='Time offset of start of second task with respect to start of first task.')  
  
    class Meta:
        # ensure there are no duplicate scheduling relations between tasks
        constraints = [UniqueConstraint(fields=['first', 'second'], name='TaskSchedulingRelationBlueprint_unique_relation')]

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        if self.first == self.second:
            raise ValidationError("First and Second Task Draft must be different.")
        if (str(self.placement) == SchedulingRelationPlacement.Choices.BEFORE.value or str(self.placement) == SchedulingRelationPlacement.Choices.AFTER.value) and self.time_offset<0:
            raise ValidationError("Time_offset must be >= 0")
        super().save(force_insert, force_update, using, update_fields)
    

class TaskSchedulingRelationDraft(BasicCommon):
    first = ForeignKey('TaskDraft', related_name='first_scheduling_relation', on_delete=CASCADE, help_text='First Task Draft to connect.')
    second = ForeignKey('TaskDraft', related_name='second_scheduling_relation', on_delete=CASCADE, help_text='Second Task Draft to connect.')
    placement = ForeignKey('SchedulingRelationPlacement', null=False, on_delete=PROTECT, help_text='Task scheduling relation placement.')
    time_offset = IntegerField(default=60, help_text='Time offset of start of second task with respect to start of first task.')  

    class Meta:
        # ensure there are no duplicate scheduling relations between tasks
        constraints = [UniqueConstraint(fields=['first', 'second'], name='TaskSchedulingRelationDraft_unique_relation')]

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        if self.first == self.second:
            raise ValidationError("First and Second Task Draft must be different.")
        if (str(self.placement) == SchedulingRelationPlacement.Choices.BEFORE.value or str(self.placement) == SchedulingRelationPlacement.Choices.AFTER.value) and self.time_offset<0:
            raise ValidationError("Time_offset must be >= 0")
        super().save(force_insert, force_update, using, update_fields)


class Reservation(NamedCommon, TemplateSchemaMixin):
    project = ForeignKey('Project', null=True, related_name='reservations', on_delete=CASCADE, help_text='Reservation will be accounted for this project.')
    description = CharField(max_length=255, help_text='Short description for this reservation, used in overviews')
    start_time = DateTimeField(help_text='Start of this reservation.')
    stop_time = DateTimeField(null=True, help_text='Stop of this reservation. If null, then this reservation is indefinitely.')
    specifications_doc = JSONField(help_text='Properties of this reservation')
    specifications_template = ForeignKey('ReservationTemplate', on_delete=CASCADE, help_text='Schema used for specifications_doc.')

    @property
    def duration(self) -> int:
        '''return the overall duration (in seconds) of this task, if stop_time in None than duration ia also None
        '''
        if self.stop_time:
            return (self.stop_time - self.start_time).total_seconds()
        else:
            return None

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        self.annotate_validate_add_defaults_to_doc_using_template('specifications_doc', 'specifications_template')
        super().save(force_insert, force_update, using, update_fields)

