import os
import unittest
import requests
import json

import logging
logger = logging.getLogger(__name__)
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

from lofar.common.test_utils import skip_integration_tests
if skip_integration_tests():
    exit(3)

from lofar.messaging.messagebus import TemporaryExchange, BusListenerJanitor
from lofar.sas.tmss.services.workflow_service import create_workflow_service, SchedulingUnitEventMessageHandler

from time import sleep
from datetime import datetime, timedelta
import uuid
from threading import Thread, Event
from lofar.sas.tmss.client.tmssbuslistener import *



class SchedulingUnitFlowTest(unittest.TestCase):


    @classmethod
    def setUpClass(cls) -> None:
        cls.TEST_UUID = uuid.uuid1()

        cls.tmp_exchange = TemporaryExchange("%s_%s" % (cls.__name__, cls.TEST_UUID))
        cls.tmp_exchange.open()

        # override DEFAULT_BUSNAME
        import lofar
        lofar.messaging.config.DEFAULT_BUSNAME = cls.tmp_exchange.address

        # import here, and not at top of module, because DEFAULT_BUSNAME needs to be set before importing
        from lofar.sas.resourceassignment.resourceassigner.test.ra_test_environment import RATestEnvironment
        from lofar.sas.tmss.test.test_environment import TMSSTestEnvironment

        cls.ra_test_env = RATestEnvironment(exchange=cls.tmp_exchange.address)
        cls.ra_test_env.start()

        cls.tmss_test_env = TMSSTestEnvironment(exchange=cls.tmp_exchange.address, populate_schemas=True, populate_test_data=False,
                                                start_subtask_scheduler=False, start_postgres_listener=True, start_ra_test_environment=False,
                                                start_dynamic_scheduler=False, enable_viewflow=True, start_workflow_service=False)
        cls.tmss_test_env.start()
        
        cls.BASE_URL_WF_API = cls.tmss_test_env.django_server.url.rstrip('/').replace('/api','/workflow_api')
        cls.AUTH = requests.auth.HTTPBasicAuth(cls.tmss_test_env.ldap_server.dbcreds.user, cls.tmss_test_env.ldap_server.dbcreds.password)

        import time
        
        cls.sync_event_bp_scheduled = Event()
        cls.sync_event_bp_cannot_proceed = Event()

        class TestSchedulingUnitEventMessageHandler(SchedulingUnitEventMessageHandler):
            def onSchedulingUnitBlueprintStatusChanged(self, id: int, status: str):
                super().onSchedulingUnitBlueprintStatusChanged(id=id, status=status)
                if status == "scheduled":
                    cls.sync_event_bp_scheduled.set()

            def onSchedulingUnitBlueprintCannotProceed(self, id: int):
                super().onSchedulingUnitBlueprintCannotProceed(id=id)
                cls.sync_event_bp_cannot_proceed.set()
        


    @classmethod
    def tearDownClass(cls) -> None:
        cls.tmss_test_env.stop()
        cls.ra_test_env.stop()
        cls.tmp_exchange.close()

    def test_qa_workflow_complete(self):
        from lofar.sas.tmss.tmss.workflowapp.flows.schedulingunitflow import SchedulingUnitFlow

        from lofar.sas.tmss.tmss.tmssapp import models
        from lofar.sas.tmss.tmss.tmssapp.models import TaskType

        from lofar.sas.tmss.tmss.tmssapp.tasks import create_task_blueprints_and_subtasks_from_scheduling_unit_draft
        from lofar.sas.tmss.test.tmss_test_data_django_models import SchedulingSet_test_data

        from lofar.sas.tmss.tmss.workflowapp.models.schedulingunitflow import SchedulingUnitProcess
        from viewflow.models import Task
        
        sync_event_bp_scheduled = Event()
        sync_event_bp_cannot_proceed = Event()


        class TestSchedulingUnitEventMessageHandler(SchedulingUnitEventMessageHandler):
            def onSchedulingUnitBlueprintStatusChanged(self, id: int, status: str):
                super().onSchedulingUnitBlueprintStatusChanged(id=id, status=status)
               
               
                if status == "scheduled":
                    logging.info("Status is %s, sending sync event",status)
                    sync_event_bp_scheduled.set()

            def onSchedulingUnitBlueprintCannotProceed(self, id: int):
                super().onSchedulingUnitBlueprintCannotProceed(id=id)
                logging.info("Scheduling Unit Blueprint with id %s cannot proceed, sending sync event",id)
                sync_event_bp_cannot_proceed.set()

        service = create_workflow_service(handler_type=TestSchedulingUnitEventMessageHandler,
                                          exchange=self.tmp_exchange.address)
        with BusListenerJanitor(service):
            strategy_template = models.SchedulingUnitObservingStrategyTemplate.objects.get(name="UC1 CTC+pipelines")

            scheduling_unit_draft = models.SchedulingUnitDraft.objects.create(
                                            name="Test Scheduling Unit UC1",
                                            requirements_doc=strategy_template.template,
                                            requirements_template=strategy_template.scheduling_unit_template,
                                            observation_strategy_template=strategy_template,
                                            copy_reason=models.CopyReason.objects.get(value='template'),
                                            generator_instance_doc="para",
                                            copies=None,
                                            scheduling_set=models.SchedulingSet.objects.create(**SchedulingSet_test_data()))
            
            scheduling_unit_blueprint = create_task_blueprints_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
            #ingest_subtask = models.Subtask.objects.get(task_blueprints__scheduling_unit_blueprint__id=scheduling_unit_blueprint.id,
            #    specifications_template__type__value=TaskType.Choices.INGEST.value)
            scheduling_unit_draft.refresh_from_db()
            
            # there is no signal that SchedulingUnitProcess instance was created,
            # so we have to wait an poll before we can proceed with the test
            poll_starttime = datetime.utcnow()
            while True:
                if SchedulingUnitProcess.objects.filter(su=scheduling_unit_blueprint).count() > 0:
                    break
                sleep(0.1)
                if datetime.utcnow()-poll_starttime > timedelta(seconds=10):
                    raise TimeoutError("SchedulingUnitProcess not created within expected time")

            # Yes! the SchedulingUnitProcess was created, let's get it.
            scheduling_unit_process_id = SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).id
            
            prev_ingest_permission_granted_since = models.SchedulingUnitBlueprint.objects.get(pk=scheduling_unit_process_id).ingest_permission_granted_since
            self.assertEqual(prev_ingest_permission_granted_since, None)

            #check the active task name
            self.assertEqual("wait_scheduled", SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name)
                
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[0].flow_task.name, 'start')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[0].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].flow_task.name, 'wait_scheduled')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].status, 'NEW')


            #Change subtask status to scheduled
            from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions
            for task_blueprint in scheduling_unit_blueprint.task_blueprints.all():
                if task_blueprint.specifications_template.type.value != TaskType.Choices.INGEST.value:
                    for subtask in task_blueprint.subtasks.all():
                        set_subtask_state_following_allowed_transitions(subtask, 'scheduled')

            # wait until scheduling unit is scheduled
            if not sync_event_bp_scheduled.wait(timeout=10):
                logging.info("sync_event_bp_scheduled event not received, raising TimeoutError")
                raise TimeoutError()
            else:
                logging.info("Received sync_event_bp_scheduled event")
                poll_starttime = datetime.utcnow()
                while True:
                    if SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name == "wait_processed":
                        break
                    sleep(0.1)
                    if datetime.utcnow()-poll_starttime > timedelta(seconds=10):
                        raise TimeoutError("Task not activated within expected time")

            sync_event_bp_scheduled.clear()

           
            #check the active task name
            self.assertEqual("wait_processed", SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name)
                
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].flow_task.name, 'wait_scheduled')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].flow_task.name, 'wait_processed')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].status, 'NEW')

            #Change subtask status to finished
            from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions
            for task_blueprint in scheduling_unit_blueprint.task_blueprints.all():
                task_blueprint.output_pinned=True
                task_blueprint.save()
                
                for subtask in task_blueprint.subtasks.all():
                    set_subtask_state_following_allowed_transitions(subtask, 'finished')

            if not sync_event_bp_cannot_proceed.wait(timeout=10):
                logging.info("sync_event_bp_cannot_proceed event not received, raising TimeoutError")
                raise TimeoutError()
            else:
                logging.info("Received sync_event_bp_cannot_proceed event")
                poll_starttime = datetime.utcnow()
                while True:
                    if SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name == "qa_reporting_to":
                        break
                    sleep(0.1)
                    if datetime.utcnow()-poll_starttime > timedelta(seconds=10):
                        raise TimeoutError("Task not activated within expected time")

            sync_event_bp_cannot_proceed.clear()

            #check the active task name
            self.assertEqual("qa_reporting_to", SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].flow_task.name, 'wait_processed')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].flow_task.name, 'qa_reporting_to')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].status, 'NEW')

            #API: Get current task
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            task_id=r_dict[0]['pk']

            #API: Assign to an user
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].flow_task.name, 'qa_reporting_to')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].status, 'ASSIGNED')
           
            #API: Perform qa_reporting_to step
            headers = {'content-type': 'application/json'}
            data = '{"operator_report": "Test report", "operator_accept": true}'
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            self.assertEqual(200,response.status_code)
           
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].flow_task.name, 'qa_reporting_to')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[4].flow_task.name, 'check_operator_accept')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[4].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].flow_task.name, 'qa_reporting_sos')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].status, 'NEW')
            
            #API: Get current task
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            task_id=r_dict[0]['pk']

            #API: Assign to an user
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            #API: Perform qa_reporting_sos step
            headers = {'content-type': 'application/json'}
            data = '{"sos_report": "Test report", "quality_within_policy": true, "sos_accept_show_pi": true}'
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].flow_task.name, 'qa_reporting_sos')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].status, 'DONE')
           
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[6].flow_task.name, 'check_sos_accept_show_pi')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[6].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].flow_task.name, 'pi_verification')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].status, 'NEW')

            #API: Get current task
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            task_id=r_dict[0]['pk']

            #API: Assign to an user
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)
      
            #API: Perform pi_verification step
            headers = {'content-type': 'application/json'}
            data = '{"pi_report": "Test report", "pi_accept": true}'
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].flow_task.name, 'pi_verification')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[8].flow_task.name, 'decide_acceptance')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[8].status, 'NEW')

            #API: Get current task
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            task_id=r_dict[0]['pk']

            #API: Assign to an user
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            #API: Perform decide_acceptance step
            headers = {'content-type': 'application/json'}
            data = '{"sos_accept_after_pi": true}'
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[8].flow_task.name, 'decide_acceptance')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[8].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[9].flow_task.name, 'check_sos_accept_after_pi')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[9].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[10].flow_task.name, 'allow_ingest')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[10].status, 'DONE')
            
            #verify that ingest_permission_granted_since is now a valid datetime
            ingest_permission_granted_since = models.SchedulingUnitBlueprint.objects.get(pk=scheduling_unit_process_id).ingest_permission_granted_since
            self.assertEqual(True,isinstance(ingest_permission_granted_since, datetime))
            
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[11].flow_task.name, 'ingest_done')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[11].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[12].flow_task.name, 'mark_sub')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[12].status, 'DONE')
            
            output_pinned = models.SchedulingUnitBlueprint.objects.get(pk=scheduling_unit_process_id).output_pinned
            self.assertEqual(True,output_pinned)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[13].flow_task.name, 'check_data_pinned')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[13].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[14].flow_task.name, 'unpin_data')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[14].status, 'NEW')

            #API: Get current task
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            task_id=r_dict[0]['pk']

            #API: Assign to an user
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            #API: Perform decide_acceptance step
            headers = {'content-type': 'application/json'}
            data = '{"unpin_data": true}'
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[14].flow_task.name, 'unpin_data')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[14].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[15].flow_task.name, 'delete_data')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[15].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[16].flow_task.name, 'end')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[16].status, 'DONE')         
    

    def test_qa_workflow_qa_reporting_to_no(self):
        from lofar.sas.tmss.tmss.workflowapp.flows.schedulingunitflow import SchedulingUnitFlow

        from lofar.sas.tmss.tmss.tmssapp import models
        from lofar.sas.tmss.tmss.tmssapp.models import TaskType

        from lofar.sas.tmss.tmss.tmssapp.tasks import create_task_blueprints_and_subtasks_from_scheduling_unit_draft
        from lofar.sas.tmss.test.tmss_test_data_django_models import SchedulingSet_test_data

        from lofar.sas.tmss.tmss.workflowapp.models.schedulingunitflow import SchedulingUnitProcess
        from viewflow.models import Task

        sync_event_bp_scheduled = Event()
        sync_event_bp_cannot_proceed = Event()


        class TestSchedulingUnitEventMessageHandler(SchedulingUnitEventMessageHandler):
            def onSchedulingUnitBlueprintStatusChanged(self, id: int, status: str):
                super().onSchedulingUnitBlueprintStatusChanged(id=id, status=status)
               
               
                if status == "scheduled":
                    logging.info("Status is %s, sending sync event",status)
                    sync_event_bp_scheduled.set()

            def onSchedulingUnitBlueprintCannotProceed(self, id: int):
                super().onSchedulingUnitBlueprintCannotProceed(id=id)
                logging.info("Scheduling Unit Blueprint with id %s cannot proceed, sending sync event",id)
                sync_event_bp_cannot_proceed.set()

        service = create_workflow_service(handler_type=TestSchedulingUnitEventMessageHandler,
                                          exchange=self.tmp_exchange.address)
        with BusListenerJanitor(service):
            
            strategy_template = models.SchedulingUnitObservingStrategyTemplate.objects.get(name="UC1 CTC+pipelines")

            scheduling_unit_draft = models.SchedulingUnitDraft.objects.create(
                                            name="Test Scheduling Unit UC1",
                                            requirements_doc=strategy_template.template,
                                            requirements_template=strategy_template.scheduling_unit_template,
                                            observation_strategy_template=strategy_template,
                                            copy_reason=models.CopyReason.objects.get(value='template'),
                                            generator_instance_doc="para",
                                            copies=None,
                                            scheduling_set=models.SchedulingSet.objects.create(**SchedulingSet_test_data()))
            
            scheduling_unit_blueprint = create_task_blueprints_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
            #ingest_subtask = models.Subtask.objects.get(task_blueprints__scheduling_unit_blueprint__id=scheduling_unit_blueprint.id,
            #    specifications_template__type__value=TaskType.Choices.INGEST.value)
            scheduling_unit_draft.refresh_from_db()
            
            # there is no signal that SchedulingUnitProcess instance was created,
            # so we have to wait an poll before we can proceed with the test
            poll_starttime = datetime.utcnow()
            while True:
                if SchedulingUnitProcess.objects.filter(su=scheduling_unit_blueprint).count() > 0:
                    break
                sleep(0.1)
                if datetime.utcnow()-poll_starttime > timedelta(seconds=10):
                    raise TimeoutError("SchedulingUnitProcess not created within expected time")

            # Yes! the SchedulingUnitProcess was created, let's get it.
            scheduling_unit_process_id = SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).id
            
            prev_ingest_permission_granted_since = models.SchedulingUnitBlueprint.objects.get(pk=scheduling_unit_process_id).ingest_permission_granted_since
            self.assertEqual(prev_ingest_permission_granted_since, None)

          
            self.assertEqual("wait_scheduled", SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name)
                
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[0].flow_task.name, 'start')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[0].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].flow_task.name, 'wait_scheduled')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].status, 'NEW')


            #Change subtask status to scheduled
            from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions
            for task_blueprint in scheduling_unit_blueprint.task_blueprints.all():
                if task_blueprint.specifications_template.type.value != TaskType.Choices.INGEST.value:
                    for subtask in task_blueprint.subtasks.all():
                        set_subtask_state_following_allowed_transitions(subtask, 'scheduled')

            # wait until scheduling unit is scheduled
            if not sync_event_bp_scheduled.wait(timeout=10):
                logging.info("sync_event_bp_scheduled event not received, raising TimeoutError")
                raise TimeoutError()
            else:
                logging.info("Received sync_event_bp_scheduled event")
                poll_starttime = datetime.utcnow()
                while True:
                    if SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name == "wait_processed":
                        break
                    sleep(0.1)
                    if datetime.utcnow()-poll_starttime > timedelta(seconds=10):
                        raise TimeoutError("Task not activated within expected time")

            sync_event_bp_scheduled.clear()

           
            #check the active task name
            self.assertEqual("wait_processed", SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name)
                
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].flow_task.name, 'wait_scheduled')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].flow_task.name, 'wait_processed')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].status, 'NEW')

            #Change subtask status to finished
            from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions
            for task_blueprint in scheduling_unit_blueprint.task_blueprints.all():
                task_blueprint.output_pinned=True
                task_blueprint.save()
                
                for subtask in task_blueprint.subtasks.all():
                    set_subtask_state_following_allowed_transitions(subtask, 'finished')

            if not sync_event_bp_cannot_proceed.wait(timeout=10):
                logging.info("sync_event_bp_cannot_proceed event not received, raising TimeoutError")
                raise TimeoutError()
            else:
                logging.info("Received sync_event_bp_cannot_proceed event")
                poll_starttime = datetime.utcnow()
                while True:
                    if SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name == "qa_reporting_to":
                        break
                    sleep(0.1)
                    if datetime.utcnow()-poll_starttime > timedelta(seconds=10):
                        raise TimeoutError("Task not activated within expected time")

            sync_event_bp_cannot_proceed.clear()

           
            #check the active task name
            self.assertEqual("qa_reporting_to", SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].flow_task.name, 'wait_processed')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].flow_task.name, 'qa_reporting_to')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].status, 'NEW')

            #API: Get current task
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            task_id=r_dict[0]['pk']

            #API: Assign to an user
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].flow_task.name, 'qa_reporting_to')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].status, 'ASSIGNED')
           
            #API: Perform qa_reporting_to step
            headers = {'content-type': 'application/json'}
            data = '{"operator_report": "Test report", "operator_accept": false}'
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            self.assertEqual(200,response.status_code)
           
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].flow_task.name, 'qa_reporting_to')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].status, 'DONE')
            
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[4].flow_task.name, 'check_operator_accept')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[4].status, 'DONE')
            
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].flow_task.name, 'mark_sub')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[6].flow_task.name, 'check_data_pinned')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[6].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].flow_task.name, 'unpin_data')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].status, 'NEW')
            

            #API: Get current task
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            task_id=r_dict[0]['pk']

            #API: Assign to an user
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            #API: Perform decide_acceptance step
            headers = {'content-type': 'application/json'}
            data = '{"unpin_data": true}'
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].flow_task.name, 'unpin_data')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[8].flow_task.name, 'delete_data')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[8].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[9].flow_task.name, 'end')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[9].status, 'DONE')         


    def test_qa_workflow_qa_reporting_sos_no(self):
        from lofar.sas.tmss.tmss.workflowapp.flows.schedulingunitflow import SchedulingUnitFlow

        from lofar.sas.tmss.tmss.tmssapp import models
        from lofar.sas.tmss.tmss.tmssapp.models import TaskType

        from lofar.sas.tmss.tmss.tmssapp.tasks import create_task_blueprints_and_subtasks_from_scheduling_unit_draft
        from lofar.sas.tmss.test.tmss_test_data_django_models import SchedulingSet_test_data

        from lofar.sas.tmss.tmss.workflowapp.models.schedulingunitflow import SchedulingUnitProcess
        from viewflow.models import Task
        
        sync_event_bp_scheduled = Event()
        sync_event_bp_cannot_proceed = Event()


        class TestSchedulingUnitEventMessageHandler(SchedulingUnitEventMessageHandler):
            def onSchedulingUnitBlueprintStatusChanged(self, id: int, status: str):
                super().onSchedulingUnitBlueprintStatusChanged(id=id, status=status)
               
               
                if status == "scheduled":
                    logging.info("Status is %s, sending sync event",status)
                    sync_event_bp_scheduled.set()

            def onSchedulingUnitBlueprintCannotProceed(self, id: int):
                super().onSchedulingUnitBlueprintCannotProceed(id=id)
                logging.info("Scheduling Unit Blueprint with id %s cannot proceed, sending sync event",id)
                sync_event_bp_cannot_proceed.set()

        service = create_workflow_service(handler_type=TestSchedulingUnitEventMessageHandler,
                                          exchange=self.tmp_exchange.address)
        with BusListenerJanitor(service):
            strategy_template = models.SchedulingUnitObservingStrategyTemplate.objects.get(name="UC1 CTC+pipelines")

            scheduling_unit_draft = models.SchedulingUnitDraft.objects.create(
                                            name="Test Scheduling Unit UC1",
                                            requirements_doc=strategy_template.template,
                                            requirements_template=strategy_template.scheduling_unit_template,
                                            observation_strategy_template=strategy_template,
                                            copy_reason=models.CopyReason.objects.get(value='template'),
                                            generator_instance_doc="para",
                                            copies=None,
                                            scheduling_set=models.SchedulingSet.objects.create(**SchedulingSet_test_data()))
            
            scheduling_unit_blueprint = create_task_blueprints_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
            #ingest_subtask = models.Subtask.objects.get(task_blueprints__scheduling_unit_blueprint__id=scheduling_unit_blueprint.id,
            #    specifications_template__type__value=TaskType.Choices.INGEST.value)
            scheduling_unit_draft.refresh_from_db()
            
            # there is no signal that SchedulingUnitProcess instance was created,
            # so we have to wait an poll before we can proceed with the test
            poll_starttime = datetime.utcnow()
            while True:
                if SchedulingUnitProcess.objects.filter(su=scheduling_unit_blueprint).count() > 0:
                    break
                sleep(0.1)
                if datetime.utcnow()-poll_starttime > timedelta(seconds=10):
                    raise TimeoutError("SchedulingUnitProcess not created within expected time")

            # Yes! the SchedulingUnitProcess was created, let's get it.
            scheduling_unit_process_id = SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).id
            
            prev_ingest_permission_granted_since = models.SchedulingUnitBlueprint.objects.get(pk=scheduling_unit_process_id).ingest_permission_granted_since
            self.assertEqual(prev_ingest_permission_granted_since, None)

            #check the active task name
            self.assertEqual("wait_scheduled", SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name)
                
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[0].flow_task.name, 'start')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[0].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].flow_task.name, 'wait_scheduled')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].status, 'NEW')


            #Change subtask status to scheduled
            from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions
            for task_blueprint in scheduling_unit_blueprint.task_blueprints.all():
                if task_blueprint.specifications_template.type.value != TaskType.Choices.INGEST.value:
                    for subtask in task_blueprint.subtasks.all():
                        set_subtask_state_following_allowed_transitions(subtask, 'scheduled')

            # wait until scheduling unit is scheduled
            if not sync_event_bp_scheduled.wait(timeout=10):
                logging.info("sync_event_bp_scheduled event not received, raising TimeoutError")
                raise TimeoutError()
            else:
                logging.info("Received sync_event_bp_scheduled event")
                poll_starttime = datetime.utcnow()
                while True:
                    if SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name == "wait_processed":
                        break
                    sleep(0.1)
                    if datetime.utcnow()-poll_starttime > timedelta(seconds=10):
                        raise TimeoutError("Task not activated within expected time")

            sync_event_bp_scheduled.clear()

           
            #check the active task name
            self.assertEqual("wait_processed", SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name)
                
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].flow_task.name, 'wait_scheduled')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].flow_task.name, 'wait_processed')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].status, 'NEW')

            #Change subtask status to finished
            from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions
            for task_blueprint in scheduling_unit_blueprint.task_blueprints.all():
                task_blueprint.output_pinned=True
                task_blueprint.save()
                
                for subtask in task_blueprint.subtasks.all():
                    set_subtask_state_following_allowed_transitions(subtask, 'finished')

            if not sync_event_bp_cannot_proceed.wait(timeout=10):
                logging.info("sync_event_bp_cannot_proceed event not received, raising TimeoutError")
                raise TimeoutError()
            else:
                logging.info("Received sync_event_bp_cannot_proceed event")
                poll_starttime = datetime.utcnow()
                while True:
                    if SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name == "qa_reporting_to":
                        break
                    sleep(0.1)
                    if datetime.utcnow()-poll_starttime > timedelta(seconds=10):
                        raise TimeoutError("Task not activated within expected time")

            sync_event_bp_cannot_proceed.clear()

            #check the active task name
            self.assertEqual("qa_reporting_to", SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].flow_task.name, 'wait_processed')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].flow_task.name, 'qa_reporting_to')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].status, 'NEW')

            #API: Get current task
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            task_id=r_dict[0]['pk']

            #API: Assign to an user
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].flow_task.name, 'qa_reporting_to')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].status, 'ASSIGNED')
           
            #API: Perform qa_reporting_to step
            headers = {'content-type': 'application/json'}
            data = '{"operator_report": "Test report", "operator_accept": true}'
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            self.assertEqual(200,response.status_code)
           
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].flow_task.name, 'qa_reporting_to')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[4].flow_task.name, 'check_operator_accept')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[4].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].flow_task.name, 'qa_reporting_sos')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].status, 'NEW')
            
            #API: Get current task
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            task_id=r_dict[0]['pk']

            #API: Assign to an user
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            #API: Perform qa_reporting_sos step
            headers = {'content-type': 'application/json'}
            data = '{"sos_report": "Test report", "quality_within_policy": true, "sos_accept_show_pi": false}'
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].flow_task.name, 'qa_reporting_sos')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].status, 'DONE')
           
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[6].flow_task.name, 'check_sos_accept_show_pi')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[6].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].flow_task.name, 'mark_sub')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].status, 'DONE')
            
            output_pinned = models.SchedulingUnitBlueprint.objects.get(pk=scheduling_unit_process_id).output_pinned
            self.assertEqual(True,output_pinned)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[8].flow_task.name, 'check_data_pinned')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[8].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[9].flow_task.name, 'unpin_data')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[9].status, 'NEW')

            #API: Get current task
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            task_id=r_dict[0]['pk']

            #API: Assign to an user
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            #API: Perform decide_acceptance step
            headers = {'content-type': 'application/json'}
            data = '{"unpin_data": true}'
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[9].flow_task.name, 'unpin_data')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[9].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[10].flow_task.name, 'delete_data')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[10].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[11].flow_task.name, 'end')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[11].status, 'DONE')         

    def test_qa_workflow_qa_quality_acceptable_no(self):
        from lofar.sas.tmss.tmss.workflowapp.flows.schedulingunitflow import SchedulingUnitFlow

        from lofar.sas.tmss.tmss.tmssapp import models
        from lofar.sas.tmss.tmss.tmssapp.models import TaskType

        from lofar.sas.tmss.tmss.tmssapp.tasks import create_task_blueprints_and_subtasks_from_scheduling_unit_draft
        from lofar.sas.tmss.test.tmss_test_data_django_models import SchedulingSet_test_data

        from lofar.sas.tmss.tmss.workflowapp.models.schedulingunitflow import SchedulingUnitProcess
        from viewflow.models import Task
        
        sync_event_bp_scheduled = Event()
        sync_event_bp_cannot_proceed = Event()


        class TestSchedulingUnitEventMessageHandler(SchedulingUnitEventMessageHandler):
            def onSchedulingUnitBlueprintStatusChanged(self, id: int, status: str):
                super().onSchedulingUnitBlueprintStatusChanged(id=id, status=status)
               
               
                if status == "scheduled":
                    logging.info("Status is %s, sending sync event",status)
                    sync_event_bp_scheduled.set()

            def onSchedulingUnitBlueprintCannotProceed(self, id: int):
                super().onSchedulingUnitBlueprintCannotProceed(id=id)
                logging.info("Scheduling Unit Blueprint with id %s cannot proceed, sending sync event",id)
                sync_event_bp_cannot_proceed.set()

        service = create_workflow_service(handler_type=TestSchedulingUnitEventMessageHandler,
                                          exchange=self.tmp_exchange.address)
        with BusListenerJanitor(service):
            strategy_template = models.SchedulingUnitObservingStrategyTemplate.objects.get(name="UC1 CTC+pipelines")

            scheduling_unit_draft = models.SchedulingUnitDraft.objects.create(
                                            name="Test Scheduling Unit UC1",
                                            requirements_doc=strategy_template.template,
                                            requirements_template=strategy_template.scheduling_unit_template,
                                            observation_strategy_template=strategy_template,
                                            copy_reason=models.CopyReason.objects.get(value='template'),
                                            generator_instance_doc="para",
                                            copies=None,
                                            scheduling_set=models.SchedulingSet.objects.create(**SchedulingSet_test_data()))
            
            scheduling_unit_blueprint = create_task_blueprints_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
            #ingest_subtask = models.Subtask.objects.get(task_blueprints__scheduling_unit_blueprint__id=scheduling_unit_blueprint.id,
            #    specifications_template__type__value=TaskType.Choices.INGEST.value)
            scheduling_unit_draft.refresh_from_db()
            
            # there is no signal that SchedulingUnitProcess instance was created,
            # so we have to wait an poll before we can proceed with the test
            poll_starttime = datetime.utcnow()
            while True:
                if SchedulingUnitProcess.objects.filter(su=scheduling_unit_blueprint).count() > 0:
                    break
                sleep(0.1)
                if datetime.utcnow()-poll_starttime > timedelta(seconds=10):
                    raise TimeoutError("SchedulingUnitProcess not created within expected time")

            # Yes! the SchedulingUnitProcess was created, let's get it.
            scheduling_unit_process_id = SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).id
            
            prev_ingest_permission_granted_since = models.SchedulingUnitBlueprint.objects.get(pk=scheduling_unit_process_id).ingest_permission_granted_since
            self.assertEqual(prev_ingest_permission_granted_since, None)

            #check the active task name
            self.assertEqual("wait_scheduled", SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name)
                
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[0].flow_task.name, 'start')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[0].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].flow_task.name, 'wait_scheduled')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].status, 'NEW')


            #Change subtask status to scheduled
            from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions
            for task_blueprint in scheduling_unit_blueprint.task_blueprints.all():
                if task_blueprint.specifications_template.type.value != TaskType.Choices.INGEST.value:
                    for subtask in task_blueprint.subtasks.all():
                        set_subtask_state_following_allowed_transitions(subtask, 'scheduled')

            # wait until scheduling unit is scheduled
            if not sync_event_bp_scheduled.wait(timeout=10):
                logging.info("sync_event_bp_scheduled event not received, raising TimeoutError")
                raise TimeoutError()
            else:
                logging.info("Received sync_event_bp_scheduled event")
                poll_starttime = datetime.utcnow()
                while True:
                    if SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name == "wait_processed":
                        break
                    sleep(0.1)
                    if datetime.utcnow()-poll_starttime > timedelta(seconds=10):
                        raise TimeoutError("Task not activated within expected time")

            sync_event_bp_scheduled.clear()

           
            #check the active task name
            self.assertEqual("wait_processed", SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name)
                
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].flow_task.name, 'wait_scheduled')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].flow_task.name, 'wait_processed')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].status, 'NEW')

            #Change subtask status to finished
            from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions
            for task_blueprint in scheduling_unit_blueprint.task_blueprints.all():
                task_blueprint.output_pinned=True
                task_blueprint.save()
                
                for subtask in task_blueprint.subtasks.all():
                    set_subtask_state_following_allowed_transitions(subtask, 'finished')

            if not sync_event_bp_cannot_proceed.wait(timeout=10):
                logging.info("sync_event_bp_cannot_proceed event not received, raising TimeoutError")
                raise TimeoutError()
            else:
                logging.info("Received sync_event_bp_cannot_proceed event")
                poll_starttime = datetime.utcnow()
                while True:
                    if SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name == "qa_reporting_to":
                        break
                    sleep(0.1)
                    if datetime.utcnow()-poll_starttime > timedelta(seconds=10):
                        raise TimeoutError("Task not activated within expected time")

            sync_event_bp_cannot_proceed.clear()

            #check the active task name
            self.assertEqual("qa_reporting_to", SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].flow_task.name, 'wait_processed')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].flow_task.name, 'qa_reporting_to')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].status, 'NEW')

            #API: Get current task
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            task_id=r_dict[0]['pk']

            #API: Assign to an user
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].flow_task.name, 'qa_reporting_to')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].status, 'ASSIGNED')
           
            #API: Perform qa_reporting_to step
            headers = {'content-type': 'application/json'}
            data = '{"operator_report": "Test report", "operator_accept": true}'
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            self.assertEqual(200,response.status_code)
           
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].flow_task.name, 'qa_reporting_to')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[4].flow_task.name, 'check_operator_accept')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[4].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].flow_task.name, 'qa_reporting_sos')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].status, 'NEW')
            
            #API: Get current task
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            task_id=r_dict[0]['pk']

            #API: Assign to an user
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            #API: Perform qa_reporting_sos step
            headers = {'content-type': 'application/json'}
            data = '{"sos_report": "Test report", "quality_within_policy": true, "sos_accept_show_pi": true}'
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].flow_task.name, 'qa_reporting_sos')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].status, 'DONE')
           
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[6].flow_task.name, 'check_sos_accept_show_pi')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[6].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].flow_task.name, 'pi_verification')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].status, 'NEW')

            #API: Get current task
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            task_id=r_dict[0]['pk']

            #API: Assign to an user
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)
      
            #API: Perform pi_verification step
            headers = {'content-type': 'application/json'}
            data = '{"pi_report": "Test report", "pi_accept": true}'
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].flow_task.name, 'pi_verification')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[8].flow_task.name, 'decide_acceptance')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[8].status, 'NEW')

            #API: Get current task
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            task_id=r_dict[0]['pk']

            #API: Assign to an user
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            #API: Perform decide_acceptance step
            headers = {'content-type': 'application/json'}
            data = '{"sos_accept_after_pi": false}'
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[8].flow_task.name, 'decide_acceptance')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[8].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[9].flow_task.name, 'check_sos_accept_after_pi')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[9].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[10].flow_task.name, 'mark_sub')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[10].status, 'DONE')
            
            output_pinned = models.SchedulingUnitBlueprint.objects.get(pk=scheduling_unit_process_id).output_pinned
            self.assertEqual(True,output_pinned)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[11].flow_task.name, 'check_data_pinned')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[11].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[12].flow_task.name, 'unpin_data')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[12].status, 'NEW')

            #API: Get current task
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            task_id=r_dict[0]['pk']

            #API: Assign to an user
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            #API: Perform decide_acceptance step
            headers = {'content-type': 'application/json'}
            data = '{"unpin_data": true}'
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[12].flow_task.name, 'unpin_data')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[12].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[13].flow_task.name, 'delete_data')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[13].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[14].flow_task.name, 'end')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[14].status, 'DONE')         
 
    def test_qa_workflow_qa_is_data_pinned_no(self):
        from lofar.sas.tmss.tmss.workflowapp.flows.schedulingunitflow import SchedulingUnitFlow

        from lofar.sas.tmss.tmss.tmssapp import models
        from lofar.sas.tmss.tmss.tmssapp.models import TaskType

        from lofar.sas.tmss.tmss.tmssapp.tasks import create_task_blueprints_and_subtasks_from_scheduling_unit_draft
        from lofar.sas.tmss.test.tmss_test_data_django_models import SchedulingSet_test_data

        from lofar.sas.tmss.tmss.workflowapp.models.schedulingunitflow import SchedulingUnitProcess
        from viewflow.models import Task
        
        sync_event_bp_scheduled = Event()
        sync_event_bp_cannot_proceed = Event()


        class TestSchedulingUnitEventMessageHandler(SchedulingUnitEventMessageHandler):
            def onSchedulingUnitBlueprintStatusChanged(self, id: int, status: str):
                super().onSchedulingUnitBlueprintStatusChanged(id=id, status=status)
               
               
                if status == "scheduled":
                    logging.info("Status is %s, sending sync event",status)
                    sync_event_bp_scheduled.set()

            def onSchedulingUnitBlueprintCannotProceed(self, id: int):
                super().onSchedulingUnitBlueprintCannotProceed(id=id)
                logging.info("Scheduling Unit Blueprint with id %s cannot proceed, sending sync event",id)
                sync_event_bp_cannot_proceed.set()

        service = create_workflow_service(handler_type=TestSchedulingUnitEventMessageHandler,
                                          exchange=self.tmp_exchange.address)
        with BusListenerJanitor(service):
            strategy_template = models.SchedulingUnitObservingStrategyTemplate.objects.get(name="UC1 CTC+pipelines")

            scheduling_unit_draft = models.SchedulingUnitDraft.objects.create(
                                            name="Test Scheduling Unit UC1",
                                            requirements_doc=strategy_template.template,
                                            requirements_template=strategy_template.scheduling_unit_template,
                                            observation_strategy_template=strategy_template,
                                            copy_reason=models.CopyReason.objects.get(value='template'),
                                            generator_instance_doc="para",
                                            copies=None,
                                            scheduling_set=models.SchedulingSet.objects.create(**SchedulingSet_test_data()))
            
            scheduling_unit_blueprint = create_task_blueprints_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
            #ingest_subtask = models.Subtask.objects.get(task_blueprints__scheduling_unit_blueprint__id=scheduling_unit_blueprint.id,
            #    specifications_template__type__value=TaskType.Choices.INGEST.value)
            scheduling_unit_draft.refresh_from_db()
            
            # there is no signal that SchedulingUnitProcess instance was created,
            # so we have to wait an poll before we can proceed with the test
            poll_starttime = datetime.utcnow()
            while True:
                if SchedulingUnitProcess.objects.filter(su=scheduling_unit_blueprint).count() > 0:
                    break
                sleep(0.1)
                if datetime.utcnow()-poll_starttime > timedelta(seconds=10):
                    raise TimeoutError("SchedulingUnitProcess not created within expected time")

            # Yes! the SchedulingUnitProcess was created, let's get it.
            scheduling_unit_process_id = SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).id
            
            prev_ingest_permission_granted_since = models.SchedulingUnitBlueprint.objects.get(pk=scheduling_unit_process_id).ingest_permission_granted_since
            self.assertEqual(prev_ingest_permission_granted_since, None)

            #check the active task name
            self.assertEqual("wait_scheduled", SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name)
                
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[0].flow_task.name, 'start')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[0].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].flow_task.name, 'wait_scheduled')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].status, 'NEW')


            #Change subtask status to scheduled
            from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions
            for task_blueprint in scheduling_unit_blueprint.task_blueprints.all():
                if task_blueprint.specifications_template.type.value != TaskType.Choices.INGEST.value:
                    for subtask in task_blueprint.subtasks.all():
                        set_subtask_state_following_allowed_transitions(subtask, 'scheduled')

            # wait until scheduling unit is scheduled
            if not sync_event_bp_scheduled.wait(timeout=10):
                logging.info("sync_event_bp_scheduled event not received, raising TimeoutError")
                raise TimeoutError()
            else:
                logging.info("Received sync_event_bp_scheduled event")
                poll_starttime = datetime.utcnow()
                while True:
                    if SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name == "wait_processed":
                        break
                    sleep(0.1)
                    if datetime.utcnow()-poll_starttime > timedelta(seconds=10):
                        raise TimeoutError("Task not activated within expected time")

            sync_event_bp_scheduled.clear()

           
            #check the active task name
            self.assertEqual("wait_processed", SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name)
                
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].flow_task.name, 'wait_scheduled')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].flow_task.name, 'wait_processed')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].status, 'NEW')

            #Change subtask status to finished
            from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions
            for task_blueprint in scheduling_unit_blueprint.task_blueprints.all():
                task_blueprint.output_pinned=False
                task_blueprint.save()
                
                for subtask in task_blueprint.subtasks.all():
                    set_subtask_state_following_allowed_transitions(subtask, 'finished')

            if not sync_event_bp_cannot_proceed.wait(timeout=10):
                logging.info("sync_event_bp_cannot_proceed event not received, raising TimeoutError")
                raise TimeoutError()
            else:
                logging.info("Received sync_event_bp_cannot_proceed event")
                poll_starttime = datetime.utcnow()
                while True:
                    if SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name == "qa_reporting_to":
                        break
                    sleep(0.1)
                    if datetime.utcnow()-poll_starttime > timedelta(seconds=10):
                        raise TimeoutError("Task not activated within expected time")

            sync_event_bp_cannot_proceed.clear()

            #check the active task name
            self.assertEqual("qa_reporting_to", SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].flow_task.name, 'wait_processed')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].flow_task.name, 'qa_reporting_to')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].status, 'NEW')

            #API: Get current task
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            task_id=r_dict[0]['pk']

            #API: Assign to an user
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].flow_task.name, 'qa_reporting_to')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].status, 'ASSIGNED')
           
            #API: Perform qa_reporting_to step
            headers = {'content-type': 'application/json'}
            data = '{"operator_report": "Test report", "operator_accept": true}'
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            self.assertEqual(200,response.status_code)
           
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].flow_task.name, 'qa_reporting_to')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[4].flow_task.name, 'check_operator_accept')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[4].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].flow_task.name, 'qa_reporting_sos')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].status, 'NEW')
            
            #API: Get current task
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            task_id=r_dict[0]['pk']

            #API: Assign to an user
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            #API: Perform qa_reporting_sos step
            headers = {'content-type': 'application/json'}
            data = '{"sos_report": "Test report", "quality_within_policy": true, "sos_accept_show_pi": true}'
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].flow_task.name, 'qa_reporting_sos')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].status, 'DONE')
           
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[6].flow_task.name, 'check_sos_accept_show_pi')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[6].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].flow_task.name, 'pi_verification')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].status, 'NEW')

            #API: Get current task
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            task_id=r_dict[0]['pk']

            #API: Assign to an user
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)
      
            #API: Perform pi_verification step
            headers = {'content-type': 'application/json'}
            data = '{"pi_report": "Test report", "pi_accept": true}'
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].flow_task.name, 'pi_verification')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[8].flow_task.name, 'decide_acceptance')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[8].status, 'NEW')

            #API: Get current task
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            task_id=r_dict[0]['pk']

            #API: Assign to an user
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            #API: Perform decide_acceptance step
            headers = {'content-type': 'application/json'}
            data = '{"sos_accept_after_pi": true}'
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[8].flow_task.name, 'decide_acceptance')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[8].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[9].flow_task.name, 'check_sos_accept_after_pi')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[9].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[10].flow_task.name, 'allow_ingest')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[10].status, 'DONE')
            
            #verify that ingest_permission_granted_since is now a valid datetime
            ingest_permission_granted_since = models.SchedulingUnitBlueprint.objects.get(pk=scheduling_unit_process_id).ingest_permission_granted_since
            self.assertEqual(True,isinstance(ingest_permission_granted_since, datetime))
            
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[11].flow_task.name, 'ingest_done')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[11].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[12].flow_task.name, 'mark_sub')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[12].status, 'DONE')
            
            output_pinned = models.SchedulingUnitBlueprint.objects.get(pk=scheduling_unit_process_id).output_pinned
            self.assertEqual(True,output_pinned)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[13].flow_task.name, 'check_data_pinned')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[13].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[14].flow_task.name, 'unpin_data')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[14].status, 'NEW')

            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[15].flow_task.name, 'end')
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[15].status, 'DONE')
    

if __name__ == '__main__':
    #run the unit tests
    unittest.main()
