#!/usr/bin/env python3

# Copyright (C) 2018    ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

# $Id:  $


# This functional test talks to the API like a regular user would.
# It is supposed to cover all REST http methods for all ViewSets.
# I am still a bit under the impression that we re-test Django functionality that we can expect to just work
# with some of these tests. On the other hand a lot of these provide us a nice basis for differentiating out
# behavior in a controlled way.
# We should probably also fully test behavior wrt mandatory and nullable fields.

import unittest
import logging

logger = logging.getLogger(__name__)
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

from lofar.common.test_utils import skip_integration_tests
if skip_integration_tests():
    exit(3)

# Do Mandatory setup step:
# use setup/teardown magic for tmss test database, ldap server and django server
# (ignore pycharm unused import statement, python unittests does use at RunTime the tmss_test_environment_unittest_setup module)
from lofar.sas.tmss.test.tmss_test_environment_unittest_setup import *
# ---
from lofar.sas.tmss.test.test_environment import TMSSTestEnvironment
tmss_test_env = TMSSTestEnvironment(populate_schemas=True, populate_test_data=False, start_ra_test_environment=True,
                                    start_postgres_listener=False, start_subtask_scheduler=False, start_dynamic_scheduler=False,
                                    enable_viewflow=False)

try:
    tmss_test_env.start()
except:
    tmss_test_env.stop()
    exit(1)

# tell unittest to stop (and automagically cleanup) the test database once all testing is done.
def tearDownModule():
    tmss_test_env.stop()
# ---

from lofar.sas.tmss.test.tmss_test_data_django_models import *

# import and setup test data creator
from lofar.sas.tmss.test.tmss_test_data_rest import TMSSRESTTestDataCreator
test_data_creator = TMSSRESTTestDataCreator(BASE_URL, AUTH)

from lofar.sas.tmss.tmss.tmssapp.viewsets.permissions import TMSSPermissions
from lofar.sas.tmss.tmss.tmssapp.viewsets.scheduling import SubtaskViewSet
from django.contrib.auth.models import User, Group
from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions, Subtask

class SystemPermissionTestCase(unittest.TestCase):
    '''
    Tests for system permissons
    '''

    @classmethod
    def create_subtask(cls) -> int:
        # Create preparatory data
        with tmss_test_env.create_tmss_client() as client:
            test_data_creator.wipe_cache()

            cluster_url = client.get_path_as_json_object('/cluster/1')['url']

            # setup: first create an observation, so the pipeline can have input.
            obs_task_blueprint_data = test_data_creator.TaskBlueprint(
                template_url=client.get_task_template(name="target observation")['url'])
            obs_task_blueprint = test_data_creator.post_data_and_get_response_as_json_object(obs_task_blueprint_data,
                                                                                             '/task_blueprint/')
            obs_subtask_template = client.get_subtask_template("observation control")
            obs_spec = get_default_json_object_for_schema(obs_subtask_template['schema'])
            obs_spec['stations']['digital_pointings'][0]['subbands'] = [0]

            obs_subtask_data = test_data_creator.Subtask(specifications_template_url=obs_subtask_template['url'],
                                                         specifications_doc=obs_spec,
                                                         cluster_url=cluster_url,
                                                         task_blueprint_urls=[obs_task_blueprint['url']],
                                                         raw_feedback='Observation.Correlator.channelWidth=3051.7578125')
            obs_subtask = test_data_creator.post_data_and_get_response_as_json_object(obs_subtask_data, '/subtask/')
            obs_subtask_output_url = test_data_creator.post_data_and_get_url(test_data_creator.SubtaskOutput(subtask_url=obs_subtask['url'],
                                                                                                             task_blueprint_url=obs_task_blueprint['url']), '/subtask_output/')
            test_data_creator.post_data_and_get_url(test_data_creator.Dataproduct(filename="L%s_SB000.MS" % obs_subtask['id'],
                                                                                  subtask_output_url=obs_subtask_output_url), '/dataproduct/')
            return obs_subtask['id']

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()

        cls.obs_subtask_id = cls.create_subtask()

        # Create test_data_creator as regular user
        cls.test_data_creator = TMSSRESTTestDataCreator(BASE_URL, requests.auth.HTTPBasicAuth('paulus', 'pauluspass'))
        response = requests.get(cls.test_data_creator.django_api_url + '/', auth=cls.test_data_creator.auth)

        # Populate permissions
        tmss_test_env.populate_permissions()

        # Retrieve TO observer system role
        cls.to_observer_group = Group.objects.get(name='TO observer')


    def test_SubtaskViewSet_has_TMSSPermissions_in_permission_classes(self):
        # Assert SubtaskViewSet has the TMSSPermissions permission class.
        is_in_permission_classes = False
        for p in SubtaskViewSet.get_permissions(SubtaskViewSet):
            # Check TMSSPermission is included in the SubtaskViewSet permission classes
            if not is_in_permission_classes:
                if type(p).__name__ is TMSSPermissions.__name__:
                    # Check that it is actually an instance of TMSSPermissions
                    self.assertTrue(isinstance(p, TMSSPermissions))
                    is_in_permission_classes = True

        self.assertTrue(is_in_permission_classes,
                        msg='SubtaskViewSet does not have %s in its permission classes.' % TMSSPermissions.__name__)


    def test_Subtask_cannot_get_progress_without_to_observer_group(self):
        user = User.objects.get(username='paulus')
        user.groups.set([])

        # refresh user to update cache, see: https://docs.djangoproject.com/en/3.0/topics/auth/default/#permission-caching
        user = User.objects.get(username='paulus')
        while user.has_perm('tmssapp.get_progress_subtask'):
            user = User.objects.get(username='paulus')

        # Assert TO observer group has the get_progress_subtask permission
        self.assertIsNotNone(self.to_observer_group.permissions.all().filter(codename='get_progress_subtask'))
        # Assert Paulus does not have the get_progress_subtask permission
        self.assertFalse(user.has_perm('tmssapp.get_progress_subtask'))

        # Try to get_progress subtask and assert Paulus can't do it without the TO observer group permissions.
        response = GET_and_assert_equal_expected_code(self, BASE_URL + '/subtask/%s/get_progress/' % self.obs_subtask_id, 403,
                                                      auth=self.test_data_creator.auth)


    def test_Subtask_can_get_progress_with_to_observer_group(self):
        user = User.objects.get(username='paulus')
        user.groups.set([self.to_observer_group])

        # refresh user to update cache, see: https://docs.djangoproject.com/en/3.0/topics/auth/default/#permission-caching
        user = User.objects.get(username='paulus')
        while not user.has_perm('tmssapp.get_progress_subtask'):
            user = User.objects.get(username='paulus')

        # Assert TO observer group has the get_progress_subtask permission
        self.assertIsNotNone(self.to_observer_group.permissions.all().filter(codename='get_progress_subtask'))
        # Assert Paulus has the get_progress_subtask permission
        self.assertTrue(user.has_perm('tmssapp.get_progress_subtask'))

        # Try to get_progress subtask and assert Paulus can do it within the TO observer group permissions.
        response = GET_and_assert_equal_expected_code(self, BASE_URL + '/subtask/%s/get_progress/' % self.obs_subtask_id,
                                                      200,
                                                      auth=self.test_data_creator.auth)


    def test_Subtask_cannot_input_dataproducts_without_to_observer_group(self):
        user = User.objects.get(username='paulus')
        user.groups.set([])

        # refresh user to update cache, see: https://docs.djangoproject.com/en/3.0/topics/auth/default/#permission-caching
        user = User.objects.get(username='paulus')
        while user.has_perm('tmssapp.input_dataproducts_subtask'):
            user = User.objects.get(username='paulus')

        # Assert TO observer group has the input_dataproducts_subtask permission
        self.assertIsNotNone(self.to_observer_group.permissions.all().filter(codename='input_dataproducts_subtask'))
        # Assert Paulus does not have the input_dataproducts_subtask permission
        self.assertFalse(user.has_perm('tmssapp.input_dataproducts_subtask'))

        # Try to input_dataproducts subtask and assert Paulus can't do it without the TO observer group permissions.
        response = GET_and_assert_equal_expected_code(self, BASE_URL + '/subtask/%s/input_dataproducts/' % self.obs_subtask_id,
                                                      403,
                                                      auth=self.test_data_creator.auth)


    def test_Subtask_can_input_dataproducts_with_to_observer_group(self):
        user = User.objects.get(username='paulus')
        user.groups.set([self.to_observer_group])

        # refresh user to update cache, see: https://docs.djangoproject.com/en/3.0/topics/auth/default/#permission-caching
        user = User.objects.get(username='paulus')
        while not user.has_perm('tmssapp.input_dataproducts_subtask'):
            user = User.objects.get(username='paulus')

        # Assert TO observer group has the input_dataproducts_subtask permission
        self.assertIsNotNone(self.to_observer_group.permissions.all().filter(codename='input_dataproducts_subtask'))
        # Assert Paulus has the input_dataproducts_subtask permission
        self.assertTrue(user.has_perm('tmssapp.input_dataproducts_subtask'))

        # Try to input_dataproducts subtask and assert Paulus can do it within the TO observer group permissions.
        response = GET_and_assert_equal_expected_code(self, BASE_URL + '/subtask/%s/input_dataproducts/' % self.obs_subtask_id,
                                                      200,
                                                      auth=self.test_data_creator.auth)


    def test_Subtask_cannot_output_dataproducts_without_to_observer_group(self):
        user = User.objects.get(username='paulus')
        user.groups.set([])

        # refresh user to update cache, see: https://docs.djangoproject.com/en/3.0/topics/auth/default/#permission-caching
        user = User.objects.get(username='paulus')
        while user.has_perm('tmssapp.output_dataproducts_subtask'):
            user = User.objects.get(username='paulus')

        # Assert TO observer group has the output_dataproducts_subtask permission
        self.assertIsNotNone(self.to_observer_group.permissions.all().filter(codename='output_dataproducts_subtask'))
        # Assert Paulus does not have the output_dataproducts_subtask permission
        self.assertFalse(user.has_perm('tmssapp.output_dataproducts_subtask'))

        # Try to output_dataproducts subtask and assert Paulus can't do it without the TO observer group permissions.
        response = GET_and_assert_equal_expected_code(self, BASE_URL + '/subtask/%s/output_dataproducts/' % self.obs_subtask_id,
                                                      403,
                                                      auth=self.test_data_creator.auth)


    def test_Subtask_can_output_dataproducts_with_to_observer_group(self):
        user = User.objects.get(username='paulus')
        user.groups.set([self.to_observer_group])

        # refresh user to update cache, see: https://docs.djangoproject.com/en/3.0/topics/auth/default/#permission-caching
        user = User.objects.get(username='paulus')
        while not user.has_perm('tmssapp.output_dataproducts_subtask'):
            user = User.objects.get(username='paulus')

        # Assert TO observer group has the output_dataproducts_subtask permission
        self.assertIsNotNone(self.to_observer_group.permissions.all().filter(codename='output_dataproducts_subtask'))
        # Assert Paulus has the output_dataproducts_subtask permission
        self.assertTrue(user.has_perm('tmssapp.output_dataproducts_subtask'))

        # Try to output_dataproducts subtask and assert Paulus can do it within the TO observer group permissions.
        response = GET_and_assert_equal_expected_code(self, BASE_URL + '/subtask/%s/output_dataproducts/' % self.obs_subtask_id,
                                                      200,
                                                      auth=self.test_data_creator.auth)


    def test_Subtask_cannot_parset_without_to_observer_group(self):
        user = User.objects.get(username='paulus')
        user.groups.set([])

        # refresh user to update cache, see: https://docs.djangoproject.com/en/3.0/topics/auth/default/#permission-caching
        user = User.objects.get(username='paulus')
        while user.has_perm('tmssapp.parset_subtask'):
            user = User.objects.get(username='paulus')

        # Assert TO observer group has the parset_subtask permission
        self.assertIsNotNone(self.to_observer_group.permissions.all().filter(codename='parset_subtask'))
        # Assert Paulus does not have the parset_subtask permission
        self.assertFalse(user.has_perm('tmssapp.parset_subtask'))

        # Try to parset subtask and assert Paulus can't do it without the TO observer group permissions.
        response = GET_and_assert_equal_expected_code(self, BASE_URL + '/subtask/%s/parset/' % self.obs_subtask_id,
                                                      403,
                                                      auth=self.test_data_creator.auth)


    def test_Subtask_can_parset_with_to_observer_group(self):
        user = User.objects.get(username='paulus')
        user.groups.set([self.to_observer_group])

        # refresh user to update cache, see: https://docs.djangoproject.com/en/3.0/topics/auth/default/#permission-caching
        user = User.objects.get(username='paulus')
        while not user.has_perm('tmssapp.parset_subtask'):
            user = User.objects.get(username='paulus')

        # Assert TO observer group has the parset_subtask permission
        self.assertIsNotNone(self.to_observer_group.permissions.all().filter(codename='parset_subtask'))
        # Assert Paulus has the parset_subtask permission
        self.assertTrue(user.has_perm('tmssapp.parset_subtask'))

        # Try to parset subtask and assert Paulus can do it within the TO observer group permissions.
        response = GET_and_assert_equal_expected_code(self, BASE_URL + '/subtask/%s/parset/' % self.obs_subtask_id,
                                                      200,
                                                      auth=self.test_data_creator.auth)


    def test_Subtask_cannot_predecessors_without_to_observer_group(self):
        user = User.objects.get(username='paulus')
        user.groups.set([])

        # refresh user to update cache, see: https://docs.djangoproject.com/en/3.0/topics/auth/default/#permission-caching
        user = User.objects.get(username='paulus')
        while user.has_perm('tmssapp.predecessors_subtask'):
            user = User.objects.get(username='paulus')

        # Assert TO observer group has the predecessors_subtask permission
        self.assertIsNotNone(self.to_observer_group.permissions.all().filter(codename='predecessors_subtask'))
        # Assert Paulus does not have the predecessors_subtask permission
        self.assertFalse(user.has_perm('tmssapp.predecessors_subtask'))

        # Try to predecessors subtask and assert Paulus can't do it without the TO observer group permissions.
        response = GET_and_assert_equal_expected_code(self,
                                                      BASE_URL + '/subtask/%s/predecessors/' % self.obs_subtask_id,
                                                      403,
                                                      auth=self.test_data_creator.auth)


    def test_Subtask_can_predecessors_with_to_observer_group(self):
        user = User.objects.get(username='paulus')
        user.groups.set([self.to_observer_group])

        # refresh user to update cache, see: https://docs.djangoproject.com/en/3.0/topics/auth/default/#permission-caching
        user = User.objects.get(username='paulus')
        while not user.has_perm('tmssapp.predecessors_subtask'):
            user = User.objects.get(username='paulus')

        # Assert TO observer group has the predecessors_subtask permission
        self.assertIsNotNone(self.to_observer_group.permissions.all().filter(codename='predecessors_subtask'))
        # Assert Paulus has the predecessors_subtask permission
        self.assertTrue(user.has_perm('tmssapp.predecessors_subtask'))

        # Try to predecessors subtask and assert Paulus can do it within the TO observer group permissions.
        response = GET_and_assert_equal_expected_code(self, BASE_URL + '/subtask/%s/predecessors/' % self.obs_subtask_id,
                                                      200,
                                                      auth=self.test_data_creator.auth)


    def test_Subtask_cannot_process_feedback_and_set_to_finished_if_complete_without_to_observer_group(self):
        user = User.objects.get(username='paulus')
        user.groups.set([])

        # refresh user to update cache, see: https://docs.djangoproject.com/en/3.0/topics/auth/default/#permission-caching
        user = User.objects.get(username='paulus')
        while user.has_perm('tmssapp.process_feedback_and_set_to_finished_if_complete_subtask'):
            user = User.objects.get(username='paulus')

        # Assert TO observer group has the process_feedback_and_set_to_finished_if_complete_subtask permission
        self.assertIsNotNone(self.to_observer_group.permissions.all().filter(codename='process_feedback_and_set_to_finished_if_complete_subtask'))
        # Assert Paulus does not have the process_feedback_and_set_to_finished_if_complete_subtask permission
        self.assertFalse(user.has_perm('tmssapp.process_feedback_and_set_to_finished_if_complete_subtask'))

        # Try to process_feedback_and_set_to_finished_if_complete subtask and assert Paulus can't do it without the TO observer group permissions.
        response = POST_and_assert_expected_response(self, BASE_URL + '/subtask/%s/process_feedback_and_set_to_finished_if_complete/' % self.obs_subtask_id,
                                                     {}, 403, {},
                                                     auth=self.test_data_creator.auth)


    def test_Subtask_can_process_feedback_and_set_to_finished_if_complete_with_to_observer_group(self):
        user = User.objects.get(username='paulus')
        user.groups.set([self.to_observer_group])

        # refresh user to update cache, see: https://docs.djangoproject.com/en/3.0/topics/auth/default/#permission-caching
        user = User.objects.get(username='paulus')
        while not user.has_perm('tmssapp.process_feedback_and_set_to_finished_if_complete_subtask'):
            user = User.objects.get(username='paulus')

        # Assert TO observer group has the process_feedback_and_set_to_finished_if_complete_subtask permission
        self.assertIsNotNone(self.to_observer_group.permissions.all().filter(codename='process_feedback_and_set_to_finished_if_complete_subtask'))
        # Assert Paulus has the process_feedback_and_set_to_finished_if_complete_subtask permission
        self.assertTrue(user.has_perm('tmssapp.process_feedback_and_set_to_finished_if_complete_subtask'))

        obs_subtask_id = self.create_subtask()
        # Set subtask status to finishing, so it can process feedback and set to finished.
        set_subtask_state_following_allowed_transitions(Subtask.objects.get(id=obs_subtask_id), 'finishing')

        # Try to process_feedback_and_set_to_finished_if_complete subtask and assert Paulus can do it within the TO observer group permissions.
        response = POST_and_assert_expected_response(self, BASE_URL + '/subtask/%s/process_feedback_and_set_to_finished_if_complete/' % obs_subtask_id,
                                                 {}, 200, {},
                                                 auth=self.test_data_creator.auth)


    def test_Subtask_cannot_reprocess_raw_feedback_for_subtask_and_set_to_finished_if_complete_without_to_observer_group(
            self):
        user = User.objects.get(username='paulus')
        user.groups.set([])

        # refresh user to update cache, see: https://docs.djangoproject.com/en/3.0/topics/auth/default/#permission-caching
        user = User.objects.get(username='paulus')
        while user.has_perm('tmssapp.reprocess_raw_feedback_for_subtask_and_set_to_finished_if_complete_subtask'):
            user = User.objects.get(username='paulus')

        # Assert TO observer group has the reprocess_raw_feedback_for_subtask_and_set_to_finished_if_complete_subtask permission
        self.assertIsNotNone(self.to_observer_group.permissions.all().filter(codename='reprocess_raw_feedback_for_subtask_and_set_to_finished_if_complete_subtask'))
        # Assert Paulus does not have the reprocess_raw_feedback_for_subtask_and_set_to_finished_if_complete_subtask permission
        self.assertFalse(user.has_perm('tmssapp.reprocess_raw_feedback_for_subtask_and_set_to_finished_if_complete_subtask'))

        # Try to reprocess_raw_feedback_for_subtask_and_set_to_finished_if_complete subtask and assert Paulus can't do it without the TO observer group permissions.
        response = GET_and_assert_equal_expected_code(self, BASE_URL + '/subtask/%s/reprocess_raw_feedback_for_subtask_and_set_to_finished_if_complete/' % self.obs_subtask_id,
                                                      403,
                                                      auth=self.test_data_creator.auth)


    def test_Subtask_can_reprocess_raw_feedback_for_subtask_and_set_to_finished_if_complete_with_to_observer_group(self):
        user = User.objects.get(username='paulus')
        user.groups.set([self.to_observer_group])

        # refresh user to update cache, see: https://docs.djangoproject.com/en/3.0/topics/auth/default/#permission-caching
        user = User.objects.get(username='paulus')
        while not user.has_perm('tmssapp.reprocess_raw_feedback_for_subtask_and_set_to_finished_if_complete_subtask'):
            user = User.objects.get(username='paulus')

        # Assert TO observer group has the reprocess_raw_feedback_for_subtask_and_set_to_finished_if_complete_subtask permission
        self.assertIsNotNone(self.to_observer_group.permissions.all().filter(codename='reprocess_raw_feedback_for_subtask_and_set_to_finished_if_complete_subtask'))
        # Assert Paulus has the reprocess_raw_feedback_for_subtask_and_set_to_finished_if_complete_subtask permission
        self.assertTrue(user.has_perm('tmssapp.reprocess_raw_feedback_for_subtask_and_set_to_finished_if_complete_subtask'))

        obs_subtask_id = self.create_subtask()
        # Set subtask status to finishing, so we can reprocess feedback
        set_subtask_state_following_allowed_transitions(Subtask.objects.get(id=obs_subtask_id), 'finishing')

        # Try to reprocess_raw_feedback_for_subtask_and_set_to_finished_if_complete subtask and assert Paulus can do it within the TO observer group permissions.
        response = GET_and_assert_equal_expected_code(self, BASE_URL + '/subtask/%s/reprocess_raw_feedback_for_subtask_and_set_to_finished_if_complete/' % obs_subtask_id,
                                                      200,
                                                      auth=self.test_data_creator.auth)


    def test_Subtask_cannot_schedule_without_to_observer_group(self):
        user = User.objects.get(username='paulus')
        user.groups.set([])

        # refresh user to update cache, see: https://docs.djangoproject.com/en/3.0/topics/auth/default/#permission-caching
        user = User.objects.get(username='paulus')
        while user.has_perm('tmssapp.schedule_subtask'):
            user = User.objects.get(username='paulus')

        # Assert TO observer group has the schedule_subtask permission
        self.assertIsNotNone(self.to_observer_group.permissions.all().filter(codename='schedule_subtask'))
        # Assert Paulus does not have the schedule_subtask permission
        self.assertFalse(user.has_perm('tmssapp.schedule_subtask'))

        # Try to schedule subtask and assert Paulus can't do it without the TO observer group permissions.
        response = POST_and_assert_expected_response(self,
                                                     BASE_URL + '/subtask/%s/schedule/' % self.obs_subtask_id,
                                                     {},
                                                     403,
                                                     None,
                                                     auth=self.test_data_creator.auth)


    def test_Subtask_can_schedule_with_to_observer_group(self):
        user = User.objects.get(username='paulus')
        user.groups.set([self.to_observer_group])

        # refresh user to update cache, see: https://docs.djangoproject.com/en/3.0/topics/auth/default/#permission-caching
        user = User.objects.get(username='paulus')
        while not user.has_perm('tmssapp.schedule_subtask'):
            user = User.objects.get(username='paulus')

        # Assert TO observer group has the schedule_subtask permission
        self.assertIsNotNone(self.to_observer_group.permissions.all().filter(codename='schedule_subtask'))
        # Assert Paulus has the schedule_subtask permission
        self.assertTrue(user.has_perm('tmssapp.schedule_subtask'))

        obs_subtask_id = self.create_subtask()
        # Set subtask status to defined, so it can be scheduled.
        set_subtask_state_following_allowed_transitions(Subtask.objects.get(id=obs_subtask_id), 'defined')

        # Try to schedule subtask and assert Paulus can do it within the TO observer group permissions.
        response = POST_and_assert_expected_response(self, BASE_URL + '/subtask/%s/schedule/' % obs_subtask_id,
                                                     {},
                                                     200,
                                                     None,
                                                     auth=self.test_data_creator.auth)


    def test_Subtask_cannot_state_log_without_to_observer_group(self):
        user = User.objects.get(username='paulus')
        user.groups.set([])

        # refresh user to update cache, see: https://docs.djangoproject.com/en/3.0/topics/auth/default/#permission-caching
        user = User.objects.get(username='paulus')
        while user.has_perm('tmssapp.state_log_subtask'):
            user = User.objects.get(username='paulus')

        # Assert TO observer group has the state_log_subtask permission
        self.assertIsNotNone(self.to_observer_group.permissions.all().filter(codename='state_log_subtask'))
        # Assert Paulus does not have the state_log_subtask permission
        self.assertFalse(user.has_perm('tmssapp.state_log_subtask'))

        # Try to state_log subtask and assert Paulus can't do it without the TO observer group permissions.
        response = GET_and_assert_equal_expected_code(self, BASE_URL + '/subtask/%s/state_log/' % self.obs_subtask_id,
                                                      403,
                                                      auth=self.test_data_creator.auth)


    def test_Subtask_can_state_log_with_to_observer_group(self):
        user = User.objects.get(username='paulus')
        user.groups.set([self.to_observer_group])

        # refresh user to update cache, see: https://docs.djangoproject.com/en/3.0/topics/auth/default/#permission-caching
        user = User.objects.get(username='paulus')
        while not user.has_perm('tmssapp.state_log_subtask'):
            user = User.objects.get(username='paulus')

        # Assert TO observer group has the state_log_subtask permission
        self.assertIsNotNone(self.to_observer_group.permissions.all().filter(codename='state_log_subtask'))
        # Assert Paulus has the state_log_subtask permission
        self.assertTrue(user.has_perm('tmssapp.state_log_subtask'))

        # Try to state_log subtask and assert Paulus can do it within the TO observer group permissions.
        response = GET_and_assert_equal_expected_code(self, BASE_URL + '/subtask/%s/state_log/' % self.obs_subtask_id,
                                                      200,
                                                      auth=self.test_data_creator.auth)


    def test_Subtask_cannot_successors_without_to_observer_group(self):
        user = User.objects.get(username='paulus')
        user.groups.set([])

        # refresh user to update cache, see: https://docs.djangoproject.com/en/3.0/topics/auth/default/#permission-caching
        user = User.objects.get(username='paulus')
        while user.has_perm('tmssapp.successors_subtask'):
            user = User.objects.get(username='paulus')

        # Assert TO observer group has the successors_subtask permission
        self.assertIsNotNone(self.to_observer_group.permissions.all().filter(codename='successors_subtask'))
        # Assert Paulus does not have the successors_subtask permission
        self.assertFalse(user.has_perm('tmssapp.successors_subtask'))

        # Try to successors subtask and assert Paulus can't do it without the TO observer group permissions.
        response = GET_and_assert_equal_expected_code(self, BASE_URL + '/subtask/%s/successors/' % self.obs_subtask_id,
                                                      403,
                                                      auth=self.test_data_creator.auth)


    def test_Subtask_can_successors_with_to_observer_group(self):
        user = User.objects.get(username='paulus')
        user.groups.set([self.to_observer_group])

        # refresh user to update cache, see: https://docs.djangoproject.com/en/3.0/topics/auth/default/#permission-caching
        user = User.objects.get(username='paulus')
        while not user.has_perm('tmssapp.successors_subtask'):
            user = User.objects.get(username='paulus')

        # Assert TO observer group has the successors_subtask permission
        self.assertIsNotNone(self.to_observer_group.permissions.all().filter(codename='successors_subtask'))
        # Assert Paulus has the successors_subtask permission
        self.assertTrue(user.has_perm('tmssapp.successors_subtask'))

        # Try to successors subtask and assert Paulus can do it within the TO observer group permissions.
        response = GET_and_assert_equal_expected_code(self, BASE_URL + '/subtask/%s/successors/' % self.obs_subtask_id,
                                                      200,
                                                      auth=self.test_data_creator.auth)


    def test_Subtask_cannot_task_log_without_to_observer_group(self):
        user = User.objects.get(username='paulus')
        user.groups.set([])

        # refresh user to update cache, see: https://docs.djangoproject.com/en/3.0/topics/auth/default/#permission-caching
        user = User.objects.get(username='paulus')
        while user.has_perm('tmssapp.task_log_subtask'):
            user = User.objects.get(username='paulus')

        # Assert TO observer group has the task_log_subtask permission
        self.assertIsNotNone(self.to_observer_group.permissions.all().filter(codename='task_log_subtask'))
        # Assert Paulus does not have the task_log_subtask permission
        self.assertFalse(user.has_perm('tmssapp.task_log_subtask'))

        # Try to task_log subtask and assert Paulus can't do it without the TO observer group permissions.
        response = GET_and_assert_equal_expected_code(self, BASE_URL + '/subtask/%s/task_log/' % self.obs_subtask_id,
                                                      403,
                                                      auth=self.test_data_creator.auth)


    def test_Subtask_can_task_log_with_to_observer_group(self):
        user = User.objects.get(username='paulus')
        user.groups.set([self.to_observer_group])

        # refresh user to update cache, see: https://docs.djangoproject.com/en/3.0/topics/auth/default/#permission-caching
        user = User.objects.get(username='paulus')
        while not user.has_perm('tmssapp.task_log_subtask'):
            user = User.objects.get(username='paulus')

        # Assert TO observer group has the task_log_subtask permission
        self.assertIsNotNone(self.to_observer_group.permissions.all().filter(codename='task_log_subtask'))
        # Assert Paulus has the task_log_subtask permission
        self.assertTrue(user.has_perm('tmssapp.task_log_subtask'))

        # Try to task_log subtask and assert Paulus can do it within the TO observer group permissions.
        response = GET_and_assert_equal_expected_code(self, BASE_URL + '/subtask/%s/task_log/' % self.obs_subtask_id,
                                                      200,
                                                      auth=self.test_data_creator.auth)


    def test_Subtask_cannot_transformed_output_dataproduct_without_to_observer_group(self):
        user = User.objects.get(username='paulus')
        user.groups.set([])

        # refresh user to update cache, see: https://docs.djangoproject.com/en/3.0/topics/auth/default/#permission-caching
        user = User.objects.get(username='paulus')
        while user.has_perm('tmssapp.transformed_output_dataproduct_subtask'):
            user = User.objects.get(username='paulus')

        # Assert TO observer group has the transformed_output_dataproduct_subtask permission
        self.assertIsNotNone(self.to_observer_group.permissions.all().filter(codename='transformed_output_dataproduct_subtask'))
        # Assert Paulus does not have the transformed_output_dataproduct_subtask permission
        self.assertFalse(user.has_perm('tmssapp.transformed_output_dataproduct_subtask'))

        # Try to transformed_output_dataproduct subtask and assert Paulus can't do it without the TO observer group permissions.
        response = GET_and_assert_equal_expected_code(self, BASE_URL + '/subtask/%s/transformed_output_dataproduct/' % self.obs_subtask_id,
                                                      403,
                                                      auth=self.test_data_creator.auth)


    def test_Subtask_can_transformed_output_dataproduct_with_to_observer_group(self):
        user = User.objects.get(username='paulus')
        user.groups.set([self.to_observer_group])

        # refresh user to update cache, see: https://docs.djangoproject.com/en/3.0/topics/auth/default/#permission-caching
        user = User.objects.get(username='paulus')
        while not user.has_perm('tmssapp.transformed_output_dataproduct_subtask'):
            user = User.objects.get(username='paulus')

        # Assert TO observer group has the transformed_output_dataproduct_subtask permission
        self.assertIsNotNone(self.to_observer_group.permissions.all().filter(codename='transformed_output_dataproduct_subtask'))
        # Assert Paulus has the transformed_output_dataproduct_subtask permission
        self.assertTrue(user.has_perm('tmssapp.transformed_output_dataproduct_subtask'))

        # Try to transformed_output_dataproduct subtask and assert Paulus can do it within the TO observer group permissions.
        with tmss_test_env.create_tmss_client() as client:
            # NOTE: A 404 Not found error is returned if input_dataproducts are missing. It is not related to user's permissions.
            response = GET_and_assert_equal_expected_code(self, BASE_URL + '/subtask/%s/transformed_output_dataproduct?input_dataproduct_id=%s' % (
                                                          self.obs_subtask_id, 1),
                                                          404,
                                                          auth=self.test_data_creator.auth)


    def test_Subtask_cannot_unschedule_without_to_observer_group(self):
        user = User.objects.get(username='paulus')
        user.groups.set([])

        # refresh user to update cache, see: https://docs.djangoproject.com/en/3.0/topics/auth/default/#permission-caching
        user = User.objects.get(username='paulus')
        while user.has_perm('tmssapp.unschedule_subtask'):
            user = User.objects.get(username='paulus')

        # Assert TO observer group has the unschedule_subtask permission
        self.assertIsNotNone(self.to_observer_group.permissions.all().filter(codename='unschedule_subtask'))
        # Assert Paulus does not have the unschedule_subtask permission
        self.assertFalse(user.has_perm('tmssapp.unschedule_subtask'))

        # Try to unschedule subtask and assert Paulus can't do it without the TO observer group permissions.
        response = POST_and_assert_expected_response(self,
                                                     BASE_URL + '/subtask/%s/unschedule/' % self.obs_subtask_id,
                                                     {},
                                                     403,
                                                     None,
                                                     auth=self.test_data_creator.auth)

    def test_Subtask_can_unschedule_with_to_observer_group(self):
        user = User.objects.get(username='paulus')
        user.groups.set([self.to_observer_group])

        # refresh user to update cache, see: https://docs.djangoproject.com/en/3.0/topics/auth/default/#permission-caching
        user = User.objects.get(username='paulus')
        while not user.has_perm('tmssapp.unschedule_subtask'):
            user = User.objects.get(username='paulus')

        # Assert TO observer group has the unschedule_subtask permission
        self.assertIsNotNone(self.to_observer_group.permissions.all().filter(codename='unschedule_subtask'))
        # Assert Paulus has the unschedule_subtask permission
        self.assertTrue(user.has_perm('tmssapp.unschedule_subtask'))

        obs_subtask_id = self.create_subtask()
        # Set subtask status to scheduled, so it can be unscheduled.
        set_subtask_state_following_allowed_transitions(Subtask.objects.get(id=obs_subtask_id), 'scheduled')

        # Try to unschedule subtask and assert Paulus can do it within the TO observer group permissions.
        response = POST_and_assert_expected_response(self, BASE_URL + '/subtask/%s/unschedule/' % obs_subtask_id,
                                                     {},
                                                     200,
                                                     None,
                                                     auth=self.test_data_creator.auth)


if __name__ == "__main__":
    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                        level=logging.INFO)
    unittest.main()
