#!/usr/bin/env python3

# Copyright (C) 2018    ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

# $Id:  $

import os
import unittest
import requests

import logging
logger = logging.getLogger(__name__)
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

from lofar.common.test_utils import exit_with_skipped_code_if_skip_integration_tests
exit_with_skipped_code_if_skip_integration_tests()

# Do Mandatory setup step:
# use setup/teardown magic for tmss test database, ldap server and django server
# (ignore pycharm unused import statement, python unittests does use at RunTime the tmss_test_environment_unittest_setup module)
from lofar.sas.tmss.test.tmss_test_environment_unittest_setup import *
tmss_test_env.populate_schemas()

from lofar.sas.tmss.test.tmss_test_data_django_models import *

# import and setup rest test data creator
from lofar.sas.tmss.test.tmss_test_data_rest import TMSSRESTTestDataCreator
rest_data_creator = TMSSRESTTestDataCreator(BASE_URL, AUTH)

from lofar.sas.tmss.tmss.tmssapp import models
from lofar.sas.tmss.tmss.exceptions import SchemaValidationException

import requests

class SubtaskValidationTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        # create reusable instances to speed up testing
        cls.task_blueprint = models.TaskBlueprint.objects.create(**TaskBlueprint_test_data())
        cls.cluster = models.Cluster.objects.create(name="dummy cluster", location="downstairs", tags=[])
        cls.state = models.SubtaskState.objects.get(value='defining')

    @staticmethod
    def create_subtask_template(schema):
        subtask_template_data = SubtaskTemplate_test_data(schema=schema)
        return models.SubtaskTemplate.objects.create(**subtask_template_data)

    def test_validate_flawed_json_schema(self):
        with self.assertRaises(SchemaValidationException) as context:
            subtask_template = self.create_subtask_template('{ this is not a json object }')

    def test_validate_flawed_json_specification(self):
        subtask_template = self.create_subtask_template(minimal_json_schema())
        specifications_doc = '{ this is not a json object }'
        subtask_data = Subtask_test_data(subtask_template=subtask_template, specifications_doc=specifications_doc,
                                         cluster=self.cluster, state=self.state)

        with self.assertRaises(SchemaValidationException) as context:
            subtask = models.Subtask.objects.create(**subtask_data)
            subtask.task_blueprints.set([self.task_blueprint])
        self.assertTrue('invalid json' in str(context.exception).lower())

    def test_validate_correlator_schema_with_valid_specification(self):
        # fetch correlator_schema for Dupplo UC1 which should be in the initially populated database
        subtask_template = models.SubtaskTemplate.objects.get(name='observation control')
        self.assertIsNotNone(subtask_template)

        specifications_doc = get_default_json_object_for_schema(subtask_template.schema)
        subtask_data = Subtask_test_data(subtask_template=subtask_template, specifications_doc=specifications_doc,
                                         cluster=self.cluster, state=self.state)

        subtask = models.Subtask.objects.create(**subtask_data)
        self.assertIsNotNone(subtask)

    def test_validate_correlator_schema_with_invalid_specification(self):
        # fetch correlator_schema for Dupplo UC1 which should be in the initially populated database
        subtask_template = models.SubtaskTemplate.objects.get(name='observation control')
        self.assertIsNotNone(subtask_template)

        # test with invalid json
        with self.assertRaises(SchemaValidationException) as context:
            subtask_data = Subtask_test_data(subtask_template=subtask_template, specifications_doc="bogus spec",
                                             cluster=self.cluster, state=self.state)
            subtask = models.Subtask.objects.create(**subtask_data)

        # test with valid json, but not according to schema
        with self.assertRaises(SchemaValidationException) as context:
            specifications_doc = get_default_json_object_for_schema(subtask_template.schema)
            specifications_doc['COBALT']['blocksize'] = -1 # invalid value, should cause the SchemaValidationException
            subtask_data = Subtask_test_data(subtask_template=subtask_template, specifications_doc=specifications_doc,
                                             cluster=self.cluster, state=self.state)
            models.Subtask.objects.create(**subtask_data)
        self.assertTrue('-1 is less than the minimum' in str(context.exception).lower())

    def test_validate_correlator_schema_with_valid_specification_via_rest(self):
        # fetch correlator_schema for Dupplo UC1 which should be in the initially populated database
        response = requests.get(BASE_URL+"/subtask_template/", {"format": "json", "name": "observation control"}, auth=AUTH)
        self.assertEqual(200, response.status_code)
        json_response = response.json()
        self.assertEqual(1, json_response.get('count'))

        template = json_response['results'][0]
        schema_url = template['url'].split('?')[0]
        schema = template['schema']

        specifications_doc = get_default_json_object_for_schema(schema)
        subtask_test_data = rest_data_creator.Subtask(specifications_template_url=schema_url, specifications_doc=specifications_doc)

        # POST and GET a new item and assert correctness
        r_dict = POST_and_assert_expected_response(self, BASE_URL + '/subtask/', subtask_test_data, 201, subtask_test_data)
        url = r_dict['url']
        GET_OK_and_assert_equal_expected_response(self, url, subtask_test_data)

    def test_validate_correlator_schema_with_invalid_specification_via_rest(self):
        # fetch correlator_schema for Dupplo UC1 which should be in the initially populated database
        response = requests.get(BASE_URL+"/subtask_template/", {"format": "json", "name": "observation control"}, auth=AUTH)
        self.assertEqual(200, response.status_code)
        json_response = response.json()
        self.assertEqual(1, json_response.get('count'))

        template = json_response['results'][0]
        schema_url = template['url'].split('?')[0]

        specifications_doc = "bogus spec"
        subtask_test_data = rest_data_creator.Subtask(specifications_template_url=schema_url, specifications_doc=specifications_doc)

        POST_and_assert_expected_response(self, BASE_URL + '/subtask/', subtask_test_data, 500, {})

if __name__ == "__main__":
    os.environ['TZ'] = 'UTC'
    unittest.main()
