#!/usr/bin/env python3

# Copyright (C) 2018    ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

# $Id:  $

import os
import unittest

import logging
logger = logging.getLogger(__name__)
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

from lofar.common.test_utils import exit_with_skipped_code_if_skip_integration_tests
exit_with_skipped_code_if_skip_integration_tests()

# Do Mandatory setup step:
# use setup/teardown magic for tmss test database, ldap server and django server
# (ignore pycharm unused import statement, python unittests does use at RunTime the tmss_test_environment_unittest_setup module)
from lofar.sas.tmss.test.tmss_test_environment_unittest_setup import *
tmss_test_env.populate_schemas()


from lofar.sas.tmss.test.tmss_test_data_django_models import *
from lofar.sas.tmss.tmss.tmssapp import models
from lofar.sas.tmss.tmss.tmssapp.subtasks import *
from lofar.sas.tmss.tmss.tmssapp.subtasks import _get_related_target_sap_by_name, _generate_tab_ring_pointings, _filter_subbands, _add_pointings
from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions


def create_subtask_object_for_testing(subtask_type_value):
    """
    Helper function to create a subtask object for testing with given subtask value and subtask state value
    as string (no object)
    """
    template_type = models.SubtaskType.objects.get(value=subtask_type_value)
    subtask_template_obj = create_subtask_template_for_testing(template_type)
    subtask_data = Subtask_test_data(subtask_template=subtask_template_obj)
    subtask = models.Subtask.objects.create(**subtask_data)
    subtask.task_blueprints.set([models.TaskBlueprint.objects.create(**TaskBlueprint_test_data())])
    return subtask

def create_subtask_template_for_testing(template_type: object):
    """
    Helper function
    :param template_type:
    :return:
    """
    subtask_template_data = SubtaskTemplate_test_data()
    subtask_template_data['type'] = template_type
    return models.SubtaskTemplate.objects.create(**subtask_template_data)


def create_task_blueprint_object_for_testing(task_template_name="target observation", QA_enabled=False):
    """
    Helper function to create a task blueprint object for testing with given task template name value
    as string (no object)
    :param task_template_name: (Optional) name of schema target observation is target observation
    :param QA_enabled: (Optional) QA plots and file_conversion
    :return: task_blueprint_obj: Created Task Blueprint object
    """
    task_template = models.TaskTemplate.objects.get(name=task_template_name)
    task_spec = get_default_json_object_for_schema(task_template.schema)
    if 'QA' in task_spec:
        task_spec["QA"]['plots']['enabled'] = QA_enabled
        task_spec["QA"]['file_conversion']['enabled'] = QA_enabled

    task_draft_data = TaskDraft_test_data(specifications_template=task_template, specifications_doc=task_spec)
    task_draft_obj = models.TaskDraft.objects.create(**task_draft_data)

    task_name = "BlueprintTask with %s" % task_template_name
    task_blueprint_data = TaskBlueprint_test_data(name=task_name, task_draft=task_draft_obj)
    task_blueprint_obj = models.TaskBlueprint.objects.create(**task_blueprint_data)
    return task_blueprint_obj


def create_relation_task_blueprint_object_for_testing(blueprint_task_producer, blueprint_task_consumer):
    """
    Helper function to create a task relation blueprint object for testing for given task objects consumer and producer
    :param blueprint_task_producer: Blueprint task of producer, typical an observation
    :param blueprint_task_consumer: Blueprint task of consumer, typical a preprocessing pipeline
    :return: task_relation_obj: Created Task Relation Blueprint object
    """
    task_relation_data = TaskRelationBlueprint_test_data(blueprint_task_producer, blueprint_task_consumer)
    task_relation_obj = models.TaskRelationBlueprint.objects.create(**task_relation_data)
    return task_relation_obj


def create_scheduling_relation_task_blueprint_for_testing(first_task_blueprint, second_task_blueprint, placement='before'):
    """
    Helper function to create a task blueprint relation object between two task blueprint (calibrator and target observation)
    :param first_task_blueprint:
    :param second_task_blueprint:
    :return: task_relation_blueprint_obj: Created Task Relation Blueprint object
    """
    task_scheduling_rel_obj = models.TaskSchedulingRelationBlueprint.objects.create(
                                 tags=[],
                                 first=first_task_blueprint,
                                 second=second_task_blueprint,
                                 placement=models.SchedulingRelationPlacement.objects.get(value=placement),
                                 time_offset=60)
    return task_scheduling_rel_obj


class SubTasksCreationFromSubTask(unittest.TestCase):

    def test_create_qafile_subtask_from_pipeline_subtask_failed(self):
        """
        Test if creation of subtask qafile failed due to wrong type of the predecessor subtask
        Correct type should be 'observation' (for this test of course it is not)
        """
        pipeline_subtask = create_subtask_object_for_testing("pipeline")
        with self.assertRaises(ValueError):
            create_qafile_subtask_from_observation_subtask(pipeline_subtask)

    def test_create_qafile_subtask_from_observation_subtask_succeed(self):
        """
        Test if creation of subtask qafile succeed
        Subtask object is None because QA file conversion is  by default not enabled!!!!
        """
        predecessor_subtask = create_subtask_object_for_testing("observation")
        subtask = create_qafile_subtask_from_observation_subtask(predecessor_subtask)
        self.assertEqual(None, subtask)

    def test_create_qaplots_subtask_from_qafile_subtask_failed(self):
        """
        Test if creation of subtask qaplots failed due to wrong state or wrong type of the predecessor subtask
        Correct type should be 'qa_files' (for this test of course it is not)
        """
        subtasks = [create_subtask_object_for_testing("pipeline"),
                    create_subtask_object_for_testing("observation"),
                    create_subtask_object_for_testing("observation") ]
        for subtask in subtasks:
            with self.assertRaises(ValueError):
                create_qaplots_subtask_from_qafile_subtask(subtask)

    def test_create_qaplots_subtask_from_qafile_subtask_succeed(self):
        """
        Test if creation of subtask qaplots succeed
        Subtask object is None because QA plots is by default not enabled!!!!
        """
        predecessor_subtask = create_subtask_object_for_testing("qa_files")
        subtask = create_qaplots_subtask_from_qafile_subtask(predecessor_subtask)
        self.assertEqual(None, subtask)


class SubTasksCreationFromTaskBluePrint(unittest.TestCase):

    def test_create_sequence_of_subtask_from_task_blueprint(self):
        """
        Create multiple subtasks from a task blueprint, executed in correct order.
        No exception should occur, check name, type and state of the subtask
        """
        task_blueprint = create_task_blueprint_object_for_testing()

        subtask = create_observation_control_subtask_from_task_blueprint(task_blueprint)
        self.assertEqual("defined", str(subtask.state))
        self.assertEqual("observation control", str(subtask.specifications_template.name))
        self.assertEqual("observation", str(subtask.specifications_template.type))

        # Next call requires an observation subtask already created
        subtask = create_qafile_subtask_from_task_blueprint(task_blueprint)
        # subtask object is None because QA file conversion is by default not enabled!!!!
        self.assertEqual(None, subtask)

        # Next call will fail due to no qa_files object
        subtask = create_qaplots_subtask_from_task_blueprint(task_blueprint)
        # subtask object is None because QA file conversion is by default not enabled!!!!
        self.assertEqual(None, subtask)


    def test_create_sequence_of_subtask_from_task_blueprint_with_QA_enabled(self):
        """
        Create multiple subtasks from a task blueprint, executed in correct order.
        QA plots and QA file conversion enabled
        No exception should occur, check name, type and state of the subtasks
        """
        # Create Observation Task Enable QA plot and QA conversion
        task_blueprint = create_task_blueprint_object_for_testing(QA_enabled=True)
        task_blueprint_preprocessing = create_task_blueprint_object_for_testing("preprocessing pipeline")

        subtask = create_observation_control_subtask_from_task_blueprint(task_blueprint)
        self.assertEqual("defined", str(subtask.state))
        self.assertEqual("observation control", str(subtask.specifications_template.name))
        self.assertEqual("observation", str(subtask.specifications_template.type))
        # Next call requires an observation subtask already created
        subtask = create_qafile_subtask_from_task_blueprint(task_blueprint)
        self.assertEqual("defined", str(subtask.state))
        self.assertEqual("QA file conversion", str(subtask.specifications_template.name))
        self.assertEqual("qa_files", str(subtask.specifications_template.type))
        # Next call requires an qaplots subtask already created
        subtask = create_qaplots_subtask_from_task_blueprint(task_blueprint)
        self.assertEqual("defined", str(subtask.state))
        self.assertEqual("QA plots", str(subtask.specifications_template.name))
        self.assertEqual("qa_plots", str(subtask.specifications_template.type))
        # Next call will fail due to missing task relation
        with self.assertRaises(SubtaskCreationException):
            create_preprocessing_subtask_from_task_blueprint(task_blueprint_preprocessing)
        # Create that relation and check again
        create_relation_task_blueprint_object_for_testing(task_blueprint, task_blueprint_preprocessing)
        subtask = create_preprocessing_subtask_from_task_blueprint(task_blueprint_preprocessing)
        self.assertEqual("defined", str(subtask.state))
        self.assertEqual("preprocessing pipeline", str(subtask.specifications_template.name))
        self.assertEqual("pipeline", str(subtask.specifications_template.type))

    def test_create_subtasks_from_task_blueprint_succeed(self):
        """
        """
        task_blueprint = create_task_blueprint_object_for_testing(QA_enabled=True)
        subtasks = create_subtasks_from_task_blueprint(task_blueprint)
        self.assertEqual(3, len(subtasks))

    def test_create_subtasks_from_task_blueprint_translates_SAP_names(self):
        task_blueprint = create_task_blueprint_object_for_testing('target observation')
        task_blueprint.specifications_doc['SAPs'] = [{'name': 'target1', 'target': '', 'subbands': [],
                                                      'digital_pointing': {'angle1': 0.1, 'angle2': 0.1,
                                                                           'direction_type': 'J2000'}},
                                                     {'name': 'target2', 'target': '', 'subbands': [],
                                                      'digital_pointing': {'angle1': 0.2, 'angle2': 0.2,
                                                                           'direction_type': 'J2000'}}]
        subtask = create_observation_control_subtask_from_task_blueprint(task_blueprint)
        i = 0
        for sap in task_blueprint.specifications_doc['SAPs']:
            subtask_pointing = subtask.specifications_doc['stations']['digital_pointings'][i]
            self.assertEqual(sap['name'], subtask_pointing['name'])
            self.assertEqual(sap['digital_pointing']['angle1'], subtask_pointing['pointing']['angle1'])
            i += 1

class SubTasksCreationFromTaskBluePrintCalibrator(unittest.TestCase):

    def test_create_sequence_of_subtask_from_task_blueprint_calibrator_failure(self):
        """
        Create multiple subtasks from a task blueprint when task is a calibrator
        Check that exception should occur due too missing related target observation
        """
        task_blueprint = create_task_blueprint_object_for_testing(task_template_name="calibrator observation")
        with self.assertRaises(SubtaskCreationException):
            create_observation_control_subtask_from_task_blueprint(task_blueprint)

    @unittest.skip("JS 2020-09-08: Cannot reproduce SubtaskCreationException. How is this test supposed to work??")
    def test_create_sequence_of_subtask_from_task_blueprint_calibrator(self):
        """
        Create multiple subtasks from a task blueprint when task is a calibrator and is related to task blueprint
        of a target observation
        Check that exception should occur due too missing pointing setting in target observation,
        the calibrator default is AutoSelect=True
        Check NO exception, when AutoSelect=False
        """
        cal_task_blueprint = create_task_blueprint_object_for_testing(task_template_name="calibrator observation")
        target_task_blueprint = create_task_blueprint_object_for_testing()
        create_scheduling_relation_task_blueprint_for_testing(cal_task_blueprint, target_task_blueprint)

        with self.assertRaises(SubtaskCreationException):
            create_observation_control_subtask_from_task_blueprint(cal_task_blueprint)

        cal_task_blueprint.specifications_doc['autoselect'] = False
        cal_task_blueprint.specifications_doc['pointing']['angle1'] = 1.111
        cal_task_blueprint.specifications_doc['pointing']['angle2'] = 2.222
        subtask = create_observation_control_subtask_from_task_blueprint(cal_task_blueprint)
        self.assertEqual("defined", str(subtask.state))
        self.assertEqual("observation control", str(subtask.specifications_template.name))
        self.assertEqual("observation", str(subtask.specifications_template.type))
        self.assertEqual('J2000', subtask.specifications_doc['stations']['analog_pointing']['direction_type'])
        self.assertEqual(1.111, subtask.specifications_doc['stations']['analog_pointing']['angle1'])
        self.assertEqual(2.222, subtask.specifications_doc['stations']['analog_pointing']['angle2'])

    def test_create_combined_subtask_from_task_blueprints(self):
        """
        Create subtasks from a target task blueprint and a separate calibrator task blueprint.
        """
        cal_task_blueprint = create_task_blueprint_object_for_testing(task_template_name="calibrator observation")
        target_task_blueprint = create_task_blueprint_object_for_testing()
        create_scheduling_relation_task_blueprint_for_testing(cal_task_blueprint, target_task_blueprint, placement='parallel')

        subtask_1 = create_observation_control_subtask_from_task_blueprint(target_task_blueprint)
        num_pointings_target = len(subtask_1.specifications_doc['stations']['digital_pointings'])

        # assert target subtask still in defining state
        self.assertEqual("defining", str(subtask_1.state))

        subtask_2 = create_observation_control_subtask_from_task_blueprint(cal_task_blueprint)

        # assert the same subtask is returned
        self.assertEqual(subtask_1, subtask_2)

        # assert the calibrator obs was added as an additional beam
        num_pointings_calibrator = len(subtask_2.specifications_doc['stations']['digital_pointings'])
        self.assertEqual(num_pointings_target + 1, num_pointings_calibrator)

        # assert the subtask is now in defined state
        self.assertEqual("defined", str(subtask_2.state))

        # assert the subtask references both tasks
        self.assertEqual(subtask_1.task_blueprints.count(), 2)
        self.assertIn(target_task_blueprint, subtask_1.task_blueprints.all())
        self.assertIn(cal_task_blueprint, subtask_1.task_blueprints.all())

        # assert we have subtask outputs for both tasks
        self.assertEqual(subtask_1.outputs.count(), 2)
        self.assertEqual(subtask_1.outputs.filter(task_blueprint=target_task_blueprint).count(), 1)
        self.assertEqual(subtask_1.outputs.filter(task_blueprint=cal_task_blueprint).count(), 1)

    def test_create_combined_subtask_from_task_blueprints_fails_if_calibrator_handled_before_target(self):
        """
        Create subtasks from a target task blueprint and a separate calibrator task blueprint.
        Handling calibrator before target task should raise Exception.
        """
        cal_task_blueprint = create_task_blueprint_object_for_testing(task_template_name="calibrator observation")
        target_task_blueprint = create_task_blueprint_object_for_testing()
        create_scheduling_relation_task_blueprint_for_testing(cal_task_blueprint, target_task_blueprint, placement='parallel')

        with self.assertRaises(SubtaskCreationException) as cm:
            create_observation_control_subtask_from_task_blueprint(cal_task_blueprint)
            create_observation_control_subtask_from_task_blueprint(target_task_blueprint)

        self.assertIn("cannot be added to the target subtask, because it does not exist", str(cm.exception))

    def test_create_combined_subtask_from_task_blueprints_fails_if_calibrator_does_not_fit(self):
        """
        Create subtasks from a target task blueprint and a separate calibrator task blueprint.
        And exception is raised when the combined number of subbands exceeds 488.
        """
        cal_task_blueprint = create_task_blueprint_object_for_testing(task_template_name="calibrator observation")
        target_task_blueprint = create_task_blueprint_object_for_testing()
        create_scheduling_relation_task_blueprint_for_testing(cal_task_blueprint, target_task_blueprint, placement='parallel')

        target_task_blueprint.specifications_doc['SAPs'] = [{'name': 'target1', 'target': '', 'subbands': list(range(0, 150)),
                                                             'digital_pointing': {'angle1': 0.1, 'angle2': 0.1,
                                                                                  'direction_type': 'J2000'}},
                                                            {'name': 'target2', 'target': '', 'subbands': list(range(150, 300)),
                                                             'digital_pointing': {'angle1': 0.2, 'angle2': 0.2,
                                                                                  'direction_type': 'J2000'}}]
        target_task_blueprint.save()

        with self.assertRaises(SubtaskCreationException) as cm:
            create_observation_control_subtask_from_task_blueprint(target_task_blueprint)
            create_observation_control_subtask_from_task_blueprint(cal_task_blueprint)

        self.assertIn("results in 600 total subbands, but only 488 are possible", str(cm.exception))


class SubTasksCreationFromTaskBluePrintCalibrator(unittest.TestCase):

    def test_create_sequence_of_subtask_from_task_blueprint_calibrator_failure(self):
        """
        Create multiple subtasks from a task blueprint when task is a calibrator
        Check that exception should occur due too missing related target observation
        """
        task_blueprint = create_task_blueprint_object_for_testing(task_template_name="calibrator observation")
        with self.assertRaises(SubtaskCreationException):
            create_observation_control_subtask_from_task_blueprint(task_blueprint)

    @unittest.skip("JS 2020-09-08: Cannot reproduce SubtaskCreationException. How is this test supposed to work??")
    def test_create_sequence_of_subtask_from_task_blueprint_calibrator(self):
        """
        Create multiple subtasks from a task blueprint when task is a calibrator and is related to task blueprint
        of a target observation
        Check that exception should occur due too missing pointing setting in target observation,
        the calibrator default is AutoSelect=True
        Check NO exception, when AutoSelect=False
        """
        cal_task_blueprint = create_task_blueprint_object_for_testing(task_template_name="calibrator observation")
        target_task_blueprint = create_task_blueprint_object_for_testing()
        create_scheduling_relation_task_blueprint_for_testing(cal_task_blueprint, target_task_blueprint)

        with self.assertRaises(SubtaskCreationException):
            create_observation_control_subtask_from_task_blueprint(cal_task_blueprint)

        cal_task_blueprint.specifications_doc['autoselect'] = False
        cal_task_blueprint.specifications_doc['pointing']['angle1'] = 1.111
        cal_task_blueprint.specifications_doc['pointing']['angle2'] = 2.222
        subtask = create_observation_control_subtask_from_task_blueprint(cal_task_blueprint)
        self.assertEqual("defined", str(subtask.state))
        self.assertEqual("observation control", str(subtask.specifications_template.name))
        self.assertEqual("observation", str(subtask.specifications_template.type))
        self.assertEqual('J2000', subtask.specifications_doc['stations']['analog_pointing']['direction_type'])
        self.assertEqual(1.111, subtask.specifications_doc['stations']['analog_pointing']['angle1'])
        self.assertEqual(2.222, subtask.specifications_doc['stations']['analog_pointing']['angle2'])


class SubTaskCreationFromTaskBlueprintPipelines(unittest.TestCase):

    def test_create_subtask_from_task_blueprint_preprocessing_pipeline(self):
        """
        Test that a preprocessing task blueprint can be turned into a preprocessing pipeline subtask
        """

        # setup
        observation_task_blueprint = create_task_blueprint_object_for_testing()
        pipeline_task_blueprint = create_task_blueprint_object_for_testing(task_template_name="preprocessing pipeline")
        create_relation_task_blueprint_object_for_testing(observation_task_blueprint, pipeline_task_blueprint)

        create_observation_control_subtask_from_task_blueprint(observation_task_blueprint)

        # trigger
        subtask = create_preprocessing_subtask_from_task_blueprint(pipeline_task_blueprint)

        # assert
        self.assertEqual("defined", str(subtask.state))
        self.assertEqual("preprocessing pipeline", str(subtask.specifications_template.name))
        self.assertEqual(models.SubtaskType.Choices.PIPELINE.value, str(subtask.specifications_template.type))

    def test_create_subtask_from_task_blueprint_preprocessing_pipeline(self):
        """
        Test that ia pulsar task blueprint can be turned into a pulsar pipeline subtask
        """

        # setup
        observation_task_blueprint = create_task_blueprint_object_for_testing()
        pipeline_task_blueprint = create_task_blueprint_object_for_testing(task_template_name="pulsar pipeline")
        create_relation_task_blueprint_object_for_testing(observation_task_blueprint, pipeline_task_blueprint)

        create_observation_control_subtask_from_task_blueprint(observation_task_blueprint)

        # trigger
        subtask = create_pulsar_pipeline_subtask_from_task_blueprint(pipeline_task_blueprint)

        # assert
        self.assertEqual("defined", str(subtask.state))
        self.assertEqual("pulsar pipeline", str(subtask.specifications_template.name))
        self.assertEqual(models.SubtaskType.Choices.PIPELINE.value, str(subtask.specifications_template.type))


class SubTaskCreationFromTaskBlueprintIngest(unittest.TestCase):

    def test_create_subtask_from_task_blueprint_ingest(self):
        """
        Test that ingest task blueprint can be turned into a ingest control subtask
        """

        # setup
        ingest_task_blueprint = create_task_blueprint_object_for_testing(task_template_name="ingest")

        # trigger
        subtask = create_ingest_subtask_from_task_blueprint(ingest_task_blueprint)

        # assert
        self.assertEqual("defined", str(subtask.state))
        self.assertEqual("ingest control", str(subtask.specifications_template.name))
        self.assertEqual(models.SubtaskType.Choices.INGEST.value, str(subtask.specifications_template.type))


class SubtaskInputSelectionFilteringTest(unittest.TestCase):

    def setUp(self) -> None:
        # make sure we're allowed to schedule
        setting = Setting.objects.get(name='dynamic_scheduling_enabled')
        setting.value = True
        setting.save()

    def test_specifications_doc_meets_selection_doc_returns_true_on_empty_filter(self):
        specs = {}
        selection = {}
        self.assertTrue(specifications_doc_meets_selection_doc(specs, selection))

    def test_specifications_doc_meets_selection_doc_returns_true_when_filter_applies(self):
        # simple selection matches specs
        specs = {'sap': ['target0']}
        selection = {'sap': ['target0']}
        self.assertTrue(specifications_doc_meets_selection_doc(specs, selection))

        # extra specs are ignored
        specs = {'sap': ['target0'], 'not': 'relevant'}
        selection = {'sap': ['target0']}
        self.assertTrue(specifications_doc_meets_selection_doc(specs, selection))

        # complex selection matches specs; multiple keys and values
        specs = {'sap': ['target0'], 'is_relevant': True}
        selection = {'sap': ['target0', 'target1'], 'is_relevant': True}
        self.assertTrue(specifications_doc_meets_selection_doc(specs, selection))

    def test_specifications_doc_meets_selection_doc_returns_true_when_filter_does_not_apply(self):
        # selection mismatches specs
        specs = {'sap': ['target0']}
        selection = {'sap': ['target1']}
        self.assertFalse(specifications_doc_meets_selection_doc(specs, selection))

        # spec only partially selected
        specs = {'sap': ['target0', 'target1']}
        selection = {'sap': ['target1']}
        self.assertFalse(specifications_doc_meets_selection_doc(specs, selection))

        # selection not in specs
        specs = {'sap': ['target0']}
        selection = {'sap': ['target0'], 'is_relevant': True}
        self.assertFalse(specifications_doc_meets_selection_doc(specs, selection))

    def test_links_to_log_files(self):
        """
        Test redirect urls to subtask logfiles.
        """

        # the link to log files is a 'view' on the subtask, and NOT part of the subtask model.
        # the link is served as an action on the REST API, redirecting to externally served log files.
        # check/test the redirect urls.
        with tmss_test_env.create_tmss_client() as client:
            # observation
            subtask_observation = create_subtask_object_for_testing("observation")
            response = client.session.get(url=client.get_full_url_for_path('/subtask/%s/task_log' % (subtask_observation.id,)), allow_redirects=False)
            self.assertTrue(response.is_redirect)
            self.assertIn("proxy.lofar.eu", response.headers['Location'])
            self.assertIn("rtcp-%s.errors" % subtask_observation.id, response.headers['Location'])

            # pipeline
            subtask_pipeline = create_subtask_object_for_testing("pipeline")
            response = client.session.get(url=client.get_full_url_for_path('/subtask/%s/task_log' % (subtask_pipeline.id,)), allow_redirects=False)
            self.assertEqual(404, response.status_code) # no log (yet) for unscheduled pipeline

            # other (qa_plots)
            subtask_qa_plots = create_subtask_object_for_testing("qa_plots")
            self.assertEqual(404, response.status_code) # no log for other subtasktypes


class SettingTest(unittest.TestCase):

    def test_schedule_observation_subtask_raises_when_flag_is_false(self):
        setting = Setting.objects.get(name='dynamic_scheduling_enabled')
        setting.value = False
        setting.save()
        obs_st = create_subtask_object_for_testing('observation')

        with self.assertRaises(SubtaskSchedulingException):
            schedule_observation_subtask(obs_st)


class SubTaskCreationFromTaskBlueprintBeamformer(unittest.TestCase):
    saps = [{"name": "target1", "target": "", "subbands": [349, 372],
             "digital_pointing": {"angle1": 0.24, "angle2": 0.25, "direction_type": "J2000"}},
            {"name": "target2", "target": "", "subbands": [309, 302],
             "digital_pointing": {"angle1": 0.42, "angle2": 0.52, "direction_type": "J2000"}}
    ]
    beamformers = [{"name": "beamformer1",
                    "coherent": {"settings": {"stokes": "I", "time_integration_factor": 8, "subbands_per_file": 244,
                                              "channels_per_subband": 8, "quantisation_enabled": False, "quantisation": {}},
                    "SAPs": [{"name": "target1",
                              "tabs": [{"relative": True,
                                        "pointing": {"direction_type": "J2000", "angle1": 0.1,  "angle2": 0.1}},
                                       {"relative": False, "pointing": {"direction_type": "J2000", "angle1": 0.2,  "angle2": 0.2}}],
                              "tab_rings": {"count": 8, "width": 0.02},
                              "subbands": {"list":[1,2,3], "method": "copy"}},
                             {"name": "target1",
                              "tabs": [{"relative": True,
                                        "pointing": {"direction_type": "J2000", "angle1": 0.1,  "angle2": 0.1}},
                                       {"relative": False, "pointing": {"direction_type": "J2000", "angle1": 0.2,  "angle2": 0.2}}],
                              "tab_rings": {"count": 7, "width": 0.03},
                              "subbands": {"list":[10,20,30], "method": "copy"}
                              }]},
                    "incoherent": {"settings": {"stokes": "I", "time_integration_factor": 4, "subbands_per_file": 244,
                                                "channels_per_subband": 8, "quantisation_enabled": False, "quantisation": {}},
                                   "SAPs": [{"name": "target1", "subbands": {"list":[4,5,6], "method": "copy"}}]},
                    "stations": ["CS001"]},
                   {"name": "beamformer2",
                    "flys_eye_enabled": True,
                    "flys eye": {"settings": {"stokes": "I", "time_integration_factor": 2, "subbands_per_file": 122,
                                              "channels_per_subband": 16, "quantisation_enabled": False, "quantisation": {}}},
                    "stations": ["DE609"]}]

    # todo: fix and enable test once we have a switch between standalone/add-on
    @unittest.skip('currently we hardcode standalone mode, where this test cannot be expected to fail')
    def test_create_sequence_of_subtask_from_task_blueprint_beamformer_failure(self):
        """
        Create a subtasks from a task blueprint when task is beamformer add-on.
        Assert that this fails without related target observation.
        # todo: mark as add-on when we have a switch
        """
        task_blueprint = create_task_blueprint_object_for_testing(task_template_name="beamforming observation")
        task_blueprint.specifications_doc['SAPs'] = self.saps
        task_blueprint.specifications_doc['beamformers'] = self.beamformers
        with self.assertRaises(SubtaskCreationException):
            create_observation_control_subtask_from_task_blueprint(task_blueprint)

    def test_get_related_target_sap_by_name(self):
        beamformer_task_blueprint = create_task_blueprint_object_for_testing(task_template_name="beamforming observation")
        beamformer_task_blueprint.specifications_doc['SAPs'] = self.saps
        beamformer_task_blueprint.specifications_doc['beamformers'] = self.beamformers
        target_task_blueprint = create_task_blueprint_object_for_testing()
        target_task_blueprint.specifications_doc['SAPs'] = self.saps
        target_task_blueprint.save()
        create_scheduling_relation_task_blueprint_for_testing(beamformer_task_blueprint, target_task_blueprint)

        # TODO: If we start using beamforming observations in parallel with target imaging observations, then we need to search for saps in the target imaging obs spec.
        # See git history for an initial implementation.
        # Also see git history for a test 'test_create_sequence_of_subtask_from_task_blueprint_beamformer' where we have a beamforming observation in parallel with target imaging observation
        sap = _get_related_target_sap_by_name(beamformer_task_blueprint, 'target2')
        self.assertEqual(sap, self.saps[1])

    def test_generate_tab_ring_pointings_returns_correct_pointings(self):

        pointing = {"angle1": 0.11, "angle2": 0.22, "direction_type": "J2000"}
        tab_rings = {"width": 1, "count": 1}

        # assert center pointing is returned
        tab_pointings = _generate_tab_ring_pointings(pointing, tab_rings)
        self.assertIn(pointing, tab_pointings)

        # assert correct number of pointings is returned
        self.assertEqual(len(tab_pointings), 1+6)  # center + 1 ring
        tab_rings.update({'count': 3})
        tab_pointings = _generate_tab_ring_pointings(pointing, tab_rings)
        self.assertEqual(len(tab_pointings), 1+6+12+18)  # center + 3 rings

        # assert width is considered
        tab_rings.update({'width': 42})
        tab_pointings = _generate_tab_ring_pointings(pointing, tab_rings)
        pointing.update({'angle2': pointing['angle2']+tab_rings['width']})
        self.assertIn(pointing, tab_pointings)

    def test_add_pointings_adds_correctly(self):

        pointing_a = {"angle1": 0.11, "angle2": 0.22, "direction_type": "J2000"}
        pointing_b = {"angle1": 0.88, "angle2": 0.66, "direction_type": "J2000"}
        pointing_sum = _add_pointings(pointing_a, pointing_b)
        self.assertEqual(pointing_sum, {"angle1": 0.99, "angle2": 0.88, "direction_type": "J2000"})

    def test_filter_subbands_filters_correctly(self):
        subbands = [1,3,4,5,10,11,12,13,19,20]

        # copy
        subband_selection = {'method': 'copy'}
        filtered_subbands = _filter_subbands(subbands, subband_selection)
        self.assertEqual(filtered_subbands, subbands)

        # subset
        subband_selection = {'method': 'subset', 'list': [1,2,3,4,5,6,7,8,9,10]}
        filtered_subbands = _filter_subbands(subbands, subband_selection)
        self.assertEqual(filtered_subbands, [1,3,4,5,10])

        # largest continuous subset
        subband_selection = {'method': 'largest continuous subset'}
        filtered_subbands = _filter_subbands(subbands, subband_selection)
        self.assertEqual(filtered_subbands, [10,11,12,13])

class SubtaskAllowedStateTransitionsTest(unittest.TestCase):
    def test_successful_path(self):
        subtask = models.Subtask.objects.create(**Subtask_test_data())
        for state_value in (SubtaskState.Choices.DEFINING.value,
                            SubtaskState.Choices.DEFINED.value,
                            SubtaskState.Choices.SCHEDULING.value,
                            SubtaskState.Choices.SCHEDULED.value,
                            SubtaskState.Choices.QUEUEING.value,
                            SubtaskState.Choices.QUEUED.value,
                            SubtaskState.Choices.STARTING.value,
                            SubtaskState.Choices.STARTED.value,
                            SubtaskState.Choices.FINISHING.value,
                            SubtaskState.Choices.FINISHED.value):
            subtask.state = SubtaskState.objects.get(value=state_value)
            # no SubtaskIllegalStateTransitionException should be raised upon save. If it is raised, then test fails. No need for asserts.
            subtask.save()

    def test_helper_method_set_subtask_state_following_allowed_transitions_successful_path(self):
        for state_value in (SubtaskState.Choices.DEFINING.value,
                            SubtaskState.Choices.DEFINED.value,
                            SubtaskState.Choices.SCHEDULING.value,
                            SubtaskState.Choices.SCHEDULED.value,
                            SubtaskState.Choices.QUEUEING.value,
                            SubtaskState.Choices.QUEUED.value,
                            SubtaskState.Choices.STARTING.value,
                            SubtaskState.Choices.STARTED.value,
                            SubtaskState.Choices.FINISHING.value,
                            SubtaskState.Choices.FINISHED.value):
            # start with subtask in defining state each time
            subtask = models.Subtask.objects.create(**Subtask_test_data(state=SubtaskState.objects.get(value=SubtaskState.Choices.DEFINING.value)))
            self.assertEqual(SubtaskState.Choices.DEFINING.value, subtask.state.value)

            set_subtask_state_following_allowed_transitions(subtask, state_value)
            self.assertEqual(state_value, subtask.state.value)

    def test_helper_method_set_subtask_state_following_allowed_transitions_error_path(self):
        for intermediate_state_value in (SubtaskState.Choices.DEFINING.value,
                                         SubtaskState.Choices.SCHEDULING.value,
                                         SubtaskState.Choices.UNSCHEDULING.value,
                                         SubtaskState.Choices.QUEUEING.value,
                                         SubtaskState.Choices.STARTING.value,
                                         SubtaskState.Choices.STARTED.value,
                                         SubtaskState.Choices.FINISHING.value):
            # start with subtask in defining state each time
            subtask = models.Subtask.objects.create(**Subtask_test_data(state=SubtaskState.objects.get(value=SubtaskState.Choices.DEFINING.value)))
            self.assertEqual(SubtaskState.Choices.DEFINING.value, subtask.state.value)

            # then go to the requested intermediate state
            set_subtask_state_following_allowed_transitions(subtask, intermediate_state_value)
            self.assertEqual(intermediate_state_value, subtask.state.value)

            # then go to the error state (should be allowed from any of these intermediate states)
            subtask.state = SubtaskState.objects.get(value=SubtaskState.Choices.ERROR.value)
            subtask.save()
            self.assertEqual(SubtaskState.Choices.ERROR.value, subtask.state.value)

    def test_helper_method_set_subtask_state_following_allowed_transitions_cancel_path(self):
        for desired_end_state_value in (SubtaskState.Choices.CANCELLING.value,SubtaskState.Choices.CANCELLED.value):
            for state_value in (SubtaskState.Choices.DEFINED.value,
                                SubtaskState.Choices.SCHEDULED.value,
                                SubtaskState.Choices.QUEUED.value,
                                SubtaskState.Choices.STARTED.value):
                # start with subtask in defining state each time
                subtask = models.Subtask.objects.create(**Subtask_test_data(state=SubtaskState.objects.get(value=SubtaskState.Choices.DEFINING.value)))
                self.assertEqual(SubtaskState.Choices.DEFINING.value, subtask.state.value)

                # then go to the requested intermediate state
                set_subtask_state_following_allowed_transitions(subtask, state_value)
                self.assertEqual(state_value, subtask.state.value)

                # then go to the error state (should be allowed from any of these intermediate states)
                set_subtask_state_following_allowed_transitions(subtask, desired_end_state_value)
                self.assertEqual(desired_end_state_value, subtask.state.value)

    def test_helper_method_set_subtask_state_following_allowed_transitions_unscheduling_path(self):
        # start with subtask in defining state
        subtask = models.Subtask.objects.create(**Subtask_test_data(state=SubtaskState.objects.get(value=SubtaskState.Choices.DEFINING.value)))
        self.assertEqual(SubtaskState.Choices.DEFINING.value, subtask.state.value)

        # use helper method to follow the allowed path to 'unscheduling'
        set_subtask_state_following_allowed_transitions(subtask, SubtaskState.Choices.UNSCHEDULING.value)
        self.assertEqual(SubtaskState.Choices.UNSCHEDULING.value, subtask.state.value)

        # check transition path
        state_log = SubtaskStateLog.objects.filter(subtask=subtask).order_by('created_at').all()
        self.assertEqual(SubtaskState.Choices.DEFINING.value, state_log[0].new_state.value)
        self.assertEqual(SubtaskState.Choices.DEFINED.value, state_log[1].new_state.value)
        self.assertEqual(SubtaskState.Choices.SCHEDULING.value, state_log[2].new_state.value)
        self.assertEqual(SubtaskState.Choices.SCHEDULED.value, state_log[3].new_state.value)
        self.assertEqual(SubtaskState.Choices.UNSCHEDULING.value, state_log[4].new_state.value)

    def test_end_states(self):
        '''Check if the end states that we cannot get out of are according to the design'''
        # there should be no state to go to from ERROR
        self.assertEqual(0, SubtaskAllowedStateTransitions.objects.filter(old_state__value=SubtaskState.Choices.UNSCHEDULABLE.value).count())

        # there should be no state to go to from FINISHED
        self.assertEqual(0, SubtaskAllowedStateTransitions.objects.filter(old_state__value=SubtaskState.Choices.FINISHED.value).count())

        # there should be no state to go to from CANCELLED
        self.assertEqual(0, SubtaskAllowedStateTransitions.objects.filter(old_state__value=SubtaskState.Choices.CANCELLED.value).count())

    def test_illegal_state_transitions(self):
        for state_value in [choice.value for choice in SubtaskState.Choices]:
            # assume helper method set_subtask_state_following_allowed_transitions is working (see other tests above)
            # use it to create subtask in desired initial state
            subtask = models.Subtask.objects.create(**Subtask_test_data(state=SubtaskState.objects.get(value=SubtaskState.Choices.DEFINING.value)))
            subtask = set_subtask_state_following_allowed_transitions(subtask, state_value)
            self.assertEqual(state_value, subtask.state.value)

            # derive the allowed and illegal state transitions states
            allowed_new_states = SubtaskAllowedStateTransitions.allowed_new_states(subtask.state)
            illegal_new_states = SubtaskAllowedStateTransitions.illegal_new_states(subtask.state)
            logger.info("test_illegal_state_transitions: old_state='%s' allowed_new_states=%s illegal_new_states=%s", state_value, [s.value for s in allowed_new_states], [s.value for s in illegal_new_states])

            for illegal_new_state in illegal_new_states:
                subtask.state = illegal_new_state
                # check that the SubtaskIllegalStateTransitionException is raise for this illegal new state
                with self.assertRaises(SubtaskIllegalStateTransitionException):
                    subtask.save()

                # state in database should still be the original
                subtask.refresh_from_db()
                self.assertEqual(state_value, subtask.state.value)

if __name__ == "__main__":
    os.environ['TZ'] = 'UTC'
    unittest.main()
