#!/usr/bin/env python3

# Copyright (C) 2018    ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

# $Id:  $

# This test makes verifies that a user talking to the API can authenticate programmatically, e.g. from a script.
# The other functional tests use

from datetime import datetime
import os
import unittest
import logging
logger = logging.getLogger(__name__)
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

from lofar.common.test_utils import exit_with_skipped_code_if_skip_integration_tests
exit_with_skipped_code_if_skip_integration_tests()

# Do Mandatory setup step:
# use setup/teardown magic for tmss test database, ldap server and django server
# (ignore pycharm unused import statement, python unittests does use at RunTime the tmss_test_environment_unittest_setup module)
from lofar.sas.tmss.test.tmss_test_environment_unittest_setup import *
tmss_test_env.populate_permissions()

# import and setup test data creator
from lofar.sas.tmss.test.tmss_test_data_rest import TMSSRESTTestDataCreator
test_data_creator = TMSSRESTTestDataCreator(BASE_URL, AUTH)

from lofar.sas.tmss.client.tmss_http_rest_client import TMSSsession

from lofar.common.test_utils import integration_test

# todo: figure out why csrftoken is missing when using the TMSSRESTTestDataCreator while it is present when...
# todo: ...running tmss manually and referring to that by overriding BASE_URL here:
# BASE_URL = 'http://localhost:8008/api'

class LDAPSession(unittest.TestCase):

    def test_failure_using_no_credentials(self):
        with requests.Session() as session:
            import time
            time.sleep(20)
            session.verify = False
            r = session.get(BASE_URL + '/task_draft/?format=api')
            self.assertEqual(r.status_code, 401)
            self.assertTrue("Authentication credentials were not provided" in r.content.decode('utf8'))

    def test_failure_using_wrong_credentials(self):
        with requests.Session() as session:
            session.verify = False
            session.auth = (AUTH.username, 'wrong')
            r = session.get(BASE_URL + '/task_draft/?format=api')
            self.assertEqual(r.status_code, 401)
            self.assertTrue("Invalid username/password" in r.content.decode('utf8'))

    def test_success_using_correct_credentials(self):
        with requests.Session() as session:
            session.verify = False
            session.auth = (AUTH.username, AUTH.password)
            r = session.get(BASE_URL + '/task_draft/?format=api')
            self.assertEqual(r.status_code, 200)
            self.assertTrue("Task Draft List" in r.content.decode('utf8'))

@integration_test
class OIDCSession(unittest.TestCase):
    def setUp(self):
        if "OIDC_RP_CLIENT_ID" not in os.environ.keys():
            raise unittest.SkipTest("Cannot run OIDCSession-test when OIDC_RP_CLIENT_ID is not set in the environment to configure django settings")

    @integration_test
    def test_failure_without_authentication(self):
        with requests.Session() as session:
            session.verify = False
            r = session.get(BASE_URL + '/task_draft/?format=api')
            self.assertEqual(r.status_code, 401)
            self.assertTrue("Authentication credentials were not provided" in r.content.decode('utf8'))

    @integration_test
    def test_failure_using_wrong_credentials(self):
        with self.assertRaises(ValueError) as err:
            with TMSSsession(AUTH.username, 'wrong', BASE_URL.replace('/api', '')) as session:
                pass

        self.assertTrue('The username and/or password you specified are not correct' in str(err.exception))

    @integration_test
    def test_success_using_correct_credentials(self):
        with TMSSsession(AUTH.username, AUTH.password, BASE_URL.replace('/api', '')).session as session:
            r = session.get(BASE_URL + '/task_draft/?format=api')
            self.assertEqual(r.status_code, 200)
            self.assertTrue("Task Draft List" in r.content.decode('utf8'))


class TokenAuthTestCase(unittest.TestCase):

    def test_failure_using_no_credentials(self):
        r = requests.post(url=BASE_URL + '/token-auth/', json={'username':'', 'password':''})
        self.assertEqual(r.status_code, 400)

    def test_failure_using_wrong_credentials(self):
        r = requests.post(url=BASE_URL + '/token-auth/', json={'username': 'somebeautifulname', 'password': 'whatever'})
        self.assertEqual(r.status_code, 400)

    def test_success_using_correct_credentials(self):
        r = requests.post(url=BASE_URL + '/token-auth/', json={'username': AUTH.username, 'password': AUTH.password})
        self.assertEqual(r.status_code, 200)
        self.assertTrue("token" in r.content.decode('utf8'))

    def test_failure_using_no_token(self):
        r = requests.get(url=BASE_URL)
        self.assertEqual(r.status_code, 401)

    def test_failure_using_wrong_token(self):
        r = requests.post(url=BASE_URL + '/token-auth/', json={'username': AUTH.username, 'password': AUTH.password})
        token = r.json()['token']
        wrong_token = '97f538d65f0c58f1cd04f34809d79abc35cc4bf9'
        wrong_token = '87f538d65f0c58f1cd04f34809d79abc35cc4bf9' if token == wrong_token else wrong_token

        headers = {'Authorization': f'Token {wrong_token}'}
        r = requests.get(url=BASE_URL, headers=headers)
        self.assertEqual(r.status_code, 401)

    def test_success_using_correct_token(self):
        r = requests.post(url=BASE_URL + '/token-auth/', json={'username': AUTH.username, 'password': AUTH.password})
        token = r.json()['token']

        headers = {'Authorization': f'Token {token}'}
        r = requests.get(url=BASE_URL, headers=headers)
        self.assertEqual(r.status_code, 200)


class TokenDeauthTestCase(unittest.TestCase):

    def test_failure_using_no_token(self):
        r = requests.delete(url=BASE_URL + '/token-deauth/')
        self.assertEqual(r.status_code, 401)

    def test_failure_using_wrong_token(self):
        r = requests.post(url=BASE_URL + '/token-auth/', json={'username': AUTH.username, 'password': AUTH.password})
        token = r.json()['token']
        wrong_token = '97f538d65f0c58f1cd04f34809d79abc35cc4bf9'
        wrong_token = '87f538d65f0c58f1cd04f34809d79abc35cc4bf9' if token == wrong_token else wrong_token

        headers = {'Authorization': f'Token {wrong_token}'}
        r = requests.delete(url=BASE_URL + '/token-deauth/', headers=headers)
        self.assertEqual(r.status_code, 401)

    def test_success_using_correct_token(self):
        r = requests.post(url=BASE_URL + '/token-auth/', json={'username': AUTH.username, 'password': AUTH.password})
        token = r.json()['token']

        headers = {'Authorization': f'Token {token}'}
        r = requests.delete(url=BASE_URL + '/token-deauth/', headers=headers)
        self.assertEqual(r.status_code, 204)
        # Check if deleted token is no more valid
        r = requests.get(url=BASE_URL, headers=headers)
        self.assertEqual(r.status_code, 401)


if __name__ == "__main__":
    unittest.main()

